#!/usr/bin/perl -w

use strict;

my $config_out = 'configure.make';
my $cp         = 'cp';
my $Makefile   = 'Makefile';

###############################################################################
# initialize
my $command = {};
my $option  = {};
my $dirs    = {};
my %lookup  = ();

###############################################################################
# template list
my $def_temp_dir = 'template/def';
my @def_template = qw( constant_math.def
		       constant_physics.def
		       equations_CReSS.def
		       equations_radar_amedas.def
		       level_pressure.def
		       model_geo_info_CReSS.def
		       model_geo_info_CReSS_parallel.def
		       model_info_CReSS.def
		       radar_amedas.def
		       );

my @namelist_temp = qw( test.namelist );

###############################################################################
# directory list
my @dirlist = qw( Doc bin cpt data def out perl src template util );

###############################################################################
# requirement check for making program
my @FC = qw( f95 f90 ifort pgf90 pgf95 frt ifc g95 );
my %FCFLAGS = (
	       'f95'      => '',
	       'f90'      => '',
	       'ifc'      => '-assume byterecl -convert big_endian',
	       'pgf90'    => '-O -byteswapio',
	       'pgf95'    => '-O -byteswapio',
	       'frt'      => '-Am -X9',
	       'ifort'    => '-assume byterecl -convert big_endian',
	       'g95'      => '-fendian=big -DNOFFT',
	       'gfortran' => '-convert=big-endian -DNOFFT'
	       );
my @AR = qw( ar );
my @LATEX = qw( platex );
my @DVIPDF = qw( dvipdfmx );
my @CDFINC = qw(
		/usr/local/netcdf_ifc/include
		/usr/local/netcdf_ifort/include
		/usr/local/netcdf/include
		/usr/include
		);
my @CDFLIB = qw(
		/usr/local/netcdf_ifc/lib
		/usr/local/netcdf_ifort/lib
		/usr/local/netcdf/lib
		/usr/lib
		);

$command->{FC}     = \@FC;
$option->{FCFLAGS} = \%FCFLAGS;
$lookup{FCFLAGS}   = 'FC';
$command->{AR}     = \@AR;
$command->{LATEX}  = \@LATEX;
$command->{DVIPDF} = \@DVIPDF;
$dirs->{CDFINC} = \@CDFINC;
$dirs->{CDFLIB} = \@CDFLIB;

my @check  = qw( FC FCFLAGS AR CDFINC CDFLIB );

print "PATH:\n";
print $ENV{PATH}, "\n\n";
my @PATH = split( /\:/, $ENV{PATH} );

# cp check
my $cp_bak = $cp;
$cp = &InquireVariable( $cp_bak, \@PATH );
if( $cp eq '' ){
    print STDERR "copy command $cp_bak not found\n";
    print STDERR "please chekc \$cp in configure.pl\n";
    exit;
}

###############################################################################
# directory check
foreach my $dir ( @dirlist ){
    unless( -d $dir ){
	mkdir( $dir, 0755 );
    }
}

###############################################################################
# copy template file
foreach my $def ( @def_template ){
    my $from = "${def_temp_dir}/${def}";
    my $to   = "def/${def}";
    next unless ( -f $from );
    unless( -f $to ){
	system( "$cp $from $to" );
    }
}

foreach my $namelist ( @namelist_temp ){
    my $from = "template/${namelist}";
    my $to   = "./${namelist}";
    next unless ( -f $from );
    unless( -f $to ){
	system( "$cp $from $to" );
    }
}

###############################################################################
# make dependency check
if( -f $config_out ){
    print STDERR "$config_out already exists\n";
    exit;
}

if( -f $Makefile ){
    my $bak = $Makefile . '.bak';
    system( "$cp $Makefile $bak" );
}

my $config = {};
foreach my $component ( @check ){
    no strict 'refs';
    print "check $component ... ";

   # program check
   if( defined( $command->{$component} ) ){
       next if( ref( $command->{$component} ) ne 'ARRAY' );
       $config->{$component} = '';
       foreach my $command ( @{$command->{$component}} ){
	   ( $config->{$component} )
	       = &InquireVariable( $command, \@PATH );
	   last if( $config->{$component} ne '' );
       }
       print "[$config->{$component}]: ";
       chomp( my $stdin = <STDIN> );
       if( $stdin eq '' && $config->{$component} eq '' ){
	   print STDERR "Error: unknown $component option\n";
	   exit;
       }elsif( $stdin ne '' ){
	   if( $stdin =~ /\// ){
	       if( -e $stdin ){
		   $config->{$component} = $stdin;
	       }
	   }else{
	       ( $config->{$component} ) = &InquireVariable( $stdin, \@PATH );
	   }
	   if( $config->{$component} eq '' ){
	       print STDERR "Error: unknown $component option\n";
	       exit;
	   }
       }

       if( $component eq 'FC' ){
	   if( $config->{$component} =~ /g95/ ||
	       $config->{$component} =~ /gfortran/ ){
	       system( "$cp src/makefile.nofft src/Makefile" );
	   }else{
	       system( "$cp src/makefile.default src/Makefile" );
	   }
       }

    # directry check
   }elsif( defined( $dirs->{$component} ) ){
       $config->{$component} = '';
       my $ndirs = @{$dirs->{$component}};
       for( my $i = 0 ; $i < $ndirs ; $i++ ){
	   my $dir = ${$dirs->{$component}}[$i];
	   if( -d $dir ){
	       $config->{$component} = $dir;
	       last;
	   }
       }

#       foreach my $dir ( @{$dirs->{$component}} ){
#	   if( -d $dir ){
#	       $config->{$component} = $dir;
#	   }
#       }

       print "[$config->{$component}]: ";
       chomp( my $stdin = <STDIN> );
       if( $stdin eq '' && $config->{$component} eq '' ){
	   print STDERR "Error: unknown $component option\n";
	   exit;
       }elsif( $stdin ne '' ){
	   if( $stdin =~ /\// ){
	       if( -e $stdin ){
		   $config->{$component} = $stdin;
	       }
	   }else{
	       ( $config->{$component} ) = &InquireVariable( $stdin, \@PATH );
	   }
	   if( $config->{$component} eq '' ){
	       print STDERR "Error: unknown $component option\n";
	       exit;
	   }
       }

    # option check
    }elsif( defined( $option->{$component} ) ){
	if( defined( $lookup{$component} ) ){
	    my $lookup = $lookup{$component};
	    $config->{$lookup} =~ /\/([^\/]+)$/;
	    my $command = $1;

	    if( defined( $option->{$component}{$command} ) ){
		$config->{$component} = $option->{$component}{$command};
	    }else{
		$config->{$component} = '';
	    }
	}else{
	    $config->{$component} = '';
	}

	print "[$config->{$component}]:";
	chomp( my $stdin = <STDIN> );
	if( $stdin ne '' ){
	    $config->{$component} = $stdin;
	}
    }

    print " $config->{$component}\n";
}

###############################################################################
# conigure.make output
open( CONFIG, "> $config_out" ) || die;
while( <DATA> ){
    chomp;
    if( /\@([^\@]+)\@/ ){
	my $opt = $1;
	my $out = '';
	if( defined( $config->{$opt} ) ){
	    $out = $config->{$opt};
	}
	s/\@${opt}\@/$out/g;
    }
    print CONFIG "$_\n";
}
close( CONFIG );

# copy Makefile
if( -f "Makefile.in" ){
    system( "$cp Makefile.in $Makefile" );
}else{
    print STDERR "Makefile.in is not found.";
}
exit;

# command check
sub InquireVariable{
    my ( $var, $PATH ) = @_;

    foreach my $path ( @{$PATH} ){
	my $command = "${path}/$var";
	$command =~ s/\/\//\//g;
	if( -e $command ){
	    return ( $command );
	}
    }

    return ( '' );
}

__DATA__

#
# Compiler configurations
#

#
# fortran90 compiler
#
FC       = @FC@
FCFLAGS  = @FCFLAGS@
#
# C compiler
#
CC       = gcc
CFLAGS   =
#CFLAGS   = -DDEC
#
# General comand
#
AR       = @AR@
INCLUDE = -Ilibmm5 -I. -I@CDFINC@
LIBDIR  = -Llibmm5 -L@CDFLIB@
LIBRARY = -lmm5tool -lnetcdf
#INCLUDE  = -I. -Ilibmm5 -I/usr/local/netcdf_ifc/include
#LIBDIR   = -Llibmm5 -L/usr/local/netcdf_ifc/lib
#LIBRARY  = -lnetcdf -lmm5tool

#
# user function
#
USER_FUNC = NO_USER_FUNC
#USER_FUNC = USER_FUNC
USER_OBJ  =

#
# for document
#
LATEX  = @LATEX@
DVIPDF = @DVIPDF@

.SUFFIXES: .f90 .F .F90 .o .f .c

.c.o:
	$(CC) $(CFLAGS) -c $*.c
.f90.o:
	$(FC) $(FCFLAGS) $(INCLUDE) -c $*.f90
.f.o:
	$(FC) $(FCFLAGS) $(INCLUDE) -c $*.f
.F.o:
	$(FC) $(FCFLAGS) $(INCLUDE) -c $*.F
.F90.o:
	$(FC) $(FCFLAGS) -D$(USER_FUNC) $(INCLUDE) -c $*.F90
