module GRIB1
  use BIN
  implicit none
  private

  type GRIB1Sec0
     integer :: Edition
     integer :: TotalLength
  end type GRIB1Sec0

  type GRIB1Sec1
     integer :: Length
     integer :: TableVersion
     integer :: IdentificationCenter
     integer :: GPID ! Generating process ID number
     integer :: GID  ! Grid identification
     integer :: flag ! Flag specifying the presence or absence a GDS or a BMS
     integer :: IPU  ! Indicator of parameter and units
     integer :: LevID ! Indicator of level or layer
     integer :: Lev1
     integer :: Lev2
     integer :: Year
     integer :: Month
     integer :: Day
     integer :: Hour
     integer :: Minute
     integer :: TermUnit ! Forecast time unit
     integer :: P1
     integer :: P2
     integer :: TermID ! Time range indicator
     integer :: TermNum ! Number included in average
     integer :: TermLack ! Number missing from averages or accumulations
     integer :: IdSubCenter
     integer :: D
  end type GRIB1Sec1

  type GRIB1Sec2
     integer :: Length
     integer :: NV ! the number of coordinate parameters
     integer :: PV ! the location of the list of vertical coordinate parameters
     integer :: Form ! Data representation type
     integer :: Nx
     integer :: Ny
     integer :: La1 ! Latitude of first grid point
     integer :: Lo1 ! Longitude of first grid point
     integer :: ResAndComp ! Resolution and component flags
     integer :: IDi
     integer :: IVec
     integer :: Lov
     integer :: Dx
     integer :: Dy
     integer :: ProjCenter ! projection center flag
     integer :: scan
     integer :: is
     integer :: js
     integer :: seq
     integer :: Latin1 ! the first true latitude
     integer :: Latin2 ! the second true latitude
     integer :: SPlat  ! latitude of south pole  ( millidegrees )
     integer :: SPlon  ! longitude of south pole ( millidegrees )
  end type GRIB1Sec2

  type GRIB1Sec3
     integer :: Length
  end type GRIB1Sec3

  type GRIB1Sec4
     integer :: Length
     integer :: SkipBit
     logical :: SHC
     logical :: SmplPack
     logical :: OrgReal
     logical :: AddFlag
     integer :: E
     real    :: RefReal
     integer :: BitNum
  end type GRIB1Sec4

  type GRIB1List
     integer :: no
     integer :: HeaderTop
     integer :: HeaderLine
     integer :: MaxLine
     integer :: TotalLength
     integer :: S0rec, S0pos, S0length
     integer :: S1rec, S1pos, S1length
     integer :: S2rec, S2pos, S2length
     integer :: S3rec, S3pos, S3length
     integer :: S4rec, S4pos, S4length
     integer :: S5rec, S5pos, S5length
     character(len=24) :: date
     integer           :: IPU
     integer           :: LevID
     integer           :: Lev1
     integer           :: Lev2
     type(GRIB1List), pointer :: next
  end type GRIB1List

  type GRIB1DAT
     integer         :: HeaderTop
     integer         :: HeaderLine
     integer         :: MaxLine
     logical         :: ByteSwap
     type(GRIB1Sec0) :: Sec0
     type(GRIB1Sec1) :: Sec1
     type(GRIB1Sec2) :: Sec2
     type(GRIB1Sec3) :: Sec3
     type(GRIB1Sec4) :: Sec4
     type(GRIB1List) :: GT
     integer :: pole
     real    :: er
     real    :: of
     real, dimension(:,:), pointer :: value
  end type GRIB1DAT

  type LayerInfo
     character(len=80), dimension(0:255) :: type
     character(len= 8), dimension(0:255) :: abbr
     character(len=16), dimension(0:255) :: LevUnit1, LevUnit2
     real             , dimension(0:255) :: LevUtimes1, LevUtimes2
     integer          , dimension(0:255) :: LevNum
  end type LayerInfo
  type(LayerInfo) :: GL
  character(len=128), parameter :: LIFile = 'TABLE/LayerInfo'

!  integer, parameter :: BufferSize = 1024
!  integer, parameter :: BufferSize = 62691
  integer, parameter :: BufferSize = 125382
  integer, parameter :: SEEK_SET   = 0
  integer, parameter :: SEEK_CUR   = 1

  integer, parameter :: GRIB1_NOERROR            =    0
  integer, parameter :: GRIB1_FILE_NOT_OPEN      = 5001
  integer, parameter :: GRIB1_NO_GRIB1_HEADER    = 5002
  integer, parameter :: GRIB1_SEEK_ERROR         = 5003
  integer, parameter :: GRIB1_SEC0_READ_ERROR    = 5004
  integer, parameter :: GRIB1_SEC1_READ_ERROR    = 5005
  integer, parameter :: GRIB1_SEC2_READ_ERROR    = 5006
  integer, parameter :: GRIB1_SEC3_READ_ERROR    = 5007
  integer, parameter :: GRIB1_SEC4_READ_ERROR    = 5008
  integer, parameter :: GRIB1_SEC5_READ_ERROR    = 5009
  integer, parameter :: GRIB1_LAYERINFO_NOT_OPEN = 5010
  integer, parameter :: GRIB1_LAST_LINES         = 5011
  integer, parameter :: GRIB1_CLOSE_ERROR        = 5012
  integer, parameter :: GRIB1_CANNOT_READ_DATA   = 5013

  type(BINFO) :: BI

  logical :: debug = .false.
  logical :: Initialized = .false.
  logical :: IEEE = .true.
  integer :: GRIB1_BufferSize = 62691

  public :: IEEE

  public :: GRIB1_BufferSize
  public :: GRIB1DAT
  public :: GRIB1List
  public :: GRIB1Sec0
  public :: GRIB1Sec1
  public :: GRIB1Sec2
  public :: GRIB1Sec3
  public :: GRIB1Sec4
  public :: GRIB1Init
  public :: GRIB1_Open
  public :: GRIB1_Scan
  public :: GRIB1_ReadData
  public :: GRIB1_Close

contains
  subroutine GRIB1Init( IO, status )
    integer, intent(IN)  :: IO
    integer, intent(OUT) :: status

    character(len=80) :: param
    character(len=16) :: unit, unit2
    character(len= 8) :: abbr
    integer :: id, num
    real    :: times1, times2
    logical :: FileExist

    if( GRIB1_BufferSize < 1 .or. BufferSize < GRIB1_BufferSize ) then
       GRIB1_BufferSize = BufferSize
    end if

    inquire( file=trim(LIFile), Exist=FileExist )
    if( .not. FileExist ) then
       status = GRIB1_LAYERINFO_NOT_OPEN
       return
    end if
    open( IO, file=trim(LIFile), IOSTAT=status )
    if( status /= 0 ) return
    do
       read(IO,*,IOSTAT=status) &
            & id, num, unit, unit2, times1, times2, param, abbr
       if( status /= 0 ) exit
       GL%type(id)       = param
       GL%abbr(id)       = abbr
       GL%LevNum(id)     = num
       GL%LevUnit1(id)   = unit
       GL%LevUnit2(id)   = unit2
       GL%LevUtimes1(id) = times1
       GL%LevUtimes2(id) = times2
    end do
    close( IO )

    Initialized = .true.
    status      = GRIB1_NOERROR

    return
  end subroutine GRIB1Init

  subroutine GRIB1ListInit( GT )
    type(GRIB1List), intent(INOUT) :: GT

    GT%no         = 0
    GT%HeaderTop  = -1
    GT%HeaderLine = -1
    GT%MaxLine    = -1
    GT%S0rec      = -1
    GT%S0pos      = -1
    GT%S0length   = -1
    GT%S1rec      = -1
    GT%S1pos      = -1
    GT%S1length   = -1
    GT%S2rec      = -1
    GT%S2pos      = -1
    GT%S2length   = -1
    GT%S3rec      = -1
    GT%S3pos      = -1
    GT%S3length   = -1
    GT%S4rec      = -1
    GT%S4pos      = -1
    GT%S4length   = -1
    GT%S5rec      = -1
    GT%S5pos      = -1
    GT%S5length   = -1

    return
  end subroutine GRIB1ListInit

  integer function GRIB1_Open( IO, grib1file, GRIB )
    integer         , intent(IN)    :: IO
    character(len=*), intent(IN)    :: grib1file
    type(GRIB1DAT)  , intent(INOUT) :: GRIB

    integer :: status
    logical :: FileExist

    integer :: i

    GRIB1_Open = GRIB1_FILE_NOT_OPEN

    if( .not. Initialized ) then
       call GRIB1Init( IO, status )
       if( status /= GRIB1_NOERROR ) then
          GRIB1_Open = status
          return
       end if
    end if

    GRIB1_Open = BINOpen( BI, IO, GRIB1_BufferSize, grib1file )
    if( GRIB1_Open /= BIN_NOERROR ) return

    call GRIB1ListInit( GRIB%GT )
    GRIB1_Open = GRIB1_HeadSearch( SEEK_SET )

    if( BI%cpos < 4 ) then
       GRIB%HeaderTop  = GRIB1_BufferSize - ( BI%cpos - 4 ) + 1
       GRIB%HeaderLine = BI%irec - 1
    else
       GRIB%HeaderTop  = BI%cpos - 4
       GRIB%HeaderLine = BI%irec
    end if

    GRIB%MaxLine = 0
    GRIB1_Open   = GRIB1_NOERROR
    return
  end function GRIB1_Open

  integer function GRIB1_Close( GRIB )
    type(GRIB1DAT) :: GRIB

    type(GRIB1List), pointer :: next, next2

    GRIB1_CLOSE = GRIB1_CLOSE_ERROR

    GRIB1_CLOSE = BINClose( BI )
    if( GRIB1_CLOSE /= BIN_NOERROR ) return

!    if( associated( GRIB%value ) ) nullify( GRIB%value )
    if( associated( GRIB%value ) ) deallocate( GRIB%value )

    if( associated( GRIB%GT%next ) ) then
       next => GRIB%GT%next
       nullify( GRIB%GT%next )
       do
          if( associated( next%next ) ) then
             next2 => next%next
             nullify( next )
             next => next2
          else
             exit
          end if
       end do
       if( associated( next%next ) ) nullify( next%next )
       nullify( next )
       if( associated( next2 ) ) then
          if( associated( next2%next ) ) nullify( next2%next )
          nullify( next2 )
       end if
    end if

    GRIB1_CLOSE = GRIB1_NOERROR
    return
  end function GRIB1_Close

  integer function GRIB1_HeadSearch( seek )
    integer, intent(IN)    :: seek

    integer :: i, status
    character(len=4) :: term4

    GRIB1_HeadSearch = GRIB1_NO_GRIB1_HEADER

    if( seek == SEEK_SET ) then
       BI%cpos = 1
       BI%rec1 = -1
       BI%rec2 = -1
    end if

    do
       GRIB1_HeadSearch = BINConvert( BI, term4 )
       if( BI%irec > BI%MaxLines ) then
          GRIB1_HeadSearch = GRIB1_LAST_LINES
          return
       end if
       if( GRIB1_HeadSearch /= BIN_NOERROR ) return

       if( term4 == 'GRIB' ) then
          GRIB1_HeadSearch = GRIB1_NOERROR
          return
       end if

       BI%cpos = BI%cpos - 3
    end do

    return
  end function GRIB1_HeadSearch

  integer function GRIB1_Scan( GRIB )
    type(GRIB1DAT), intent(INOUT) :: GRIB

    integer :: status, no, fsize
    type(GRIB1List), pointer :: cur, next

    no = 0
    GRIB1_Scan = GRIB1_SEEK_ERROR

    do
       ! Section 0
       ! Header scan
       if( no > 0 ) then
          GRIB1_Scan = GRIB1_HeadSearch( SEEK_CUR )
       else
          GRIB1_Scan = GRIB1_HeadSearch( SEEK_SET )
       end if
       if( GRIB1_Scan == GRIB1_NOERROR ) then
          no = no + 1
          allocate( cur )
          if( BI%cpos < 4 ) then
             GRIB%HeaderTop  = GRIB1_BufferSize - ( BI%cpos - 4 ) + 1
             GRIB%HeaderLine = BI%irec - 1
          else
             GRIB%HeaderTop  = BI%cpos - 4
             GRIB%HeaderLine = BI%irec
          end if

          cur%no         = no
          cur%HeaderTop  = GRIB%HeaderTop
          cur%HeaderLine = GRIB%HeaderLine
       else
          if( GRIB1_Scan == GRIB1_LAST_LINES ) then
             fsize = BI%BufferSize * ( BI%irec - 1 ) + BI%cpos - 5
             write(0,*)
             write(0,*) '***WARNING***'
             write(0,*) 'BUFFER SIZE SHOULD BE A FACTOR OF FILESIZE!'
             write(0,*) '  CURRENT BUFFER SIZE: ', BI%BufferSize
             write(0,*) '            FILE SIZE: ', fsize
             write(0,*)
             exit
          else
             return
          end if
       end if
       cur%S0pos  = BI%cpos
       cur%S0rec  = BI%irec
       GRIB1_Scan = GRIB1_ReadSection0( GRIB )
       if( GRIB1_Scan /= GRIB1_NOERROR ) return
       cur%TotalLength = GRIB%Sec0%TotalLength
       cur%MaxLine     = GRIB%MaxLine
       cur%S0length    = 8

       ! Section 1
       cur%S1pos  = BI%cpos
       cur%S1rec  = BI%irec
       GRIB1_Scan = GRIB1_ReadSection1( GRIB )
       if( GRIB1_Scan /= GRIB1_NOERROR ) return
       cur%S1length = GRIB%Sec1%Length

       ! Section 2
       cur%S2pos    = -1
       cur%S2rec    = -1
       cur%S2length = -1
       if( btest( GRIB%Sec1%flag, 7 ) ) then
          cur%S2pos = BI%cpos
          cur%S2rec = BI%irec
          GRIB1_Scan = GRIB1_ReadSection2( GRIB )
          if( GRIB1_Scan /= GRIB1_NOERROR ) return
          cur%S2length = GRIB%Sec2%Length
       end if
       
       ! Section 3
       cur%S3pos    = -1
       cur%S3rec    = -1
       cur%S3length = -1
       if( btest( GRIB%Sec1%flag, 6 ) ) then
          cur%S3pos = BI%cpos
          cur%S3rec = BI%irec
          GRIB1_Scan = GRIB1_ReadSection3( GRIB )
          if( GRIB1_Scan /= GRIB1_NOERROR ) return
          cur%S3length = GRIB%Sec3%Length
       end if
       
       ! Section 4
       cur%S4pos = BI%cpos
       cur%S4rec = BI%irec
       GRIB1_Scan = GRIB1_ReadSection4( GRIB, .true. )
       if( GRIB1_Scan /= GRIB1_NOERROR ) return
       cur%S4length = GRIB%Sec4%Length

       ! Section 5
       cur%S5pos = BI%cpos
       cur%S5rec = BI%irec
       GRIB1_Scan = GRIB1_ReadSection5( GRIB )
       if( GRIB1_Scan /= GRIB1_NOERROR ) return
       cur%S5length = 4

       ! date
       if( GRIB%Sec1%P1 == 0 ) then
          write(cur%date, '(i4.4,a1,i2.2,a1,i2.2,a1,i2.2,a1,i2.2,a8)' ) &
               & GRIB%Sec1%Year  , '-',                                 &
               & GRIB%Sec1%Month , '-',                                 &
               & GRIB%Sec1%Day   , '_',                                 &
               & GRIB%Sec1%Hour  , ':',                                 &
               & GRIB%Sec1%Minute, ':00.0000'
       else
          cur%date = 'YYYY-MM-DD_hh:mm_ss.ssss'
       end if

       cur%IPU   = GRIB%Sec1%IPU
       cur%LevID = GRIB%Sec1%LevID
       cur%Lev1  = GRIB%Sec1%Lev1
       cur%Lev2  = GRIB%Sec1%Lev2

       ! summary
       if( debug ) then
          write(6,*) cur%no
          write(6,*) cur%HeaderTop, cur%HeaderLine, cur%MaxLine, cur%TotalLength
          write(6,*) cur%S0pos, cur%S0rec, cur%S0length
          write(6,*) cur%S1pos, cur%S1rec, cur%S1length
          write(6,*) cur%S2pos, cur%S2rec, cur%S2length
          write(6,*) cur%S3pos, cur%S3rec, cur%S3length
          write(6,*) cur%S4pos, cur%S4rec, cur%S4length
          write(6,*) cur%S5pos, cur%S5rec, cur%S5length
          write(6,*)
       end if

       if( associated( GRIB%GT%next ) ) then
          next => GRIB%GT%next
          do

             if( .not. associated( next%next ) ) exit
             next => next%next
          end do
          next%next => cur

       else
          if( no == 1 ) then
             GRIB%GT%no          = cur%no
             GRIB%GT%HeaderTop   = cur%HeaderTop
             GRIB%GT%HeaderLine  = cur%HeaderLine
             GRIB%GT%MaxLine     = cur%MaxLine
             GRIB%GT%TotalLength = cur%TotalLength
             GRIB%GT%S0pos       = cur%S0pos
             GRIB%GT%S0rec       = cur%S0rec
             GRIB%GT%S0length    = cur%S0length
             GRIB%GT%S1pos       = cur%S1pos
             GRIB%GT%S1rec       = cur%S1rec
             GRIB%GT%S1length    = cur%S1length
             GRIB%GT%S2pos       = cur%S2pos
             GRIB%GT%S2rec       = cur%S2rec
             GRIB%GT%S2length    = cur%S2length
             GRIB%GT%S3pos       = cur%S3pos
             GRIB%GT%S3rec       = cur%S3rec
             GRIB%GT%S3length    = cur%S3length
             GRIB%GT%S4pos       = cur%S4pos
             GRIB%GT%S4rec       = cur%S4rec
             GRIB%GT%S4length    = cur%S4length
             GRIB%GT%S5pos       = cur%S5pos
             GRIB%GT%S5rec       = cur%S5rec
             GRIB%GT%S5length    = cur%S5length
             GRIB%GT%date        = cur%date
             GRIB%GT%IPU         = cur%IPU
             GRIB%GT%LevID       = cur%LevID
             GRIB%GT%Lev1        = cur%Lev1
             GRIB%GT%Lev2        = cur%Lev2
          else
             GRIB%GT%next => cur
          end if
       end if

       if( BI%irec-1 == BI%MaxLines .and. BI%cpos == 1 ) then
          exit
       end if

    end do

    GRIB1_Scan = GRIB1_NOERROR
    return
  end function GRIB1_Scan

  integer function GRIB1_ReadSection0( GRIB )
    type(GRIB1DAT), intent(INOUT) :: GRIB
!
! Octet No.
! --------------
!  1-4    GRIB
!  5-7    Total length of GRIB message in octets
!    8    GRIB Edition Number ( currently 2 )
!

    integer :: i, status
    character(len=1) :: term1
    character(len=3) :: term3

    if( debug ) then
       write(6,*)
       write(6,*) 'SECTION 0'
    end if

    GRIB1_ReadSection0 = GRIB1_SEC0_READ_ERROR
    
    ! Octet 5-7
    GRIB1_ReadSection0 = BINConvert( BI,                    &
         &                           term3,                 &
         &                           GRIB%Sec0%TotalLength, &
         &                           GRIB%ByteSwap          )
    if( debug ) write(6,*) 5, GRIB%Sec0%TotalLength

    ! Octet 8
    GRIB1_ReadSection0 = BINConvert( BI,                    &
         &                           term1,                 &
         &                           GRIB%Sec0%Edition,     &
         &                           GRIB%ByteSwap          )
    if( debug ) write(6,*) 8, GRIB%Sec0%Edition

    GRIB%MaxLine = GRIB%Sec0%TotalLength / BI%BufferSize + BI%irec - 1
    if( debug ) write(6,*) 'MAX REC: ', GRIB%MaxLine

    GRIB1_ReadSection0 = GRIB1_NOERROR
    return
  end function GRIB1_ReadSection0

  integer function GRIB1_ReadSection1( GRIB )
    type(GRIB1DAT), intent(INOUT) :: GRIB
!
! Octet No.
! ---------------------
!   1-3    Length of section in octets
!     4    Parameter Table Version number

!
    integer :: status, is, ie, itmp
    character(len=1) :: term1
    character(len=2) :: term2
    character(len=3) :: term3

    if( debug ) then
       write(6,*)
       write(6,*) 'SECTION 1'
    end if

    GRIB1_ReadSection1 = GRIB1_SEC1_READ_ERROR

    ! Octet 1-3
    GRIB1_ReadSection1 = BINConvert( BI,                    &
         &                           term3,                 &
         &                           GRIB%Sec1%Length,      &
         &                           GRIB%ByteSwap          )
    if( debug ) write(6,*) 1, GRIB%Sec1%Length

    ! Octet 4
    GRIB1_ReadSection1 = BINConvert( BI,                     &
         &                           term1,                  &
         &                           GRIB%Sec1%TableVersion, &
         &                           GRIB%ByteSwap           )
    if( debug ) write(6,*) 4, GRIB%Sec1%TableVersion

    ! Octet 5
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%IdentificationCenter, &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 5, GRIB%Sec1%IdentificationCenter

    ! Octet 6
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%GPID,                 &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 6, GRIB%Sec1%GPID

    ! Octet 7
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%GID,                  &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 7, GRIB%Sec1%GID

    ! Octet 8
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%flag,                 &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 8, GRIB%Sec1%flag

    ! Octet 9
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%IPU,                  &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 9, GRIB%Sec1%IPU

    ! Octet 10
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%LevID,                &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 10, GRIB%Sec1%LevID

    ! Octet 11
    GRIB%Sec1%Lev1 = -999
    GRIB%Sec1%Lev2 = -999
    if( GL%LevNum(GRIB%Sec1%LevID) == 1 ) then
       GRIB1_ReadSection1 = BINConvert( BI,                            &
            &                           term2,                         &
            &                           GRIB%Sec1%Lev1,                &
            &                           GRIB%ByteSwap                  )

    else if( GL%LevNum(GRIB%Sec1%LevID) == 2 ) then
       GRIB1_ReadSection1 = BINConvert( BI,                            &
            &                           term1,                         &
            &                           GRIB%Sec1%Lev1,                &
            &                           GRIB%ByteSwap                  )

       GRIB1_ReadSection1 = BINConvert( BI,                            &
            &                           term1,                         &
            &                           GRIB%Sec1%Lev2,                &
            &                           GRIB%ByteSwap                  )

    else
       BI%cpos = BI%cpos + 2
    end if
    if( debug ) write(6,*) 11, GRIB%Sec1%Lev1, GRIB%Sec1%Lev2

    ! Octet 13
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%Year,                 &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 13, GRIB%Sec1%Year

    ! Octet 14
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%Month,                &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 14, GRIB%Sec1%Month

    ! Octet 15
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%Day,                  &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 15, GRIB%Sec1%Day

    ! Octet 16
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%Hour,                 &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 16, GRIB%Sec1%Hour

    ! Octet 17
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%Minute,               &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 17, GRIB%Sec1%Minute

    ! Octet 18
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%TermUnit,             &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 18, GRIB%Sec1%TermUnit

    ! Octet 19
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%P1,                   &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 19, GRIB%Sec1%P1

    ! Octet 20
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%P2,                   &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 20, GRIB%Sec1%P2

    ! Octet 21
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%TermID,               &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 21, GRIB%Sec1%TermID

    ! Octet 22-23
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term2,                          &
         &                           GRIB%Sec1%TermNum,              &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 22, GRIB%Sec1%TermNum

    ! Octet 24
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%TermLack,             &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 24, GRIB%Sec1%TermLack

    ! Octet 25
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           itmp,                           &
         &                           GRIB%ByteSwap                   )
    GRIB%Sec1%Year = ( itmp - 1 ) * 100 + GRIB%Sec1%Year
    if( debug ) write(6,*) 25, GRIB%Sec1%Year

    ! Octet 26
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec1%IdSubCenter,          &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 26, GRIB%Sec1%IdSubCenter

    ! Octet 27-28
    GRIB1_ReadSection1 = BINConvert( BI,                             &
         &                           term2,                          &
         &                           GRIB%Sec1%D,                    &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 27, GRIB%Sec1%D

    GRIB1_ReadSection1 = GRIB1_NOERROR
    return
  end function GRIB1_ReadSection1

  integer function GRIB1_ReadSection2( GRIB )
    type(GRIB1DAT), intent(INOUT) :: GRIB
!
! Octet No.
! ---------------------
!   1-3    Length of section in octets
!
    integer :: status, is, ie, itmp
    character(len=1) :: term1
    character(len=2) :: term2
    character(len=3) :: term3

    if( debug ) then
       write(6,*)
       write(6,*) 'SECTION 2'
    end if

    GRIB1_ReadSection2 = GRIB1_SEC2_READ_ERROR

    ! Octet 1-3
    GRIB1_ReadSection2 = BINConvert( BI,                             &
         &                           term3,                          &
         &                           GRIB%Sec2%Length,               &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 1, GRIB%Sec2%Length

    ! Octet 4
    GRIB1_ReadSection2 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec2%NV,                   &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 4, GRIB%Sec2%NV

    ! Octet 5
    GRIB1_ReadSection2 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec2%PV,                   &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 5, GRIB%Sec2%PV

    ! Octet 6
    GRIB1_ReadSection2 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec2%Form,                 &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 6, GRIB%Sec2%Form

    ! Lat/Lon grid
    if( GRIB%Sec2%Form == 0 ) then
       return
    ! Lambert comformal, secant or tangent, conical or bipolar projection
    else if( GRIB%Sec2%Form == 3 ) then
       ! Octet 7-8
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term2,                          &
            &                           GRIB%Sec2%Nx,                   &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 7, GRIB%Sec2%Nx

       ! Octet 9-10
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term2,                          &
            &                           GRIB%Sec2%Ny,                   &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 9, GRIB%Sec2%Ny

       ! Octet 11-13
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%La1,                  &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 11, GRIB%Sec2%La1

       ! Octet 14-16
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%Lo1,                  &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 14, GRIB%Sec2%Lo1

       ! Octet 17
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term1,                          &
            &                           GRIB%Sec2%ResAndComp,           &
            &                           GRIB%ByteSwap                   )
       GRIB%Sec2%IDi = iand( GRIB%Sec2%ResAndComp, 128 ) / 128
       itmp          = iand( GRIB%Sec2%ResAndComp,  64 ) /  64
       GRIB%Sec2%IDi = iand( GRIB%Sec2%ResAndComp,   8 ) /   8
       if( itmp == 0 ) then
          GRIB%er = 6.36744e+6
          GRIB%of = 0.0
       else
          GRIB%er = 6.356775e+6
          GRIB%of = 1. / 297.0
       end if
       if( debug ) write(6,*) 17, GRIB%Sec2%ResAndComp, GRIB%er, GRIB%of

       ! Octet 18-20
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%Lov,                  &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 18, GRIB%Sec2%Lov

       ! Octet 21-23
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%Dx,                   &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 21, GRIB%Sec2%Dx

       ! Octet 24-26
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%Dy,                   &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 24, GRIB%Sec2%Dy

       ! Octet 27
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term1,                          &
            &                           GRIB%Sec2%ProjCenter,           &
            &                           GRIB%ByteSwap                   )
       if( iand( GRIB%Sec2%ProjCenter, 128 ) == 128 ) then
          GRIB%pole = -90
       else
          GRIB%pole =  90
       end if
       if( debug ) write(6,*) 27, GRIB%Sec2%ProjCenter, GRIB%pole

       ! Octet 28
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term1,                          &
            &                           GRIB%Sec2%scan,                 &
            &                           GRIB%ByteSwap                   )
       GRIB%Sec2%is  = iand( GRIB%Sec2%scan, 128 ) / 128
       GRIB%Sec2%js  = iand( GRIB%Sec2%scan,  64 ) /  64
       GRIB%Sec2%seq = iand( GRIB%Sec2%scan,  32 ) /  32
       if( debug ) write(6,*) 28, GRIB%Sec2%scan, GRIB%Sec2%is, GRIB%Sec2%js, GRIB%Sec2%seq

       ! Octet 29-31
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%Latin1,               &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 29, GRIB%Sec2%Latin1

       ! Octet 32-34
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%Latin2,               &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 32, GRIB%Sec2%Latin2

       ! Octet 35-37
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%SPlat,                &
            &                           GRIB%ByteSwap                   )
       GRIB%Sec2%SPlat = CheckSign( GRIB%Sec2%SPlat, 23 )
       if( debug ) write(6,*) 35, GRIB%Sec2%SPlat

       ! Octet 38-40
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term3,                          &
            &                           GRIB%Sec2%SPlon,                &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 38, GRIB%Sec2%SPlon

       ! Octet 41-42
       GRIB1_ReadSection2 = BINConvert( BI,                             &
            &                           term2,                          &
            &                           itmp,                           &
            &                           GRIB%ByteSwap                   )
       if( debug ) write(6,*) 41, itmp
    else
       return
    end if

    GRIB1_ReadSection2 = GRIB1_NOERROR
    return
  end function GRIB1_ReadSection2

  integer function GRIB1_ReadSection3( GRIB )
    type(GRIB1DAT), intent(INOUT) :: GRIB
!
! Octet No.
! ---------------------
!   1-3    Length of section in octets
! ( currently unsupported )
! 4-
!
    integer :: status, i, tmp
    character(len=3) :: term3

    if( debug ) then
       write(6,*)
       write(6,*) 'SECTION 3'
    end if

    GRIB1_ReadSection3 = GRIB1_SEC3_READ_ERROR

    ! Octet 1-3
    GRIB1_ReadSection3 = BINConvert( BI,                             &
         &                           term3,                          &
         &                           GRIB%Sec3%Length,               &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 1, GRIB%Sec3%Length

    GRIB1_ReadSection3 = GRIB1_NOERROR
    return
  end function GRIB1_ReadSection3

  integer function GRIB1_ReadSection4( GRIB, skip )
    type(GRIB1DAT), intent(INOUT) :: GRIB
    logical       , intent(IN)    :: skip
!
! Octet No.
! ---------------------
!   1-3    Length of section in octets
!     4    Bits 1-4: Flag
!          Bits 5-8: Number of unused bits at end of Section 4
!

    integer :: status, i, itmp
    character(len=1) :: term1
    character(len=2) :: term2
    character(len=3) :: term3
    character(len=4) :: term4

    if( debug ) then
       write(6,*)
       write(6,*) 'SECTION 4'
    end if

    GRIB1_ReadSection4 = GRIB1_SEC4_READ_ERROR

    ! Octet 1-3
    GRIB1_ReadSection4 = BINConvert( BI,                             &
         &                           term3,                          &
         &                           GRIB%Sec4%Length,               &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 1, GRIB%Sec4%Length

    ! Octet 4
    GRIB1_ReadSection4 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           itmp,                           &
         &                           GRIB%ByteSwap                   )
    if( btest( itmp, 7 ) ) then
       GRIB%Sec4%SHC = .true.
    else
       GRIB%Sec4%SHC = .false.
    end if
    if( btest( itmp, 6 ) ) then
       GRIB%Sec4%SmplPack = .true.
    else
       GRIB%Sec4%SmplPack = .false.
    end if
    if( btest( itmp, 5 ) ) then
       GRIB%Sec4%OrgReal = .true.
    else
       GRIB%Sec4%OrgReal = .false.
    end if
    if( btest( itmp, 4 ) ) then
       GRIB%Sec4%AddFlag = .true.
    else
       GRIB%Sec4%AddFlag = .false.
    end if
    GRIB%Sec4%SkipBit = &
         & iand(itmp,8) + iand(itmp,4) + iand(itmp,2) + iand(itmp,1)
    if( debug ) write(6,*) 4, GRIB%Sec4%SHC, GRIB%Sec4%SmplPack, GRIB%Sec4%OrgReal, GRIB%Sec4%AddFlag, GRIB%Sec4%SkipBit

    ! Octet 5-6
    GRIB1_ReadSection4 = BINConvert( BI,                             &
         &                           term2,                          &
         &                           GRIB%Sec4%E,                    &
         &                           GRIB%ByteSwap                   )
    GRIB%Sec4%E = CheckSign( GRIB%Sec4%E, 15 )
    if( debug ) write(6,*) 5, GRIB%Sec4%E

    ! Octet 7-10
    GRIB1_ReadSection4 = BINConvert( BI,                             &
         &                           term4,                          &
         &                           GRIB%Sec4%RefReal,              &
         &                           IEEE,                           &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 7, GRIB%Sec4%RefReal

   ! Octet 11
    GRIB1_ReadSection4 = BINConvert( BI,                             &
         &                           term1,                          &
         &                           GRIB%Sec4%BitNum,               &
         &                           GRIB%ByteSwap                   )
    if( debug ) write(6,*) 11, GRIB%Sec4%BitNum

    if( GRIB%Sec2%NV == 0 .and. GRIB%Sec2%PV == 255 ) then
       if( skip ) then
          BI%cpos = BI%cpos + GRIB%Sec4%Length - 11
          BI%irec = BI%irec + BI%cpos / BI%BufferSize
          BI%cpos = mod( BI%cpos, BI%BufferSize )
       else
          call ReadRegularSinglePack( GRIB )
          if( mod( GRIB%Sec4%SkipBit, 8 ) == 0 ) then
             BI%cpos = BI%cpos + ( GRIB%Sec4%SkipBit / 8 )
          else
             BI%cpos = BI%cpos + ( GRIB%Sec4%SkipBit / 8 ) + 1
          end if
       end if
    else
       stop 9999
    end if

    GRIB1_ReadSection4 = GRIB1_NOERROR
    return
  end function GRIB1_ReadSection4

  subroutine ReadRegularSinglePack( GRIB )
    type(GRIB1DAT), intent(INOUT) :: GRIB 

    integer :: i, j, nij, ix, jx, status, ival, itmp
    integer :: bpos, ip, jp
    integer :: BitInDat, BitMod
    integer :: TotalBit, TotalByte, ReadByte, ByteInRead
    integer :: BitHead, SkipByte

    character(len=10) :: term

    ix  = GRIB%Sec2%Nx
    jx  = GRIB%Sec2%Ny
    nij = ix * jx

!    if( associated( GRIB%value ) ) nullify( GRIB%value )
    if( associated( GRIB%value ) ) deallocate( GRIB%value )
    allocate( GRIB%value(ix,jx) )

    ! default values
    TotalBit = ix * jx * GRIB%Sec4%BitNum
    if( mod( TotalBit, 8 ) == 0 ) then
       TotalByte = TotalBit / 8
    else
       TotalByte = TotalBit / 8 + 1
    end if
    BitInDat   = Grib%Sec4%BitNum - 1

    if( mod( GRIB%Sec4%BitNum, 8 ) == 0 ) then
       ReadByte = GRIB%Sec4%BitNUm / 8
       SkipByte = ReadByte
    else
       ReadByte = GRIB%Sec4%BitNUm / 8 + 1
       SkipByte = ReadByte - 1
    end if
    BitMod = GRIB%Sec4%BitNum - ReadByte * 4

    BitHead = 0
    do i = 1, nij
       status = BINConvert( BI,                             &
            &               term(1:ReadByte),               &
            &               itmp,                           &
            &               GRIB%ByteSwap                   )

       ival = 0
       bpos = ( ReadByte * 8 - 1 ) - BitHead
       do j = BitInDat, 0, -1
          if( btest( itmp, bpos) ) ival = ival + 2**j
          bpos = bpos - 1
       end do

       if( GRIB%Sec2%seq == 0 ) then
          ip = mod( i, GRIB%Sec2%Nx )
          if( ip == 0 ) ip = GRIB%Sec2%Nx
          jp = ( i - ip ) / GRIB%Sec2%Nx + 1
       else
          jp = mod( i, GRIB%Sec2%Ny )
          if( jp == 0 ) jp = GRIB%Sec2%Ny
          ip = ( i - jp ) / GRIB%Sec2%Ny + 1
       end if

       GRIB%value(ip,jp) = ( GRIB%Sec4%RefReal                      &
            &              + real( ival ) * 2.**real(GRIB%Sec4%E) ) &
            &            / ( 10.**real(GRIB%Sec1%D) )

       BitHead = mod( GRIB%Sec4%BitNum * i, 8 )
       if( BitHead == 0 .and. ReadByte /= SkipByte ) then
          BI%cpos = BI%cpos + SkipByte + 1 - ReadByte
       else
          BI%cpos = BI%cpos + SkipByte - ReadByte
       end if

    end do

    return
  end subroutine ReadRegularSinglePack

  integer function GRIB1_ReadSection5( GRIB )
    type(GRIB1DAT), intent(INOUT) :: GRIB
!
! Octet No.
! ---------------------
!   1-4    '7777'
!

    character(len=4) :: term, term4
    integer :: status, is, ie, in

    if( debug ) then
       write(6,*)
       write(6,*) 'SECTION 5'
    end if

    GRIB1_ReadSection5 = GRIB1_SEC5_READ_ERROR

    GRIB1_ReadSection5 = BINConvert( BI, term )
    if( debug ) write(6,*) 1, term

    GRIB1_ReadSection5 = GRIB1_NOERROR
    return
  end function GRIB1_ReadSection5

  integer function GRIB1_ReadData( GRIB, rec, IPU, LevID, Lev1, Lev2, date )
    type(GRIB1DAT)             , intent(INOUT) :: GRIB
    integer          , optional, intent(IN) :: rec
    integer          , optional, intent(IN) :: IPU
    integer          , optional, intent(IN) :: LevID
    integer          , optional, intent(IN) :: Lev1
    integer          , optional, intent(IN) :: Lev2
    character(len=24), optional, intent(IN) :: date

    type(GRIB1List), pointer :: next
    integer :: rLev1, rLev2

    rLev1 = -999
    rLev2 = -999

    GRIB1_ReadData = GRIB1_CANNOT_READ_DATA

    if( present( rec ) ) then
       if( rec == 1 ) then
          BI%cpos = GRIB%GT%S4pos
          BI%irec = GRIB%GT%S4rec
          BI%rec1 = -1
       else
          if( .not. associated( GRIB%GT%next ) ) return
          next => GRIB%GT%next
          do
             if( next%no == rec ) exit
             if( .not. associated( next%next ) ) return
             next => next%next
          end do
          BI%cpos = next%S4pos
          BI%irec = next%S4rec
          BI%rec1 = -1
       end if

    else
       if( .not. present( IPU   ) ) return
       if( .not. present( LevID ) ) return
       if( present( Lev1  ) ) rLev1 = Lev1
       if( present( Lev2  ) ) rLev2 = Lev2
       if( .not. present( date  ) ) return

       if(    GRIB%GT%IPU   == IPU   .and. &
            & GRIB%GT%LevID == LevID .and. &
            & GRIB%GT%Lev1  == rLev1 .and. &
            & GRIB%GT%Lev2  == rLev2 .and. &
            & GRIB%GT%date  == date ) then
          BI%cpos = GRIB%GT%S4pos
          BI%irec = GRIB%GT%S4rec
          BI%rec1 = -1

       else
          if( .not. associated( GRIB%GT%next ) ) return
          next => GRIB%GT%next
          do
             if(    next%IPU   == IPU   .and. &
                  & next%LevID == LevID .and. &
                  & next%Lev1  == rLev1 .and. &
                  & next%Lev2  == rLev2 .and. &
                  & next%date  == date        ) exit

             if( .not. associated( next%next ) ) return
             next => next%next
          end do
          BI%cpos = next%S4pos
          BI%irec = next%S4rec
          BI%rec1 = -1
       end if

    end if

    GRIB1_ReadData = GRIB1_ReadSection4( GRIB, .false. )

    GRIB1_ReadData = GRIB1_NOERROR
    return
  end function GRIB1_ReadData

  integer function CheckSign( val, bit )
    integer, intent(IN)    :: bit
    integer, intent(INOUT) :: val

    integer :: tmp
    tmp = 0

    call mvbits( val, 0, bit+1, tmp, 0 )
    if( btest( tmp, bit ) ) then
       tmp = ibclr( tmp, bit )
       tmp = -tmp
    end if
    CheckSign = tmp
    return
  end function CheckSign
end module GRIB1
