module GRIB2
  use BIN
  implicit none
  private

  type Section0
     integer   :: cpos
     integer   :: irec
     integer   :: MasterTableNumber
     integer   :: Edition
     integer*8 :: TotalLength
  end type Section0

  type Section1
     integer :: cpos
     integer :: irec
     integer :: Length
     integer :: IdentificationCenter
     integer :: IdentificationSubCenter
     integer :: MasterTablesVersion
     integer :: LocalTablesVersion
     integer :: ReferenceTime
     integer :: Year
     integer :: Month
     integer :: Day
     integer :: Hour
     integer :: Minute
     integer :: Second
     integer :: ProductionStatus
     integer :: TypeOfProcessedData
  end type Section1

  type Section2
     integer :: cpos
     integer :: irec
     integer :: Length
  end type Section2

  type Section3
     integer :: cpos
     integer :: irec
     integer :: Length
     integer :: SourceOfGrid
     integer :: NumberOfData
     integer :: OptionalList
     integer :: InterpretationOfList
     integer :: GridDefinitionNumber
     real    :: er
     real    :: ellip_a
     real    :: ellip_b
     real    :: ellip_f
     integer :: Nx
     integer :: Ny
     integer :: BasicAngle
     integer :: UnitOfBaseAngle
     real    :: La1
     real    :: Lo1
     real    :: La2
     real    :: Lo2
     real    :: Dx
     real    :: Dy
     integer :: ResCompFlag
     integer :: ScanMode
  end type Section3

  type Section4
     integer :: cpos
     integer :: irec
     integer :: Length
     integer :: NumberOfCoordinateValues
     integer :: TemplateNumber
     integer :: ParameterCategory
     integer :: ParameterNumber
     integer :: GenerationProcess
     integer :: BGProcessID
     integer :: AFProcessID
     integer :: CutOffHour
     integer :: CutOffMinute
     integer :: TimeUnit
     integer :: ForcastTime
     integer :: FirstSurfaceType
     integer :: FirstSurfaceScale
     integer :: FirstSurfaceValue
     integer :: SecondSurfaceType
     integer :: SecondSurfaceScale
     integer :: SecondSurfaceValue
     integer :: LastYear
     integer :: LastMonth
     integer :: LastDay
     integer :: LastHour
     integer :: LastMinute
     integer :: LastSecond
     integer :: nDesc
     integer :: nMissing
     integer :: StatProcess
     integer :: TimeIncrementType
     integer :: TimeUnitID
     integer :: TimeRangeLength
     integer :: TimeIncrementID
     integer :: SuccessiveTimeIncrement
  end type Section4

  type Section5
     integer :: cpos
     integer :: irec
     integer :: Length
     integer :: DataPoints
     integer :: TemplateNumber
     real    :: RefValue
     integer :: BinScale
     integer :: DecScale
     integer :: Nbits
     integer :: FieldType
     integer :: CompMaxLevel
     integer :: MaxLevel
     real, dimension(:), pointer :: vtable
  end type Section5

  type Section6
     integer :: cpos
     integer :: irec
     integer :: Length
     integer :: BitMapIndicator
  end type Section6

  type Section7
     integer :: cpos
     integer :: irec
     integer :: Length
  end type Section7

  type GRIB2DAT
     integer        :: HeaderTop
     integer        :: HeaderLine
     integer        :: MaxLine
     logical        :: ByteSwap
     type(Section0) :: Sec0
     type(Section1) :: Sec1
     type(Section2) :: Sec2
     type(Section3) :: Sec3
     type(Section4) :: Sec4
     type(Section5) :: Sec5
     type(Section6) :: Sec6
     type(Section7) :: Sec7
  end type GRIB2DAT

  type(Section0) :: Sec0
  type(Section1) :: Sec1
  type(Section2) :: Sec2
  type(Section3) :: Sec3
  type(Section4) :: Sec4
  type(Section5) :: Sec5
  type(Section6) :: Sec6
  type(Section7) :: Sec7


  integer, parameter :: SEEK_SET   = 0
  integer, parameter :: SEEK_CUR   = 1

  integer, parameter :: GRIB2_NOERROR            =    0
  integer, parameter :: GRIB2_FILE_NOT_OPEN      = 1001
  integer, parameter :: GRIB2_NO_GRIB2_HEADER    = 1002
  integer, parameter :: GRIB2_SEEK_ERROR         = 1003
  integer, parameter :: GRIB2_SEC0_READ_ERROR    = 1004
  integer, parameter :: GRIB2_SEC1_READ_ERROR    = 1005
  integer, parameter :: GRIB2_SEC2_READ_ERROR    = 1006
  integer, parameter :: GRIB2_SEC3_READ_ERROR    = 1007
  integer, parameter :: GRIB2_SEC4_READ_ERROR    = 1008
  integer, parameter :: GRIB2_SEC5_READ_ERROR    = 1009
  integer, parameter :: GRIB2_SEC6_READ_ERROR    = 1010
  integer, parameter :: GRIB2_SEC7_READ_ERROR    = 1011
  integer, parameter :: GRIB2_SEC8_READ_ERROR    = 1012
  integer, parameter :: GRIB2_LAST_LINES         = 1013
  integer, parameter :: GRIB2_NOT_YET_SUPPORTED  = 1014
  integer, parameter :: GRIB2_VARIABLE_NOT_FOUND = -1015
  integer, parameter :: GRIB2_VARIABLE_NOT_READ  = 1016
  integer, parameter :: GRIB2_FILE_NOT_CLOSED    = 1017

  integer                   :: irec
  integer                   :: cpos

  type(BINFO) :: BI

  integer :: GRIB2_BufferSize = 1024

  logical :: debug = .false.

  public :: GRIB2DAT
  public :: GRIB2_Open
  public :: GRIB2_Close
  public :: GRIB2_Scan
  public :: GRIB2_InquireVariable
  public :: GRIB2_ReadVariable

  public :: GRIB2_BufferSize

  public :: Section0
  public :: Section1
  public :: Section2
  public :: Section3
  public :: Section4
  public :: Section5
  public :: Section6
  public :: Section7

  integer :: HeaderTop  = 0
  integer :: HeaderLine = 0

  character(len=1) :: cval1
  character(len=2) :: cval2
  character(len=4) :: cval4
  character(len=8) :: cval8

contains
  !
  ! Read variable data
  !
  integer function GRIB2_ReadVariable( IO,                          &
       &                               ByteSwap,                    &
       &                               GRIB,                        &
       &                               nval,                        &
       &                               n,                           &
       &                               val,                         &
       &                               nx,                          &
       &                               ny,                          &
       &                               category,                    &
       &                               number,                      &
       &                               time,                        &
       &                               level1,                      &
       &                               level2,                      &
       &                               resample                     )
    integer                 , intent(IN)    :: IO
    logical                 , intent(IN)    :: ByteSwap
    integer                 , intent(IN)    :: nval
    type(GRIB2DAT)          , intent(INOUT) :: GRIB(nval)
    integer                 , intent(IN)    :: n
    integer                 , intent(IN)    :: nx
    integer                 , intent(IN)    :: ny
    real                    , intent(OUT)   :: val(nx,ny)
    integer       , optional, intent(IN)    :: category
    integer       , optional, intent(IN)    :: number
    integer       , optional, intent(IN)    :: time
    integer       , optional, intent(IN)    :: level1
    integer       , optional, intent(IN)    :: level2
    logical       , optional, intent(IN)    :: resample

    integer :: nn, nnx, nny
    logical :: rspl

    integer :: i, j
    integer :: x1, x2, y1, y2
    real    :: x, y, x21, y21, xa, xb, ya, yb

    real, dimension(:,:), allocatable :: rtmp

    GRIB2_ReadVariable = GRIB2_VARIABLE_NOT_READ

    nn   = n
    rspl = .false.
    if( present( resample ) ) rspl = resample

    if(    present( category ) .and. &
         & present( number   ) .and. &
         & present( time     ) .and. &
         & present( level1   )       ) then
       if( present( level2 ) ) then
          nn = GRIB2_InquireVariable( GRIB, nval, category,        &
               &                      number, time, level1, level2 )
       else
          nn = GRIB2_InquireVariable( GRIB, nval, category,        &
               &                      number, time, level1         )
       end if

       if( nn == GRIB2_VARIABLE_NOT_FOUND ) return
    end if

    nnx = GRIB(nn)%Sec3%Nx
    nny = GRIB(nn)%Sec3%Ny
    if( nnx /= nx .or. nny /= ny ) then
       if( rspl ) then
          allocate( rtmp(nnx,nny) )

          BI%cpos = GRIB(nn)%Sec7%cpos
          BI%irec = GRIB(nn)%Sec7%irec
          GRIB2_ReadVariable = GRIB2_ReadSection7( IO,                    &
               &                                   ByteSwap,              &
               &                                   .false.,               &
               &                                   GRIB(nn),              &
               &                                   rtmp                   )

          do j = 1, ny
             y   = 1.0 + real( j - 1 ) * real( nny - 1 ) / real( ny - 1 )
             y1  = int( y )
             y2  = y1 + 1
             y2  = min( y2, nny )
             y21 = real( y2 - y1 )
             ya  = real(y2) - y
             yb  = y - real(y1)
             do i = 1, nx
                x   = 1.0 + real( i - 1 ) * real( nnx - 1 ) / real( nx - 1 )
                x1  = int( x )
                x2  = x1 + 1
                x2  = min( x2, nnx )
                x21 = real( x2 - x1 )
                xa  = real(x2) - x
                xb  = x - real(x1)

                if( abs(x21) > 1.e-4 .and. abs(y21) > 1.e-4 ) then
                   val(i,j) = ( xa * ( ya*rtmp(x1,y1) + yb*rtmp(x1,y2) )   &
                        &     + xb * ( ya*rtmp(x2,y1) + yb*rtmp(x2,y2) ) ) &
                        &   / ( y21 * x21 )
                else if( abs(x21) < 1.e-4 .and. abs(y21) > 1.e-4 ) then
                   val(i,j) = ( ya * rtmp(x1,y1) + yb * rtmp(x1,y2) ) / y21
                else if( abs(y21) < 1.e-4 .and. abs(x21) > 1.e-4 ) then
                   val(i,j) = ( xa * rtmp(x1,y1) + xb * rtmp(x2,y1) ) / x21
                else
                   val(i,j) = rtmp(x1,y1)
                end if
             end do
          end do

          deallocate( rtmp )
       else
          return
       end if
    else
       BI%cpos = GRIB(nn)%Sec7%cpos
       BI%irec = GRIB(nn)%Sec7%irec
       GRIB2_ReadVariable = GRIB2_ReadSection7( IO,                    &
            &                                   ByteSwap,              &
            &                                   .false.,               &
            &                                   GRIB(nn),              &
            &                                   val                    )
    end if

    return
  end function GRIB2_ReadVariable

  !
  ! Inquire variable data
  !
  integer function GRIB2_InquireVariable( GRIB,                     &
       &                                  nval,                     &
       &                                  category,                 &
       &                                  number,                   &
       &                                  time,                     &
       &                                  level1,                   &
       &                                  level2                    )
    integer                 , intent(IN) :: nval
    type(GRIB2DAT)          , intent(IN) :: GRIB(nval)
    integer                 , intent(IN) :: category
    integer                 , intent(IN) :: number
    integer                 , intent(IN) :: time
    integer                 , intent(IN) :: level1
    integer       , optional, intent(IN) :: level2

    integer :: i, il, il2

    GRIB2_InquireVariable = GRIB2_VARIABLE_NOT_FOUND

    do i = 1, nval
       if( GRIB(i)%Sec4%ForcastTime       /= time     ) cycle
       if( GRIB(i)%Sec4%ParameterCategory /= category ) cycle
       if( GRIB(i)%Sec4%ParameterNumber   /= number   ) cycle

       if( GRIB(i)%Sec4%FirstSurfaceScale > 0 ) then
          il = GRIB(i)%Sec4%FirstSurfaceValue                              &
               & / 10**(GRIB(i)%Sec4%FirstSurfaceScale)
       else if( GRIB(i)%Sec4%FirstSurfaceScale == -127 ) then
          il = level1
       else
          il = GRIB(i)%Sec4%FirstSurfaceValue                              &
               & * 10**(abs(GRIB(i)%Sec4%FirstSurfaceScale))
       end if
       if( il /= level1 ) cycle

       if( present( level2 ) ) then
          if( GRIB(i)%Sec4%SecondSurfaceScale > 0 ) then
             il2 = GRIB(i)%Sec4%SecondSurfaceValue                            &
                  & / 10**(GRIB(i)%Sec4%SecondSurfaceScale)
          else
             il2 = GRIB(i)%Sec4%SecondSurfaceValue                            &
                  & * 10**(abs(GRIB(i)%Sec4%SecondSurfaceScale))
          end if
          if( il2 /= level2 ) cycle
       end if

       GRIB2_InquireVariable = i
       return
    end do

    return
  end function GRIB2_InquireVariable

  integer function GRIB2_Close( IO )
    integer, intent(IN) :: IO
    GRIB2_Close = GRIB2_FILE_NOT_CLOSED

    GRIB2_Close = BINClose( BI )
    return
  end function GRIB2_Close

  integer function GRIB2_Open( IO, grib2file, ByteSwap, nval )
    integer                   , intent(IN)    :: IO
    character(len=*)          , intent(IN)    :: grib2file
    logical                   , intent(IN)    :: ByteSwap
    integer                   , intent(INOUT) :: nval

    integer :: status
    logical :: FileExist, skip

    integer :: esec, nsec, nlen

    nval  = 0
    cval4 = '7777'
    esec  = transfer( cval4, 7777 )

    GRIB2_Open = GRIB2_FILE_NOT_OPEN

    GRIB2_Open = BINOpen( BI, IO, GRIB2_BufferSize, grib2file )
    if( GRIB2_Open /= BIN_NOERROR ) return

    GRIB2_Open = GRIB2_HeadSearch( SEEK_SET )

    if( BI%cpos < 4 ) then
       HeaderTop  = GRIB2_BufferSize - ( BI%cpos - 4 ) + 1
       HeaderLine = BI%irec - 1
    else
       HeaderTop  = BI%cpos - 4
       HeaderLine = BI%irec
    end if

    cpos = BI%cpos
    irec = BI%irec

    BI%cpos = BI%cpos + 12

    do
       GRIB2_Open = BINConvert( BI, cval4, nlen, ByteSwap )
       if( GRIB2_Open /= BIN_NOERROR ) return
       if( nlen == esec ) exit
       GRIB2_Open = BINConvert( BI, cval1, nsec, ByteSwap )
       if( GRIB2_Open /= BIN_NOERROR ) return
       if( nsec > 7 .or. nsec < 1 ) return

       if( nsec == 4 ) nval = nval + 1
       BI%cpos = BI%cpos + nlen - 5
    end do
    BI%cpos = cpos
    BI%irec = irec

    GRIB2_Open   = GRIB2_NOERROR
    return
  end function GRIB2_Open

  integer function GRIB2_HeadSearch( seek )
    integer, intent(IN)    :: seek

    integer          :: i, status
    character(len=4) :: term4

    GRIB2_HeadSearch = GRIB2_NO_GRIB2_HEADER

    if( seek == SEEK_SET ) then
       BI%cpos = 1
       BI%rec1 = -1
       BI%rec2 = -1
    end if

    do
       GRIB2_HeadSearch = BINConvert( BI, term4 )
       if( BI%irec > BI%MaxLines ) then
          GRIB2_HeadSearch = GRIB2_LAST_LINES
          return
       end if
       if( GRIB2_HeadSearch /= BIN_NOERROR ) return

       if( term4 == 'GRIB' ) then
          GRIB2_HeadSearch = GRIB2_NOERROR
          return
       end if

       BI%cpos = BI%cpos - 3
    end do

    return
  end function GRIB2_HeadSearch

  integer function GRIB2_Scan( IO, GRIB, nval, ByteSwap )
    integer       , intent(IN)    :: IO
    integer       , intent(IN)    :: nval
    type(GRIB2DAT), intent(INOUT) :: GRIB(nval)
    logical       , intent(IN)    :: ByteSwap

    integer :: i
    integer :: in, nlen, esec

    in    = 1
    cval4 = '7777'
    esec  = transfer( cval4, 7777 )


    GRIB2_Scan      = GRIB2_SEEK_ERROR
    GRIB%HeaderTop  = HeaderTop
    GRIB%HeaderLine = irec

    ! Section 0
    Sec0%cpos  = HeaderTop
    Sec0%irec  = irec
    GRIB2_Scan = GRIB2_ReadSection0( IO, ByteSwap )
    if( GRIB2_Scan /= GRIB2_NOERROR ) return
    GRIB%Sec0 = Sec0

    ! Section 1
    GRIB2_Scan = GRIB2_ReadSection1( IO, ByteSwap )
    if( GRIB2_Scan /= GRIB2_NOERROR ) return
    GRIB%Sec1 = Sec1

    do
       ! Section 2
       GRIB2_Scan = GRIB2_ReadSection2( IO, ByteSwap )
       if( GRIB2_Scan /= GRIB2_NOERROR ) return
       if( in > 1 .and. Sec2%Length == 0 ) then
          GRIB(in)%Sec2 = GRIB(in-1)%Sec2
       else
          GRIB(in)%Sec2 = Sec2
       end if

       ! Section 3
       GRIB2_Scan = GRIB2_ReadSection3( IO, ByteSwap )
       if( GRIB2_Scan /= GRIB2_NOERROR ) return
       if( in > 1 .and. Sec3%Length == 0 ) then
          GRIB(in)%Sec3 = GRIB(in-1)%Sec3
       else
          GRIB(in)%Sec3 = Sec3
       end if

       ! Section 4
       GRIB2_Scan = GRIB2_ReadSection4( IO, ByteSwap )
       if( GRIB2_Scan /= GRIB2_NOERROR ) return
       GRIB(in)%Sec4 = Sec4

       ! Section 5
       GRIB2_Scan = GRIB2_ReadSection5( IO, ByteSwap )
       if( GRIB2_Scan /= GRIB2_NOERROR ) return
       GRIB(in)%Sec5 = Sec5

       ! Section 6
       GRIB2_Scan = GRIB2_ReadSection6( IO, ByteSwap )
       if( GRIB2_Scan /= GRIB2_NOERROR ) return
       GRIB(in)%Sec6 = Sec6

       ! Section 7
       GRIB2_Scan = GRIB2_ReadSection7( IO, ByteSwap, .true. )
       if( GRIB2_Scan /= GRIB2_NOERROR ) return
       GRIB(in)%Sec7 = Sec7

       GRIB2_Scan = BINConvert( BI, cval4, nlen, ByteSwap )
       if( GRIB2_Scan /= BIN_NOERROR ) return
       if( nlen /= esec ) then
          BI%cpos = BI%cpos - 4
          in      = in + 1
       else
          exit
       end if
    end do

    if( debug ) then
       write(6,'(a10,2i10)') 'SECTION 0:', GRIB(1)%Sec0%cpos, GRIB(1)%Sec0%irec
       write(6,'(a10,2i10)') 'SECTION 1:', GRIB(1)%Sec1%cpos, GRIB(1)%Sec1%irec
       do i = 1, nval
          if( GRIB(i)%Sec2%cpos > 0 .and. GRIB(i)%Sec2%irec > 0 ) then
             write(6,'(a10,2i10)') 'SECTION 2:', GRIB(i)%Sec2%cpos, &
                  &                          GRIB(i)%Sec2%irec
          end if
          if( GRIB(i)%Sec3%cpos > 0 .and. GRIB(i)%Sec3%irec > 0 ) then
             write(6,'(a10,2i10)') 'SECTION 3:', GRIB(i)%Sec3%cpos, &
                  &                          GRIB(i)%Sec3%irec
          end if
          write(6,'(a10,7i10)') 'SECTION 4:', GRIB(i)%Sec4%cpos,                  &
               &                          GRIB(i)%Sec4%irec,                  &
               &                          GRIB(i)%Sec4%ParameterCategory,     &
               &                          GRIB(i)%Sec4%ParameterNumber,       &
               &                          GRIB(i)%Sec4%FirstSurfaceValue,     &
               &                          GRIB(i)%Sec4%ForcastTime,           &
               &                          GRIB(i)%Sec4%GenerationProcess
          write(6,'(a10,2i10,e15.7,2i10)') 'SECTION 5:', GRIB(i)%Sec5%cpos,               &
               &                          GRIB(i)%Sec5%irec,                  &
               &                          GRIB(i)%Sec5%RefValue,              &
               &                          GRIB(i)%Sec5%BinScale,              &
               &                          GRIB(i)%Sec5%DecScale
          write(6,'(a10,2i10)') 'SECTION 6:', GRIB(i)%Sec6%cpos, GRIB(i)%Sec6%irec
          write(6,'(a10,2i10)') 'SECTION 7:', GRIB(i)%Sec7%cpos, GRIB(i)%Sec7%irec
       end do
    end if

    GRIB2_Scan = GRIB2_NOERROR
    return
  end function GRIB2_Scan

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SECTION 0
!
! Octet No.
! --------------
!  1-4    GRIB
!  5-6    Reserved
!    7    Discipline - GRIB Master Table Number
!    8    GRIB Edition Number ( currently 2 )
! 9-16    Total length of GRIB message in octets
!
  integer function GRIB2_ReadSection0( IO, ByteSwap )
    integer       , intent(IN)    :: IO
    logical       , intent(IN)    :: ByteSwap

    integer :: i, status

    GRIB2_ReadSection0 = GRIB2_SEC0_READ_ERROR

    call initialize_struct( 0 )

    ! Octet 5-6
    BI%cpos = BI%cpos + 2

    ! Octet 7
    GRIB2_ReadSection0 = BINConvert( BI, cval1, Sec0%MasterTableNumber, ByteSwap )
    if( GRIB2_ReadSection0 /= BIN_NOERROR ) return

    ! Octet 8
    GRIB2_ReadSection0 = BINConvert( BI, cval1, Sec0%Edition, ByteSwap )
    if( GRIB2_ReadSection0 /= BIN_NOERROR ) return

    ! Octet 9-16
    GRIB2_ReadSection0 = BINConvert( BI, cval8, Sec0%TotalLength, ByteSwap )
    if( GRIB2_ReadSection0 /= BIN_NOERROR ) return

    GRIB2_ReadSection0 = GRIB2_NOERROR

    return
  end function GRIB2_ReadSection0

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SECTION 1
!
! Octet No.
! ---------------------
!   1-4    Length of section in octets ( 21 or nn )
!     5    Number of Section ( 1 )
!   6-7    Identification of originating/generating centre
!   8-9    Identificatio of originating/generating sub-centre
!    10    GRIB Master Tables Version Number
!    11    GRIB Local Tables Version Number
!    12    Significance of Reference Time
! 13-14    Year ( 4 digit )
!    15    Month
!    16    Day
!    17    Hour
!    18    Minute
!    19    Second
!    20    Production status of processed data in this GRIB message
!    21    Type of prcessed data in this GRIB message
! 22-nn    Reserved: need not be present
!
  integer function GRIB2_ReadSection1( IO, ByteSwap )
    integer       , intent(IN)    :: IO
    logical       , intent(IN)    :: ByteSwap

    integer :: nsec, nlen, cpos, irec

    GRIB2_ReadSection1 = GRIB2_SEC1_READ_ERROR
    call initialize_struct( 1 )

    cpos = BI%cpos
    irec = BI%irec

    ! Octet 1-4
    GRIB2_ReadSection1 = BINConvert( BI, cval4, nlen, ByteSwap )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return


    ! Octet 5
    GRIB2_ReadSection1 = BINConvert( BI, cval1, nsec, ByteSwap )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    if( nsec == 1 ) then
       Sec1%cpos   = cpos
       Sec1%irec   = irec
       Sec1%Length = nlen
    else
       Sec1%cpos = -1
       Sec1%irec = -1
       BI%cpos = BI%cpos - 5
       return
    end if

    ! Octet 6-7
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval2,                          &
         &                           Sec1%IdentificationCenter,      &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return
    
    ! Octet 8-9
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval2,                          &
         &                           Sec1%IdentificationSubCenter,   &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 10
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%MasterTablesVersion,       &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 11
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%LocalTablesVersion,        &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 12
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%ReferenceTime,             &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 13-14
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval2,                          &
         &                           Sec1%Year,                      &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 15
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%Month,                     &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 16
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%Day,                       &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 17
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%Hour,                      &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 18
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%Minute,                    &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 19
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%Second,                    &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 20
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%ProductionStatus,          &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! Octet 21
    GRIB2_ReadSection1 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec1%TypeOfProcessedData,       &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection1 /= BIN_NOERROR ) return

    ! skip if Octet > 21
    BI%cpos = BI%cpos + ( 21 - Sec1%Length )
    
    GRIB2_ReadSection1 = GRIB2_NOERROR

    return
  end function GRIB2_ReadSection1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SECTION 2
!
! Octet No.
! ---------------------
!   1-4    Length of section in octets
!     5    Number of Section ( 2 )
!  6-nn    Local use
!
  integer function GRIB2_ReadSection2( IO, ByteSwap )
    integer       , intent(IN)    :: IO
    logical       , intent(IN)    :: ByteSwap

    integer :: nsec, nlen, cpos, irec

    GRIB2_ReadSection2 = GRIB2_SEC2_READ_ERROR
    call initialize_struct( 2 )

    cpos = BI%cpos
    irec = BI%irec

    ! Octet 1-4
    GRIB2_ReadSection2 = BINConvert( BI, cval4, nlen, ByteSwap )
    if( GRIB2_ReadSection2 /= BIN_NOERROR ) return

    ! Octet 5
    GRIB2_ReadSection2 = BINConvert( BI, cval1, nsec, ByteSwap )
    if( GRIB2_ReadSection2 /= BIN_NOERROR ) return

    if( nsec == 2 ) then
       Sec2%cpos   = cpos
       Sec2%irec   = irec
       Sec2%Length = nlen
    else
       Sec2%cpos = -1
       Sec2%irec = -1
       BI%cpos = BI%cpos - 5
       return
    end if

    ! skip if Octet >= 6
    BI%cpos = BI%cpos + ( 5 - Sec1%Length )
    
    GRIB2_ReadSection2 = GRIB2_NOERROR

    return
  end function GRIB2_ReadSection2

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SECTION 3
!
! Octet No.
! ---------------------
!   1-4    Length of section in octets
!     5    Number of Section ( 3 )
!     6    Source of grid definition
!  7-10    Number of data points
!    11    Num. of octets for optional list of num. defining num. of points
!    12    Interpretation of list of numbers defining number of points
! 13-14    Grid Definition Template Number (N)
!
! ( currently unsupported )
! 15-xx    Grid Definition Template
! [xx+1]-nn Optional list of numbers defining number of points
!
  integer function GRIB2_ReadSection3( IO, ByteSwap )
    integer       , intent(IN)    :: IO
    logical       , intent(IN)    :: ByteSwap

    integer :: status, i, tmp

    integer :: nsec, nlen, cpos, irec

    GRIB2_ReadSection3 = GRIB2_SEC3_READ_ERROR
    call initialize_struct( 3 )

    cpos = BI%cpos
    irec = BI%irec

    ! Octet 1-4
    GRIB2_ReadSection3 = BINConvert( BI, cval4, nlen, ByteSwap )
    if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

    ! Octet 5
    GRIB2_ReadSection3 = BINConvert( BI, cval1, nsec, ByteSwap )
    if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

    if( nsec == 3 ) then
       Sec3%cpos   = cpos
       Sec3%irec   = irec
       Sec3%Length = nlen
    else
       Sec3%cpos = -1
       Sec3%irec = -1
       BI%cpos = BI%cpos - 5
       return
    end if

    ! Octet 6
    GRIB2_ReadSection3 = BINConvert( BI,                           &
         &                           cval1,                        &
         &                           Sec3%SourceOfGrid,            &
         &                           ByteSwap                      )
    if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

    ! Octet 7-10
    GRIB2_ReadSection3 = BINConvert( BI,                           &
         &                           cval4,                        &
         &                           Sec3%NumberOfData,            &
         &                           ByteSwap                      )
    if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

    ! Octet 11
    GRIB2_ReadSection3 = BINConvert( BI,                           &
         &                           cval1,                        &
         &                           Sec3%OptionalList,            &
         &                           ByteSwap                      )
    if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

    ! Octet 12
    GRIB2_ReadSection3 = BINConvert( BI,                           &
         &                           cval1,                        &
         &                           Sec3%InterpretationOfList,    &
         &                           ByteSwap                      )
    if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

    ! Octet 13-14
    GRIB2_ReadSection3 = BINConvert( BI,                           &
         &                           cval2,                        &
         &                           Sec3%GridDefinitionNumber,    &
         &                           ByteSwap                      )
    if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

    if( Sec3%SourceOfGrid == 0 ) then
       if( Sec3%GridDefinitionNumber == 0 ) then
          ! Octet 15 ( Shape of the earth )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval1,                        &
               &                           tmp,                          &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return
          if( tmp == 0 ) then
             Sec3%er      = 6.36747e+6
             Sec3%ellip_a = Sec3%er
             Sec3%ellip_b = Sec3%er
             Sec3%ellip_f = 0
          else if( tmp == 4 ) then
             Sec3%er      = 6.3781370e+6
             Sec3%ellip_a = 6.3781370e+6
             Sec3%ellip_b = 6.356752314e+6
             Sec3%ellip_f = 1/298.257222101
          else if( tmp == 6 ) then
             Sec3%er      = 6.371229e+6
             Sec3%ellip_a = Sec3%er
             Sec3%ellip_b = Sec3%er
             Sec3%ellip_f = 0
          else
             GRIB2_ReadSection3 = GRIB2_NOT_YET_SUPPORTED
             return
          end if

          ! skip Octet 16 to 30
          BI%cpos = BI%cpos + 15

          ! Octet 31-34 ( Nx )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           Sec3%Nx,                      &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

          ! Octet 35-38 ( Ny )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           Sec3%Ny,                      &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

          ! Octet 39-42 ( basic angle of all lat and lon  )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           Sec3%BasicAngle,              &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

          ! Octet 43-46 ( the unit of basic angle of all lat and lon  )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           Sec3%UnitOfBaseAngle,         &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

          ! Octet 47-50 ( the latitude of first grid )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           tmp,                          &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return
          Sec3%La1 = real(tmp) * 1e-6

          ! Octet 51-54 ( the longitude of first grid )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           tmp,                          &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return
          Sec3%Lo1 = real(tmp) * 1e-6

          ! Octet 55 ( resolution and component flag )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval1,                        &
               &                           Sec3%ResCompFlag,             &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

          ! Octet 56-59 ( the latitude of the last grid )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           tmp,                          &
               &                           ByteSwap,                     &
               &                           31                            )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return
          Sec3%La2 = real(tmp) * 1e-6

          ! Octet 60-63 ( the longitude of the last grid )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           tmp,                          &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return
          Sec3%Lo2 = tmp * 1e-6

          ! Octet 64-67 ( i direction increments )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           tmp,                          &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return
          Sec3%Dx = tmp * 1e-6

          ! Octet 68-71 ( j direction increments )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval4,                        &
               &                           tmp,                          &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return
          Sec3%Dy = tmp * 1e-6

          ! Octet 72 ( scan mode )
          GRIB2_ReadSection3 = BINConvert( BI,                           &
               &                           cval1,                        &
               &                           Sec3%ScanMode,                &
               &                           ByteSwap                      )
          if( GRIB2_ReadSection3 /= BIN_NOERROR ) return

       else
          GRIB2_ReadSection3 = GRIB2_NOT_YET_SUPPORTED
          return
       end if
    else
       GRIB2_ReadSection3 = GRIB2_NOT_YET_SUPPORTED
       return
    end if

    GRIB2_ReadSection3 = GRIB2_NOERROR

    return
  end function GRIB2_ReadSection3

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SECTION 4
!
! Octet No.
! ---------------------
!   1-4    Length of section in octets
!     5    Number of Section ( 4 )
!   6-7    Number of coordinate values after Template
!   8-9    Product Definition Template Number
!
! ( currently not supported )
! 10-xx    Product Definition Template
! [xx+1]-xx  Optional list of coordinate values
!
  integer function GRIB2_ReadSection4( IO, ByteSwap )
    integer       , intent(IN)    :: IO
    logical       , intent(IN)    :: ByteSwap

    integer :: nsec, nlen

    GRIB2_ReadSection4 = GRIB2_SEC4_READ_ERROR
    call initialize_struct( 4 )

    cpos  = BI%cpos
    irec  = BI%irec
    cval4 = '    '
    cval1 = ' '

    ! Octet 1-4
    GRIB2_ReadSection4 = BINConvert( BI, cval4, nlen, ByteSwap )
    if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

    ! Octet 5
    GRIB2_ReadSection4 = BINConvert( BI, cval1, nsec, ByteSwap )
    if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

    if( nsec == 4 ) then
       Sec4%cpos   = cpos
       Sec4%irec   = irec
       Sec4%Length = nlen
    else
       Sec4%cpos = -1
       Sec4%irec = -1
       BI%cpos = BI%cpos - 5
       return
    end if

    ! Octet 6-7
    GRIB2_ReadSection4 = BINConvert( BI,                               &
         &                           cval2,                            &
         &                           Sec4%NumberOfCoordinateValues,    &
         &                           ByteSwap                          )
    if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

    ! Octet 8-9
    GRIB2_ReadSection4 = BINConvert( BI,                               &
         &                           cval2,                            &
         &                           Sec4%TemplateNumber,              &
         &                           ByteSwap                          )
    if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

    if(    Sec4%TemplateNumber == 0 .or. &
         & Sec4%TemplateNumber == 8 .or. &
         & Sec4%TemplateNumber == 50008 ) then

       ! Octet 10
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%ParameterCategory,           &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 11
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%ParameterNumber,             &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 12
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%GenerationProcess,           &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 13
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%BGProcessID,                 &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 14
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%AFProcessID,                 &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 15,16
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval2,                            &
            &                           Sec4%CutOffHour,                  &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 17
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%CutOffMinute,                &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 18
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%TimeUnit,                    &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return
       
       ! Octet 19-22
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval4,                            &
            &                           Sec4%ForcastTime,                 &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return
       Sec4%ForcastTime = CheckSign( Sec4%ForcastTime, 31 )
       
       ! Octet 23
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%FirstSurfaceType,            &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return
       
       ! Octet 24
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%FirstSurfaceScale,           &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return
       Sec4%FirstSurfaceScale = CheckSign( Sec4%FirstSurfaceScale, 7 )

       ! Octet 25-28
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval4,                            &
            &                           Sec4%FirstSurfaceValue,           &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return
       
       ! Octet 29
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%SecondSurfaceType,           &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 30
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval1,                            &
            &                           Sec4%SecondSurfaceScale,          &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       ! Octet 31-34
       GRIB2_ReadSection4 = BINConvert( BI,                               &
            &                           cval4,                            &
            &                           Sec4%SecondSurfaceValue,          &
            &                           ByteSwap                          )
       if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

       if(    Sec4%TemplateNumber == 8 .or. &
            & Sec4%TemplateNumber == 50008 ) then
          ! Octet 35-36
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval2,                            &
               &                           Sec4%LastYear,                    &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 37
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%LastMonth,                   &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 38
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%LastDay,                     &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 39
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%LastHour,                    &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 40
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%LastMinute,                  &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 41
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%LastSecond,                  &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 42
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%nDesc,                       &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 43-46
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval4,                            &
               &                           Sec4%nMissing,                    &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 47
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%StatProcess,                 &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 48
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%TimeIncrementType,           &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 49
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%TimeUnitID,                  &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 50-53
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval4,                            &
               &                           Sec4%TimeRangeLength,             &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 54
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval1,                            &
               &                           Sec4%TimeIncrementID,             &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          ! Octet 55-58
          GRIB2_ReadSection4 = BINConvert( BI,                               &
               &                           cval4,                            &
               &                           Sec4%SuccessiveTimeIncrement,     &
               &                           ByteSwap                          )
          if( GRIB2_ReadSection4 /= BIN_NOERROR ) return

          if( Sec4%TemplateNumber == 50008 ) then
             ! Octet 59-66 & 67-74 & 75-82 ( work information, currently skip )
             BI%cpos = BI%cpos +  24
          end if

          if( Sec4%nDesc > 1 ) then
             GRIB2_ReadSection4 = GRIB2_NOT_YET_SUPPORTED
             return
          end if
       end if

    else
       GRIB2_ReadSection4 = GRIB2_NOT_YET_SUPPORTED
       return
    end if

    GRIB2_ReadSection4 = GRIB2_NOERROR

    return
  end function GRIB2_ReadSection4

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SECTION 5
!
! Octet No.
! ---------------------
!   1-4    Length of section in octets
!     5    Number of Section ( 5 )
!   6-9    Number of data points where one or more values are specified in Sec7
! 10-11    Data Representation Template Number
!
! ( currently unsupported )
! 12-nn    Data Representation Template
!
  integer function GRIB2_ReadSection5( IO, ByteSwap )
    integer       , intent(IN)    :: IO
    logical       , intent(IN)    :: ByteSwap

    integer :: i
    integer :: tmp
    integer :: nsec, nlen

    GRIB2_ReadSection5 = GRIB2_SEC5_READ_ERROR
    call initialize_struct( 5 )

    cpos = BI%cpos
    irec = BI%irec

    ! Octet 1-4
    GRIB2_ReadSection5 = BINConvert( BI, cval4, nlen, ByteSwap )
    if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

    ! Octet 5
    GRIB2_ReadSection5 = BINConvert( BI, cval1, nsec, ByteSwap )
    if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

    if( nsec == 5 ) then
       Sec5%cpos   = cpos
       Sec5%irec   = irec
       Sec5%Length = nlen
    else
       Sec5%cpos = -1
       Sec5%irec = -1
       BI%cpos = BI%cpos - 5
       return
    end if

    ! Octet 6-9
    GRIB2_ReadSection5 = BINConvert( BI,                              &
         &                           cval4,                           &
         &                           Sec5%DataPoints,                 &
         &                           ByteSwap                         )
    if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

    ! Octet 10-11
    GRIB2_ReadSection5 = BINConvert( BI,                              &
         &                           cval2,                           &
         &                           Sec5%TemplateNumber,             &
         &                           ByteSwap                         )
    if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

    if( Sec5%TemplateNumber == 0 ) then
       ! Octet 12-15
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval4,                           &
            &                           Sec5%RefValue,                   &
            &                           ByteSwap,                        &
            &                           .false.                          )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

       ! Octet 16-17
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval2,                           &
            &                           Sec5%BinScale,                   &
            &                           ByteSwap                         )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return
       Sec5%BinScale = CheckSign( Sec5%BinScale, 15 )

       ! Octet 18-19
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval2,                           &
            &                           Sec5%DecScale,                   &
            &                           ByteSwap                         )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return
       Sec5%DecScale = CheckSign( Sec5%DecScale, 15 )

       ! Octet 20
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval1,                           &
            &                           Sec5%Nbits,                      &
            &                           ByteSwap                         )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

       ! Octet 21
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval1,                           &
            &                           Sec5%FieldType,                  &
            &                           ByteSwap                         )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

    else if( Sec5%TemplateNumber == 200 ) then
       ! Octet 12
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval1,                           &
            &                           Sec5%Nbits,                      &
            &                           ByteSwap                         )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

       ! Octet 13-14
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval2,                           &
            &                           Sec5%CompMaxLevel,               &
            &                           ByteSwap                         )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

       ! Octet 15-16
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval2,                           &
            &                           Sec5%MaxLevel,                   &
            &                           ByteSwap                         )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

       ! Octet 17
       GRIB2_ReadSection5 = BINConvert( BI,                              &
            &                           cval1,                           &
            &                           Sec5%DecScale,                   &
            &                           ByteSwap                         )
       if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

       if( associated( Sec5%vtable ) ) deallocate( Sec5%vtable )
       allocate( Sec5%vtable( Sec5%MaxLevel ) )

       do i = 1, Sec5%MaxLevel
          ! Octet 15-16
          GRIB2_ReadSection5 = BINConvert( BI,                              &
               &                           cval2,                           &
               &                           tmp,                             &
               &                           ByteSwap                         )
          if( GRIB2_ReadSection5 /= BIN_NOERROR ) return

          Sec5%vtable(i) = real(tmp) * 0.01
       end do

    else
       GRIB2_ReadSection5 = GRIB2_NOT_YET_SUPPORTED
       return
    end if

    GRIB2_ReadSection5 = GRIB2_NOERROR

    return
  end function GRIB2_ReadSection5

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SECTION 6
!
! Octet No.
! ---------------------
!   1-4    Length of section in octets
!     5    Number of Section ( 6 )
!     6    Bit-map indicator
!
! ( currentlry unsupported )
!  7-nn    Bit-map
!
  integer function GRIB2_ReadSection6( IO, ByteSwap )
    integer       , intent(IN)    :: IO
    logical       , intent(IN)    :: ByteSwap

    integer :: nsec, nlen

    GRIB2_ReadSection6 = GRIB2_SEC6_READ_ERROR
    call initialize_struct( 6 )

    cpos = BI%cpos
    irec = BI%irec

    ! Octet 1-4
    GRIB2_ReadSection6 = BINConvert( BI, cval4, nlen, ByteSwap )
    if( GRIB2_ReadSection6 /= BIN_NOERROR ) return

    ! Octet 5
    GRIB2_ReadSection6 = BINConvert( BI, cval1, nsec, ByteSwap )
    if( GRIB2_ReadSection6 /= BIN_NOERROR ) return

    if( nsec == 6 ) then
       Sec6%cpos   = cpos
       Sec6%irec   = irec
       Sec6%Length = nlen
    else
       Sec6%cpos = -1
       Sec6%irec = -1
       BI%cpos = BI%cpos - 5
       return
    end if

    ! Octet 6
    GRIB2_ReadSection6 = BINConvert( BI,                             &
         &                           cval1,                          &
         &                           Sec6%BitMapIndicator,           &
         &                           ByteSwap                        )
    if( GRIB2_ReadSection6 /= BIN_NOERROR ) return

    ! skip octet 7-
    BI%cpos = BI%cpos + ( 6 - Sec6%Length )

    GRIB2_ReadSection6 = GRIB2_NOERROR

    return
  end function GRIB2_ReadSection6

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SECTION 7
!
! Octet No.
! ---------------------
!   1-4    Length of section in octets
!     5    Number of Section ( 7 )
!   6-     Data
!
  integer function GRIB2_ReadSection7( IO,                          &
       &                               ByteSwap,                    &
       &                               SKIP,                        &
       &                               GRIB,                        &
       &                               val                          )
    integer                                 , intent(IN)    :: IO
    logical                                 , intent(IN)    :: ByteSwap
    logical                                 , intent(IN)    :: SKIP
    type(GRIB2DAT)                , optional, intent(INOUT) :: GRIB
    real          , dimension(:,:), optional, intent(INOUT) :: val
    

    integer :: nsec, nlen

    GRIB2_ReadSection7 = GRIB2_SEC7_READ_ERROR
    call initialize_struct( 7 )

    cpos = BI%cpos
    irec = BI%irec

    ! Octet 1-4
    GRIB2_ReadSection7 = BINConvert( BI, cval4, nlen, ByteSwap )
    if( GRIB2_ReadSection7 /= BIN_NOERROR ) return

    ! Octet 5
    GRIB2_ReadSection7 = BINConvert( BI, cval1, nsec, ByteSwap )
    if( GRIB2_ReadSection7 /= BIN_NOERROR ) return

    if( nsec == 7 ) then
       Sec7%cpos   = cpos
       Sec7%irec   = irec
       Sec7%Length = nlen
    else
       Sec7%cpos = -1
       Sec7%irec = -1
       BI%cpos = BI%cpos - 5
       return
    end if

    if( SKIP ) then
       BI%cpos = BI%cpos + ( Sec7%Length - 5 )
       return

    else
       if( present( GRIB ) .and. present( val ) ) then
          if( GRIB%Sec5%TemplateNumber == 200 ) then
             call ReadRunLengthPack( IO, GRIB, ByteSwap, val )
          else
             val = -1.
             call ReadRegularSinglePack( GRIB, ByteSwap, val )
          end if
       end if
    end if

    GRIB2_ReadSection7 = GRIB2_NOERROR

    return
  end function GRIB2_ReadSection7

  subroutine ReadRunLengthPack( IO, GRIB, ByteSwap, val )
    integer                       , intent(IN)    :: IO
    type(GRIB2DAT)                , intent(INOUT) :: GRIB
    logical                       , intent(IN)    :: ByteSwap
    real          , dimension(:,:), intent(OUT)   :: val

    integer :: nx
    integer :: ny
    integer :: M
    integer :: V
    integer :: L

    integer :: status
    integer :: itmp
    integer :: loop
    integer :: in
    integer :: i
    integer :: j
    integer :: ipos
    integer :: nmax
    integer :: ival
    integer :: n
    integer :: nc

    integer :: cpos
    integer :: irec

    nx = size( val, 1 )
    ny = size( val, 2 )
    nmax = nx * ny

    cpos = BI%cpos
    irec = BI%irec

    BI%cpos = GRIB%Sec5%cpos
    BI%irec = GRIB%Sec5%irec
    status = GRIB2_ReadSection5( IO, ByteSwap )
    GRIB%Sec5 = Sec5

    BI%cpos = cpos
    BI%irec = irec

    M = GRIB%Sec5%MaxLevel
    V = GRIB%Sec5%CompMaxLevel
    L = 2**GRIB%Sec5%Nbits - 1 - V

    if( GRIB%Sec3%Nx /= nx .or. GRIB%Sec3%Ny /= ny ) return

    ival = -1
    n    =  1
    nc   =  0
    ipos =  0
    do loop = 1, GRIB%Sec7%Length-5
       status = BINConvert( BI,                              &
            &               cval1,                           &
            &               itmp,                            &
            &               ByteSwap                         )
       if( status /= BIN_NOERROR ) return

       if( itmp <= V ) then
          if( ival >= 0 ) then
             do in = 1, n
                ipos = ipos + 1
                j    = ipos / nx + 1
                i    = ipos - ( nx * ( j - 1 ) )

                if( ipos > nmax ) then
                   write(6,*) '***GRIB2 READ ERROR (DECODE RUN LENGTH)***'
                   return
                end if

                if( ival == 0 ) then
                   val(i,j) = -1.
                else
                   val(i,j) = GRIB%Sec5%vtable(ival)
                end if

             end do
          end if

          ival = itmp
          n    =    1
          nc   =    0
       else
          nc = nc + 1
          n  = n + L**(nc-1) * ( itmp - ( V + 1 ) )
       end if

    end do

    return
  end subroutine ReadRunLengthPack

  subroutine ReadRegularSinglePack( GRIB, ByteSwap, val )
    type(GRIB2DAT)                , intent(IN)    :: GRIB 
    logical                       , intent(IN)    :: ByteSwap
    real          , dimension(:,:), intent(INOUT) :: val

    integer :: nx, ny, nxy

    integer :: i, j, nij, ix, jx, status, ival, itmp
    integer :: bpos, ip, jp
    integer :: BitInDat, BitMod
    integer :: TotalBit, TotalByte, ReadByte, ByteInRead
    integer :: BitHead, SkipByte

    character(len=10) :: term

    nx  = size( val, 1 )
    ny  = size( val, 2 )
    nxy = nx * ny

    if( GRIB%Sec3%Nx /= nx .or. GRIB%Sec3%Ny /= ny ) return

    TotalBit = nx * ny * GRIB%Sec5%Nbits
    if( mod( TotalBit, 8 ) == 0 ) then
       TotalByte = TotalBit / 8
    else
       TotalByte = TotalBit / 8 + 1
    end if
    BitInDat = GRIB%Sec5%Nbits - 1

    if( mod( GRIB%Sec5%Nbits, 8 ) == 0 ) then
       ReadByte = GRIB%Sec5%Nbits / 8
       SkipByte = ReadByte
    else
       ReadByte = GRIB%Sec5%Nbits / 8 + 1
       SkipByte = ReadByte - 1
    end if
    BitMod = GRIB%Sec5%Nbits - ReadByte * 4

    BitHead = 0
    do i = 1, nxy
       status = BINConvert( BI,                             &
            &               term(1:ReadByte),               &
            &               itmp,                           &
            &               ByteSwap                        )

       ival = 0
       bpos = ( ReadByte * 8 - 1 ) - BitHead
       do j = BitInDat, 0, -1
          if( btest( itmp, bpos) ) ival = ival + 2**j
          bpos = bpos - 1
       end do

       if( btest( GRIB%Sec3%ScanMode, 5 ) ) then
          ip = ( i - 1 ) / ny + 1
          jp = i - ( ip - 1 ) * ny
       else
          jp = ( i - 1 ) / nx + 1
          ip = i - ( jp - 1 ) * nx
       end if

       val(ip,jp) = ( GRIB%Sec5%RefValue                                   &
            &              + real( ival ) * 2.**real(GRIB%Sec5%BinScale) ) &
            &            / ( 10.**real(GRIB%Sec5%DecScale) )

       BitHead = mod( GRIB%Sec5%Nbits * i, 8 )
       if( BitHead == 0 .and. ReadByte /= SkipByte ) then
          BI%cpos = BI%cpos + SkipByte + 1 - ReadByte
       else
          BI%cpos = BI%cpos + SkipByte - ReadByte
       end if

    end do

    return
  end subroutine ReadRegularSinglePack

  subroutine initialize_struct( no )
    integer, intent(IN) :: no

    if( no == 0 ) then
       Sec0%cpos              = 0
       Sec0%irec              = 0
       Sec0%MasterTableNumber = 0
       Sec0%Edition           = 0
       Sec0%TotalLength       = 0

    else if( no == 1 ) then
       Sec1%cpos                    = 0
       Sec1%irec                    = 0
       Sec1%Length                  = 0
       Sec1%IdentificationCenter    = 0
       Sec1%IdentificationSubCenter = 0
       Sec1%MasterTablesVersion     = 0
       Sec1%LocalTablesVersion      = 0
       Sec1%ReferenceTime           = 0
       Sec1%Year                    = 0
       Sec1%Month                   = 0
       Sec1%Day                     = 0
       Sec1%Hour                    = 0
       Sec1%Minute                  = 0
       Sec1%Second                  = 0
       Sec1%ProductionStatus        = 0
       Sec1%TypeOfProcessedData     = 0

    else if( no == 2 ) then
       Sec2%cpos   = 0
       Sec2%irec   = 0
       Sec2%Length = 0

    else if( no == 3 ) then
       Sec3%cpos                 = 0
       Sec3%irec                 = 0
       Sec3%Length               = 0
       Sec3%SourceOfGrid         = 0
       Sec3%NumberOfData         = 0
       Sec3%OptionalList         = 0
       Sec3%InterpretationOfList = 0
       Sec3%GridDefinitionNumber = 0
       Sec3%er                   = 0.0
       Sec3%ellip_a              = 0.0
       Sec3%ellip_b              = 0.0
       Sec3%ellip_f              = 0.0
       Sec3%Nx                   = 0
       Sec3%Ny                   = 0
       Sec3%BasicAngle           = 0
       Sec3%UnitOfBaseAngle      = 0
       Sec3%La1                  = 0.0
       Sec3%Lo1                  = 0.0
       Sec3%La2                  = 0.0
       Sec3%Lo2                  = 0.0
       Sec3%Dx                   = 0.0
       Sec3%Dy                   = 0.0
       Sec3%ResCompFlag          = 0
       Sec3%ScanMode             = 0

    else if ( no == 4 ) then
       Sec4%cpos                     = 0
       Sec4%irec                     = 0
       Sec4%Length                   = 0
       Sec4%NumberOfCoordinateValues = 0
       Sec4%TemplateNumber           = 0
       Sec4%ParameterCategory        = 0
       Sec4%ParameterNumber          = 0
       Sec4%GenerationProcess        = 0
       Sec4%BGProcessID              = 0
       Sec4%AFProcessID              = 0
       Sec4%CutOffHour               = 0
       Sec4%CutOffMinute             = 0
       Sec4%TimeUnit                 = 0
       Sec4%ForcastTime              = 0
       Sec4%FirstSurfaceType         = 0
       Sec4%FirstSurfaceScale        = 0
       Sec4%FirstSurfaceValue        = 0
       Sec4%SecondSurfaceType        = 0
       Sec4%SecondSurfaceScale       = 0
       Sec4%SecondSurfaceValue       = 0
       Sec4%LastYear                 = 0
       Sec4%LastMonth                = 0
       Sec4%LastDay                  = 0
       Sec4%LastHour                 = 0
       Sec4%LastMinute               = 0
       Sec4%LastSecond               = 0
       Sec4%nDesc                    = 0
       Sec4%nMissing                 = 0
       Sec4%StatProcess              = 0
       Sec4%TimeIncrementType        = 0
       Sec4%TimeUnitID               = 0
       Sec4%TimeRangeLength          = 0
       Sec4%TimeIncrementID          = 0
       Sec4%SuccessiveTimeIncrement  = 0

    else if( no == 5 ) then
       Sec5%cpos           = 0
       Sec5%irec           = 0
       Sec5%Length         = 0
       Sec5%DataPoints     = 0
       Sec5%TemplateNumber = 0
       Sec5%RefValue       = 0.0
       Sec5%BinScale       = 0
       Sec5%DecScale       = 0
       Sec5%Nbits          = 0
       Sec5%FieldType      = 0
       Sec5%CompMaxLevel   = 0
       Sec5%MaxLevel       = 0
       if( associated( Sec5%vtable ) ) deallocate( Sec5%vtable )

    else if( no == 6 ) then
       Sec6%cpos            = 0
       Sec6%irec            = 0
       Sec6%Length          = 0
       Sec6%BitMapIndicator = 0

    else if( no == 7 ) then
       Sec7%cpos   = 0
       Sec7%irec   = 0
       Sec7%Length = 0

    end if

    return
  end subroutine initialize_struct

  integer function CheckSign( val, bit )
    integer, intent(IN)    :: bit
    integer, intent(INOUT) :: val

    integer :: tmp
    tmp = 0
    
    call mvbits( val, 0, bit+1, tmp, 0 )
    if( btest( tmp, bit ) ) then
       tmp = ibclr( tmp, bit )
       tmp = -tmp
    end if
    CheckSign = tmp
    return
  end function CheckSign

end module GRIB2
