module BIN
  use hlib
  implicit none
  private

  type BINFO
     integer :: IO
     integer :: BufferSize
     integer :: cpos
     integer :: irec
     integer :: rec1, rec2
     integer :: MaxLines
     logical :: swap
     character, dimension(:), allocatable :: Buffer1
     character, dimension(:), allocatable :: Buffer2
  end type BINFO

  integer, parameter :: BIN_NOERROR          =    0
  integer, parameter :: BIN_FILE_NOT_OPEN    = 4001
  integer, parameter :: BIN_LAST_LINE_UNREAD = 4002
  integer, parameter :: BIN_CANNOT_READ_CHAR = 4003
  integer, parameter :: BIN_BUFFER_UNSTORED  = 4004
  integer, parameter :: BIN_NOT_CONVERT_INT  = 4005
  integer, parameter :: BIN_CLOSE_ERROR      = 4005

  interface BINConvert
     module procedure BINReadChar
     module procedure BINCharToInteger4
     module procedure BINCharToInteger8
     module procedure BINCharToReal4
  end interface

  integer :: oflag = 0

  public :: BINFO
  public :: BIN_NOERROR
  public :: BINOpen
  public :: BINConvert
  public :: BINClose
  
contains
  ! open binary data
  integer function BINOpen( BI, IO, BufferSize, fname )
    type(BINFO)               , intent(INOUT) :: BI
    integer                   , intent(IN)    :: IO
    integer                   , intent(IN)    :: BufferSize
    character(len=*)          , intent(IN)    :: fname

    logical :: FileExist
    integer :: irec, status

    BINOpen = BIN_FILE_NOT_OPEN

    inquire( file=trim(fname), EXIST=FileExist )
    if( .not. FileExist ) return

#ifdef HLIB
    call hopen( BI%IO, trim(fname), O_RDONLY, status, oflag )
    if( status /= 0 ) return
    call hseek( BI%IO, SEEK_SET, 0, status, oflag )
#else
    open( IO,                &
         & file=trim(fname), &
         & access='direct',  &
         & recl=BufferSize,  &
         & IOSTAT=status     )
    BI%IO = IO
#endif
    if( status /= 0 ) return

    BI%BufferSize = BufferSize
    BI%cpos       = 1
    BI%irec       = 1
    BI%rec1       = -1
    BI%rec2       = -1
    BI%swap       = .false.
    BI%MaxLines   = 0
    allocate( BI%Buffer1(BufferSize), stat=status )
    if( status /= 0 ) return
    allocate( BI%Buffer2(BufferSize), stat=status )
    if( status /= 0 ) return

    irec = 0
    do
       irec = irec + 1
#ifdef HLIB
       call hread( BI%IO, BI%Buffer1, BufferSize, status, oflag )
       if( status == 0 ) exit
#else
       read( BI%IO, rec=irec, IOSTAT=status ) BI%Buffer1
       if( status /= 0 ) exit
#endif
       BI%MaxLines = BI%MaxLines + 1
    end do

    BI%Buffer1 = ''

!!$    if( irec /= BI%MaxLines ) then
!!$       BINOpen = BIN_LAST_LINE_UNREAD
!!$       return
!!$    end if

    BINOpen = BIN_NOERROR
  end function BINOpen

  integer function BINClose( BI )
    type(BINFO), intent(INOUT) :: BI

    integer :: status

    BINClose = BIN_CLOSE_ERROR

#ifdef HLIB
    call hclose( BI%IO, status, oflag )
#else
    close( BI%IO, IOSTAT=status )
#endif
    if( status /= 0 ) return

    BI%IO = -1
    BI%BufferSize = -1
    BI%cpos       = -1
    BI%irec       = -1
    BI%rec1       = -1
    BI%rec2       = -2
    BI%MaxLines   = -1
    BI%swap       = .false.
    if( allocated( BI%Buffer1 ) ) deallocate( BI%Buffer1 )
    if( allocated( BI%Buffer2 ) ) deallocate( BI%Buffer2 )

    BINClose = BIN_NOERROR
    return
  end function BINClose

  ! convert 1-4 byte characters into 4-byte integer
  integer function BINCharToReal4( BI, cval, rval, swap, IEEE )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=*)          , intent(INOUT) :: cval
    real                      , intent(INOUT) :: rval
    logical         , optional, intent(IN)    :: swap
    logical         , optional, intent(IN)    :: IEEE

    integer :: clen
    logical :: IEEEo

    if( present( IEEE ) ) then
       IEEEo = IEEE
    else
       IEEEo = .true.
    end if

    BINCharToReal4 = BIN_NOT_CONVERT_INT

    clen = len(cval)

    if( clen == 1 ) then

    else if( clen == 2 ) then

    else if( clen == 3 ) then

    else if( clen == 4 ) then
       if( present( swap ) ) then
          BINCharToReal4 = BINChar4ToReal4( BI, cval, rval, IEEEo, swap )
       else
          BINCharToReal4 = BINChar4ToReal4( BI, cval, rval, IEEEo )
       end if

    else
       return
    end if

    return
  end function BINCharToReal4

  ! convert 4-byte characters into 4-byte float
  integer function BINChar4ToReal4( BI, cval, rval, IEEE, swap )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=4)          , intent(INOUT) :: cval
    real                      , intent(OUT)   :: rval
    logical                   , intent(IN)    :: IEEE
    logical         , optional, intent(IN)    :: swap


    BINChar4ToReal4 = BIN_NOT_CONVERT_INT

    BINChar4ToReal4 = BINReadChar( BI, cval )
    if( BINChar4ToReal4 /= BIN_NOERROR ) return

    if( present( swap ) ) then
       call Char4ToReal4( cval, rval, IEEE, swap )
    else
       call Char4ToReal4( cval, rval, IEEE       )
    end if

    BINChar4ToReal4 = BIN_NOERROR
    return
  end function BINChar4ToReal4

  subroutine Char4ToReal4( cval, rval, IEEE, swap )
    character(len=4) , intent(IN)    :: cval
    real             , intent(INOUT) :: rval
    logical          , intent(IN)    :: IEEE
    logical, optional, intent(IN)    :: swap

    integer          :: i, j
    character(len=4) :: cdata

    cdata = cval
    if( present( swap ) ) then
       if( swap ) then
          do i = 1, 4
             j = 5 - i
             cdata(i:i) = cval(j:j)
          end do
       end if
    end if

    rval = transfer( cdata, 0.0_4 )
    if( IEEE ) rval = M2I3E( rval )

    return
  end subroutine Char4ToReal4

  real function M2I3E( ibm )
    real, intent(IN) :: ibm

    integer :: i, j
    integer :: sign, power, float, tmp
    sign  = 0
    power = 0
    float = 0
    tmp = transfer( ibm, tmp )
    call mvbits( tmp, 31,  1, sign , 0 )
    call mvbits( tmp, 24,  7, power, 0 )
    call mvbits( tmp,  0, 24, float, 0 )

    M2I3E = 16.**(power-64) * real( float ) / ( 16.**6 )
    if( btest(sign,0) ) M2I3E = - M2I3E

    return
  end function M2I3E

  ! convert 1-4 byte characters into 4-byte integer
  integer function BINCharToInteger4( BI, cval, ival, swap, ibit )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=*)          , intent(INOUT) :: cval
    integer                   , intent(INOUT) :: ival
    logical         , optional, intent(IN)    :: swap
    integer         , optional, intent(IN)    :: ibit

    integer :: clen

    BINCharToInteger4 = BIN_NOT_CONVERT_INT

    clen = len(cval)

    if( clen == 1 ) then
       if( present( swap ) ) then
          BINCharToInteger4 = BINChar1ToInteger4( BI, cval, ival, swap )
       else
          BINCharToInteger4 = BINChar1ToInteger4( BI, cval, ival       )
       end if

    else if( clen == 2 ) then
       if( present( swap ) ) then
          BINCharToInteger4 = BINChar2ToInteger4( BI, cval, ival, swap )
       else
          BINCharToInteger4 = BINChar2ToInteger4( BI, cval, ival       )
       end if

    else if( clen == 3 ) then
       if( present( swap ) ) then
          BINCharToInteger4 = BINChar3ToInteger4( BI, cval, ival, swap )
       else
          BINCharToInteger4 = BINChar3ToInteger4( BI, cval, ival       )
       end if

    else if( clen == 4 ) then
       if( present( swap ) ) then
          if( present( ibit ) ) then
             BINCharToInteger4 = BINChar4ToInteger4( BI, cval, ival, swap, ibit )
          else
             BINCharToInteger4 = BINChar4ToInteger4( BI, cval, ival, swap )
          end if
       else
          if( present( ibit ) ) then
             BINCharToInteger4 = BINChar4ToInteger4( BI, cval, ival, ibit=ibit )
          else
             BINCharToInteger4 = BINChar4ToInteger4( BI, cval, ival       )
          end if
       end if

    else
       return
    end if

    return
  end function BINCharToInteger4

  ! convert 1-byte characters into 4-byte integer
  integer function BINChar1ToInteger4( BI, cval, ival, swap )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=1)          , intent(INOUT) :: cval
    integer                   , intent(OUT)   :: ival
    logical         , optional, intent(IN)    :: swap

    character(len=4) :: cdata

    BINChar1ToInteger4 = BIN_NOT_CONVERT_INT

    BINChar1ToInteger4 = BINReadChar( BI, cval )
    if( BINChar1ToInteger4 /= BIN_NOERROR ) return

    cdata = transfer( 0, cdata )
    cdata(4:4) = cval(1:1)

    if( present( swap ) ) then
       call Char4ToInteger4( cdata, ival, swap )
    else
       call Char4ToInteger4( cdata, ival       )
    end if

    BINChar1ToInteger4 = BIN_NOERROR
    return
  end function BINChar1ToInteger4

  ! convert 2-byte characters into 4-byte integer
  integer function BINChar2ToInteger4( BI, cval, ival, swap )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=2)          , intent(INOUT) :: cval
    integer                   , intent(OUT)   :: ival
    logical         , optional, intent(IN)    :: swap

    character(len=4) :: cdata

    BINChar2ToInteger4 = BIN_NOT_CONVERT_INT

    BINChar2ToInteger4 = BINReadChar( BI, cval )
    if( BINChar2ToInteger4 /= BIN_NOERROR ) return

    cdata = transfer( 0, cdata )
    cdata(3:4) = cval(1:2)

    if( present( swap ) ) then
       call Char4ToInteger4( cdata, ival, swap )
    else
       call Char4ToInteger4( cdata, ival       )
    end if

    BINChar2ToInteger4 = BIN_NOERROR
    return
  end function BINChar2ToInteger4

  ! convert 3-byte characters into 4-byte integer
  integer function BINChar3ToInteger4( BI, cval, ival, swap )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=3)          , intent(INOUT) :: cval
    integer                   , intent(OUT)   :: ival
    logical         , optional, intent(IN)    :: swap

    character(len=4) :: cdata

    BINChar3ToInteger4 = BIN_NOT_CONVERT_INT

    BINChar3ToInteger4 = BINReadChar( BI, cval )
    if( BINChar3ToInteger4 /= BIN_NOERROR ) return

    cdata = transfer( 0, cdata )
    cdata(2:4) = cval(1:3)

    if( present( swap ) ) then
       call Char4ToInteger4( cdata, ival, swap )
    else
       call Char4ToInteger4( cdata, ival       )
    end if

    BINChar3ToInteger4 = BIN_NOERROR
    return
  end function BINChar3ToInteger4

  ! convert 4-byte characters into 4-byte integer
  integer function BINChar4ToInteger4( BI, cval, ival, swap, ibit )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=4)          , intent(INOUT) :: cval
    integer                   , intent(OUT)   :: ival
    logical         , optional, intent(IN)    :: swap
    integer         , optional, intent(IN)    :: ibit


    BINChar4ToInteger4 = BIN_NOT_CONVERT_INT

    BINChar4ToInteger4 = BINReadChar( BI, cval )
    if( BINChar4ToInteger4 /= BIN_NOERROR ) return

    if( present( swap ) ) then
       if( present( ibit ) ) then
          call Char4ToInteger4( cval, ival, swap, ibit )
       else
          call Char4ToInteger4( cval, ival, swap )
       end if
    else
       if( present( ibit ) ) then
          call Char4ToInteger4( cval, ival, ibit=ibit )
       else
          call Char4ToInteger4( cval, ival       )
       end if
    end if

    BINChar4ToInteger4 = BIN_NOERROR

    return
  end function BINChar4ToInteger4

  subroutine Char4ToInteger4( cval, ival, swap, ibit )
    character(len=4) , intent(IN) :: cval
    integer          , intent(INOUT) :: ival
    logical, optional, intent(IN)    :: swap
    integer, optional, intent(IN)    :: ibit

    integer          :: i, j
    character(len=4) :: cdata

    cdata = cval
    if( present( swap ) ) then
       if( swap ) then
          do i = 1, 4
             j = 5 - i
             cdata(i:i) = cval(j:j)
          end do
       end if
    end if

    ival = transfer( cdata, 0_4 )
    if( present( ibit ) ) then
       ival = CheckSign( ival, ibit )
    end if

    return
  end subroutine Char4ToInteger4

  integer function CheckSign( val, bit )
    integer, intent(IN)    :: bit
    integer, intent(INOUT) :: val

    integer :: tmp
    tmp = 0
    
    call mvbits( val, 0, bit+1, tmp, 0 )
    if( btest( tmp, bit ) ) then
       tmp = ibclr( tmp, bit )
       tmp = -tmp
    end if
    CheckSign = tmp
    return
  end function CheckSign

  ! convert 1-8 byte characters into 8-byte integer
  integer function BINCharToInteger8( BI, cval, ival, swap )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=*)          , intent(INOUT) :: cval
    integer(8)                , intent(INOUT) :: ival
    logical         , optional, intent(IN)    :: swap

    integer :: clen

    BINCharToInteger8 = BIN_NOT_CONVERT_INT

    clen = len(cval)

    if( clen == 1 ) then

    else if( clen == 2 ) then

    else if( clen == 3 ) then

    else if( clen == 8 ) then
       if( present( swap ) ) then
          BINCharToInteger8 = BINChar8ToInteger8( BI, cval, ival, swap )
       else
          BINCharToInteger8 = BINChar8ToInteger8( BI, cval, ival       )
       end if

    else
       return
    end if

    return
  end function BINCharToInteger8

  ! convert 8-byte characters into 8-byte integer
  integer function BINChar8ToInteger8( BI, cval, ival, swap )
    type(BINFO)               , intent(INOUT) :: BI
    character(len=8)          , intent(INOUT) :: cval
    integer(8)                , intent(OUT)   :: ival
    logical         , optional, intent(IN)    :: swap


    BINChar8ToInteger8 = BIN_NOT_CONVERT_INT

    BINChar8ToInteger8 = BINReadChar( BI, cval )
    if( BINChar8ToInteger8 /= BIN_NOERROR ) return

    if( present( swap ) ) then
       call Char8ToInteger8( cval, ival, swap )
    else
       call Char8ToInteger8( cval, ival       )
    end if

    BINChar8ToInteger8 = BIN_NOERROR

    return
  end function BINChar8ToInteger8

  subroutine Char8ToInteger8( cval, ival, swap )
    character(len=8) , intent(IN) :: cval
    integer(8)       , intent(INOUT) :: ival
    logical, optional, intent(IN)    :: swap

    integer          :: i, j
    character(len=4) :: cdata

    cdata = cval
    if( present( swap ) ) then
       if( swap ) then
          do i = 1, 8
             j = 9 - i
             cdata(i:i) = cval(j:j)
          end do
       end if
    end if

    ival = transfer( cdata, 0_4 )

    return
  end subroutine Char8ToInteger8

  ! extract character data from cache
  integer function BINReadChar( BI, cval )
    type(BINFO)     , intent(INOUT) :: BI
    character(len=*), intent(OUT)   :: cval

    integer :: clen, status, is, ie
    logical :: TOF

    BINReadChar = BIN_CANNOT_READ_CHAR

#ifndef HLIB
    inquire( BI%IO, OPENED=TOF )
    if( .not. TOF ) return
#endif

    status = NextReadData( BI )
    if( status /= 0 ) return

    clen = len( cval )
    is   = BI%cpos
    ie   = BI%cpos + clen - 1

    call BINReadCharacter( cval,          &
         &                 clen,          &
         &                 BI%Buffer1,    &
         &                 BI%Buffer2,    &
         &                 BI%BufferSize, &
         &                 is,            &
         &                 ie             )
    BI%cpos = BI%cpos + clen
    status = NextReadData( BI )

    BINReadChar = BIN_NOERROR
    return
  end function BINReadChar

  integer function NextReadData( BI )
    type(BINFO), intent(INOUT) :: BI

    integer :: status, ir, cpos, fpos

    NextReadData = BIN_BUFFER_UNSTORED

    if( BI%cpos < 1 ) then
       ir   = abs( BI%cpos / BI%BufferSize ) + 1
       cpos = ir * BI%BufferSize + BI%cpos
       BI%irec = BI%irec - ir
       BI%cpos = cpos

    else if( BI%cpos <= BI%BufferSize ) then
       if( BI%rec1 == BI%irec ) then
          NextReadData = BIN_NOERROR
          return
       end if

    else
       ir   = BI%cpos / BI%BufferSize
       cpos = BI%cpos - ir * BI%BufferSize
       BI%irec = BI%irec + ir
       BI%cpos = cpos
    end if

    BI%Buffer1 = ''
    BI%Buffer2 = ''

    if( BI%irec /= BI%rec1 .and. BI%irec <= BI%MaxLines ) then
#ifdef HLIB
       fpos = ( BI%irec - 1 ) * BI%BufferSize
       call hseek( BI%IO, fpos, SEEK_SET, status, oflag )
       if( status < 0 ) return
       call hread( BI%IO, BI%Buffer1, BI%BufferSize, status, oflag )
       if( status < 0 ) return
#else
       read( BI%IO, rec=BI%irec, IOSTAT=status ) BI%Buffer1
       if( status /= 0 ) return
#endif
       BI%rec1 = BI%irec
    end if

    if( BI%irec+1 /= BI%rec2 .and. BI%irec < BI%MaxLines ) then
#ifdef HLIB
       fpos = BI%irec * BI%BufferSize
       call hseek( BI%IO, fpos, SEEK_SET, status, oflag )
       if( status < 0 ) return
       call hread( BI%IO, BI%Buffer2, BI%BufferSize, status, oflag )
       if( status < 0 ) return
#else
       read( BI%IO, rec=BI%irec+1, IOSTAT=status ) BI%Buffer2
       if( status /= 0 ) return
#endif
       BI%rec2 = BI%irec + 1
    end if

    NextReadData = BIN_NOERROR
    return
  end function NextReadData

end module BIN

subroutine BINReadCharacter( cval, clen, Buffer1, Buffer2, BufferSize, is, ie )
  integer                  , intent(IN)  :: clen
  integer                  , intent(IN)  :: BufferSize
  character(len=clen)      , intent(OUT) :: cval
  character(len=BufferSize), intent(IN)  :: Buffer1
  character(len=BufferSize), intent(IN)  :: Buffer2
  integer                  , intent(IN)  :: is
  integer                  , intent(IN)  :: ie
  
  integer :: ie2
  
  if( ie <= BufferSize ) then
     cval(1:clen) = Buffer1(is:ie)
  else
     ie2 = clen - ( BufferSize - is + 1 )
     cval(1:clen) = Buffer1(is:BufferSize)//Buffer2(1:ie2)
  end if

  return
end subroutine BINReadCharacter
