!
! map module
!
!           AUTHOR: KATO Masaya <mkato@pastel.ocn.ne.jp>
!    FIRST RELEASE: 2002/03/17
!      LAST UPDATE: 2009/07/20
!
module GeoGraphic
  implicit none
  private
  
  integer, parameter :: total_geoparam = 8
  real(8), parameter :: er  = 6371000.0d0            ! Earth Radius ( m )
  real(8), parameter :: pi  = 3.141592653589793238d0 ! pi
  real(8), parameter :: d2r = 1.745329251994330d-2   ! pi / 180.
  real(8), parameter :: r2d = 57.2957795130823d0     ! 180. / pi

  type GEODETIC
    character(len=40) :: Datum
    character(len=40) :: Ellipsoid
    real(8)           :: da
    real(8)           :: df
    real(8)           :: dx
    real(8)           :: dy
    real(8)           :: dz
    real(8)           :: WGS84_a = 6378137.0d0
    real(8)           :: WGS84_f = 1.0d0 / 298.257223563d0
 end type GEODETIC

  ! earth ellipticity
  logical :: Ellipsoid = .false.

  ! map projection ID
  integer, parameter :: PSTEREO = 3       ! polar stereographic
  integer, parameter :: LAMCON  = 1       ! lambert comformal
  integer, parameter :: MERCAT  = 2       ! mercator

  interface Azimuth2Point
     module procedure Azimuth2Point_real
     module procedure Azimuth2Point_double
  end interface

  interface PointAzimuth2Point
     module procedure PointAzimuth2Point_real
     module procedure PointAzimuth2Point_double
  end interface

  interface Dis2GPoint
     module procedure Dis2GPoint_real
     module procedure Dis2GPoint_double
  end interface

  interface xyz2blh
     module procedure xyz2blh_real
     module procedure xyz2blh_double
  end interface

  interface blh2xyz
     module procedure blh2xyz_real
     module procedure blh2xyz_double
  end interface

  interface geodetic_convert
     module procedure geodetic_convert_real
     module procedure geodetic_convert_double
     module procedure geodetic_convert_direct_real
     module procedure geodetic_convert_direct_double
  end interface
  
  type(GEODETIC), dimension(total_geoparam), save :: GEOPARAM
  integer                                         :: GEOREAD = 0

!  public :: GEODETIC
!  public :: GEOPARAM
  public :: geodetic_parameter_print
  public :: geodetic_parameter
  public :: geodetic_convert
  public :: Dis2GPoint
  public :: Azimuth2Point
  public :: PointAzimuth2Point
  public :: xy2ll
  public :: ll2xy

contains
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine PointAzimuth2Point_real( phi0, lam0, theta, arc, phi1, lam1,     &
       &                              earth, major, minor, flat )
    real          , intent(IN)  :: phi0
    real          , intent(IN)  :: lam0
    real          , intent(IN)  :: theta
    real          , intent(IN)  :: arc
    real          , intent(OUT) :: phi1
    real          , intent(OUT) :: lam1
    real, optional, intent(IN)  :: earth
    real, optional, intent(IN)  :: major
    real, optional, intent(IN)  :: minor
    real, optional, intent(IN)  :: flat

    real(8) :: phi1_double
    real(8) :: lam1_double

    if( present( major ) ) then
       ! not supported
    else
       if( present( earth ) ) then
          call PointAzimuth2Point_double( dble(phi0),                       &
               &                          dble(lam0),                       &
               &                          dble(theta),                      &
               &                          dble(arc),                        &
               &                          phi1_double,                      &
               &                          lam1_double,                      &
               &                          earth=dble(earth) )
       else
          call PointAzimuth2Point_double( dble(phi0),                       &
               &                          dble(lam0),                       &
               &                          dble(theta),                      &
               &                          dble(arc),                        &
               &                          phi1_double,                      &
               &                          lam1_double )
       end if
    end if

    phi1 = real( phi1_double )
    lam1 = real( lam1_double )

    return
  end subroutine PointAzimuth2Point_real

  subroutine PointAzimuth2Point_double( phi0, lam0, theta, arc, phi1, lam1,  &
       &                                earth, major, minor, flat )
    real(8)          , intent(IN)  :: phi0
    real(8)          , intent(IN)  :: lam0
    real(8)          , intent(IN)  :: theta
    real(8)          , intent(IN)  :: arc
    real(8)          , intent(OUT) :: phi1
    real(8)          , intent(OUT) :: lam1
    real(8), optional, intent(IN)  :: earth
    real(8), optional, intent(IN)  :: major
    real(8), optional, intent(IN)  :: minor
    real(8), optional, intent(IN)  :: flat

    real(8) :: r
    real(8) :: dis
    real(8) :: rtheta
    real(8) :: rphi0
    real(8) :: rlam0

    ! earth radius
    if( present( earth ) ) then
       r = earth
    else
       r = er
    end if

    ! azimuth
    rtheta = theta
    if( rtheta < -180.0d0 ) rtheta = 360.0d0 + rtheta 
    if( rtheta >  180.0d0 ) rtheta = rtheta - 360.0d0
    rtheta = rtheta * r2d

    ! start point
    rphi0 = phi0 * d2r
    rlam0 = lam0 * d2r

    ! distance
    dis = arc / r

    phi1 = asin( sin(rphi0)*cos(dis) + cos(phi0)*sin(dis)*cos(rtheta) )
    if( abs( phi1 ) > 1d-10 ) then
       lam1 = rlam0 + asin( sin(dis)*sin(rtheta)/cos(phi1) )
    else
       lam1 = rlam0 + sign( 1.0d0, rtheta ) * dis
    end if

    phi1 = phi1 * r2d
    lam1 = lam1 * r2d

    return
  end subroutine PointAzimuth2Point_double

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine Azimuth2Point_real( theta, phi0, lam0, phi, lam, earth,       &
       &                         major, minor, flat )
    real          , intent(OUT) :: theta
    real          , intent(IN)  :: phi0
    real          , intent(IN)  :: lam0
    real          , intent(IN)  :: phi
    real          , intent(IN)  :: lam
    real, optional, intent(IN) :: earth
    real, optional, intent(IN) :: major
    real, optional, intent(IN) :: minor
    real, optional, intent(IN) :: flat

    real(8) :: theta_double

    if( present( major ) ) then
       ! not supported
    else
       if( present( earth ) ) then
          call Azimuth2Point_double( theta_double,                         &
               &                     dble(phi0),                           &
               &                     dble(lam0),                           &
               &                     dble(phi),                            &
               &                     dble(lam),                            &
               &                     earth=dble(earth) )
       else
          call Azimuth2Point_double( theta_double,                         &
               &                     dble(phi0),                           &
               &                     dble(lam0),                           &
               &                     dble(phi),                            &
               &                     dble(lam) )
       end if
    end if

    theta = theta_double

    return
  end subroutine Azimuth2Point_real

  subroutine Azimuth2Point_double( theta, phi0, lam0, phi, lam, earth )
    real(8)          , intent(OUT) :: theta
    real(8)          , intent(IN)  :: phi0
    real(8)          , intent(IN)  :: lam0
    real(8)          , intent(IN)  :: phi
    real(8)          , intent(IN)  :: lam
    real(8), optional, intent(IN)  :: earth

    real(8) :: r
    real(8) :: x
    real(8) :: y
    real(8) :: dlam
    real(8) :: rphi0
    real(8) :: rphi

    if( present( earth ) ) then
       r = earth 
    else
       r = er
    end if

    rphi0 = phi0 * d2r
    rphi  = phi * d2r
    dlam = ( lam - lam0 ) * d2r

    y    = cos( rphi ) * sin( dlam )
    x    = sin( rphi )*cos( rphi0 ) - cos( rphi )*sin( rphi0 )*cos( dlam )

    theta = atan2( y, x ) * r2d

    return
  end subroutine Azimuth2Point_double

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! This subroutine calcurates geographical distance from
! ( phi1, lam1 ) to ( phi2, lam2 ), where phi1, phi2 are
! latitude ( degree ) and lam1, lam2 are longitude (degree).
! Cubic triangle formula is used. Argument r means earth
! radius. If r were not specified, default earth radius
! ( = 6371 km ) be used.
!
  subroutine Dis2GPoint_real( dis, phi1, lam1, phi2, lam2, earth,           &
       &                      major, minor, flat )
    real          , intent(OUT) :: dis
    real          , intent(IN)  :: phi1, lam1, phi2, lam2
    real, optional, intent(IN)  :: earth
    real, optional, intent(IN)  :: major
    real, optional, intent(IN)  :: minor
    real, optional, intent(IN)  :: flat

    real(8) :: dis_double

    if( present( major ) ) then
       if( present( minor ) ) then
          call Dis2GPoint_ellipsoid_double( dis_double,                     &
               &                          dble(phi1),                       &
               &                          dble(lam1),                       &
               &                          dble(phi2),                       &
               &                          dble(lam2),                       &
               &                          dble(major),                      &
               &                          minor=dble(minor) )
       else if( present( flat ) ) then
          call Dis2GPoint_ellipsoid_double( dis_double,                     &
               &                          dble(phi1),                       &
               &                          dble(lam1),                       &
               &                          dble(phi2),                       &
               &                          dble(lam2),                       &
               &                          dble(major),                      &
               &                          flat=dble(flat) )
       else
          dis = -1.0e+37
          return
       end if
    else
       if( present( earth ) ) then
          call Dis2GPoint_double( dis_double,                               &
               &                  dble( phi1 ),                             &
               &                  dble( lam1 ),                             &
               &                  dble( phi2 ),                             &
               &                  dble( lam2 ),                             &
               &                  dble( earth ) )
       else
          call Dis2GPoint_double( dis_double,                               &
               &                  dble( phi1 ),                             &
               &                  dble( lam1 ),                             &
               &                  dble( phi2 ),                             &
               &                  dble( lam2 ) )
       end if
    end if

    dis = real( dis_double, kind=kind(4) )

    return
  end subroutine Dis2GPoint_real

  subroutine Dis2GPoint_double( dis, phi1, lam1, phi2, lam2, earth,         &
       &                        major, minor, flat )
    real(8)          , intent(OUT) :: dis
    real(8)          , intent(IN)  :: phi1, lam1, phi2, lam2
    real(8), optional, intent(IN)  :: earth
    real(8), optional, intent(IN)  :: major
    real(8), optional, intent(IN)  :: minor
    real(8), optional, intent(IN)  :: flat

    real(8) :: arc, r
    real(8) :: rphi1, rphi2, rlam1, rlam2

    dis   = -1.d+37

    ! for ellipsoid
    if( present( major ) ) then
       if( present( minor ) ) then
          call Dis2GPoint_ellipsoid_double( dis,                            &
               &                            phi1,                           &
               &                            lam1,                           &
               &                            phi2,                           &
               &                            lam2,                           &
               &                            major,                          &
               &                            minor=minor )
       else if( present( flat ) ) then
          call Dis2GPoint_ellipsoid_double( dis,                            &
               &                            phi1,                           &
               &                            lam1,                           &
               &                            phi2,                           &
               &                            lam2,                           &
               &                            major,                          &
               &                            flat=flat )
       end if
       return
    end if

    rphi1 = phi1 * d2r
    rphi2 = phi2 * d2r
    rlam1 = lam1 * d2r
    rlam2 = lam2 * d2r

    if( present( earth ) ) then
       r = earth
    else
       r = er
    end if

    arc =  sin( rphi2 ) * sin( rphi1 ) + &
         & cos( rphi2 ) * cos( rphi1 ) * cos( rlam2 - rlam1 )
    arc = acos( arc )

    dis = r * arc

    return
  end subroutine Dis2GPoint_double

  subroutine Dis2GPoint_ellipsoid_real( dis, phi1, lam1, phi2, lam2,        &
       &                                major, minor, flat )
    real          , intent(OUT) :: dis
    real          , intent(IN)  :: phi1
    real          , intent(IN)  :: lam1
    real          , intent(IN)  :: phi2
    real          , intent(IN)  :: lam2
    real          , intent(IN)  :: major
    real, optional, intent(IN)  :: minor
    real, optional, intent(IN)  :: flat

    real(8) :: dis_double

    dis = -1.0e+37

    if( present( minor ) ) then
       call Dis2GPoint_ellipsoid_double( dis_double,                        &
            &                            dble(phi1),                        &
            &                            dble(lam1),                        &
            &                            dble(phi2),                        &
            &                            dble(lam2),                        &
            &                            dble(major),                       &
            &                            minor=dble(minor) )
    else if( present( flat ) ) then
       call Dis2GPoint_ellipsoid_double( dis_double,                        &
            &                            dble(phi1),                        &
            &                            dble(lam1),                        &
            &                            dble(phi2),                        &
            &                            dble(lam2),                        &
            &                            dble(major),                       &
            &                            flat=dble(flat) )
    else
       return
    end if

    dis = real( dis_double, kind=kind(4) )

    return
  end subroutine Dis2GPoint_ellipsoid_real

  subroutine Dis2GPoint_ellipsoid_double( dis, phi1, lam1, phi2, lam2,      &
       &                                  major, minor, flat )
    real(8)          , intent(OUT) :: dis
    real(8)          , intent(IN)  :: phi1
    real(8)          , intent(IN)  :: lam1
    real(8)          , intent(IN)  :: phi2
    real(8)          , intent(IN)  :: lam2
    real(8)          , intent(IN)  :: major
    real(8), optional, intent(IN)  :: minor
    real(8), optional, intent(IN)  :: flat

    integer, parameter :: max_count = 100

    integer :: i

    real(8) :: f
    real(8) :: b
    real(8) :: u2p
    real(8) :: rphi1
    real(8) :: rlam1
    real(8) :: rphi2
    real(8) :: rlam2
    real(8) :: U1
    real(8) :: U2
    real(8) :: dL
    real(8) :: dlam
    real(8) :: dlam0

    real(8) :: sigma
    real(8) :: dsigma
    real(8) :: sin_sigma
    real(8) :: cos_sigma
    real(8) :: sin_alpha
    real(8) :: cos_alpha2
    real(8) :: cos_2sigma_m
    real(8) :: A
    real(8) :: BB
    real(8) :: C

    real(8) :: dtmp1
    real(8) :: dtmp2

    dis = -1.0d+37

    if( present( flat ) ) then
       f = flat
    else if( present( minor ) ) then
       f = ( major - minor ) / major
    else
       return
    end if

    if( present( minor ) ) then
       b = minor
    else
       b = major * ( 1.0d0 - f )
    end if

    rphi1 = phi1 * d2r
    rphi2 = phi2 * d2r
    rlam1 = lam1 * d2r
    rlam2 = lam2 * d2r

    U1  = atan( ( 1.d0 - f ) * tan( rphi1 ) )
    U2  = atan( ( 1.d0 - f ) * tan( rphi2 ) )
    dL  = rlam2 - rlam1

    i = 0
    dlam  = dL
    do
       i = i + 1
       if( i > max_count ) return
       dtmp1 = cos(U2) * sin(dlam)
       dtmp2 = cos(U1) * sin(U2) - sin(U1) * cos(U2) * cos(dlam)
       sin_sigma = sqrt( dtmp1*dtmp1 + dtmp2*dtmp2 )
       cos_sigma = sin(U1) * sin(U2) + cos(U1) * cos(U2) * cos(dlam)
       sigma = atan2( sin_sigma, cos_sigma )

       sin_alpha  = cos(U1) * cos(U2) * sin(dlam) / sin_sigma
       cos_alpha2 = max( 1.d0 - sin_alpha * sin_alpha, 0.0d0 )
       cos_2sigma_m = cos_sigma - 2.d0 * sin(U1) * sin(U2) / cos_alpha2

       C = f / 16.0d0 * cos_alpha2                                          &
            & * ( 4.0d0 + f * ( 4.0d0 - 3.0d0 * cos_alpha2 ) )
       dlam0 = dlam
       dlam = dL + ( 1.0d0 - C ) * f * sin_alpha * ( sigma                  &
            &    + C * sin_sigma * ( cos_2sigma_m + C * cos_sigma           &
            &    * ( -1.0d0 + 2.0d0 * cos_2sigma_m * cos_2sigma_m ) ) )
       if( abs( dlam - dlam0 ) < 1.0d-12 ) exit
    end do

    u2p = cos_alpha2 * ( major*major - b*b ) / ( b*b )
    A   = 1.0d0 + u2p/16384.d0 * ( 4096.0d0 + u2p                           &
         &    * ( -768.0d0 + u2p * ( 320.d0 - 175.0d0 * u2p ) ) )


    BB  = u2p/1024.0d0 * ( 256.0d0 + u2p * ( -128.0d0 + u2p                 &
         &    * ( 74.0d0 - 47.0d0 * u2p ) ) )


    dsigma = BB * sin_sigma * ( cos_2sigma_m + 0.25d0 * BB * ( cos_sigma    &
         &     * ( -1.0d0 + 2.0d0 * cos_2sigma_m * cos_2sigma_m )           &
         &     - 1.d0/6.0d0 * BB * cos_2sigma_m * ( -3.0d0 + 4.0d0          &
         &     * sin_sigma * sin_sigma ) * ( -3.0d0                         &
         &      + 4.0d0 * cos_2sigma_m * cos_2sigma_m ) ) )

    dis = b * A * ( sigma - dsigma )

    return

  end subroutine Dis2GPoint_ellipsoid_double

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine xyz2blh_real( X, Y, Z, b, l, h, major, minor, flat )
    real          , intent(IN)  :: X
    real          , intent(IN)  :: Y
    real          , intent(IN)  :: Z
    real          , intent(OUT) :: b
    real          , intent(OUT) :: l
    real          , intent(OUT) :: h
    real          , intent(IN)  :: major
    real, optional, intent(IN)  :: minor
    real, optional, intent(IN)  :: flat

    real(8) :: b_double
    real(8) :: l_double
    real(8) :: h_double

    b = -1.0e+37
    l = -1.0e+37
    h = -1.0e+37
    
    if( present( minor ) ) then
       call xyz2blh_double( dble(X),                                        &
            &               dble(Y),                                        &
            &               dble(Z),                                        &
            &               b_double,                                       &
            &               l_double,                                       &
            &               h_double,                                       &
            &               dble(major),                                    &
            &               minor=dble(minor) )
    else if( present( flat ) ) then
       call xyz2blh_double( dble(X),                                        &
            &               dble(Y),                                        &
            &               dble(Z),                                        &
            &               b_double,                                       &
            &               l_double,                                       &
            &               h_double,                                       &
            &               dble(major),                                    &
            &               flat=dble(flat) )
    else
       return
    end if

    b = real( b_double, kind=kind(4) )
    l = real( l_double, kind=kind(4) )
    h = real( h_double, kind=kind(4) )

    return
  end subroutine xyz2blh_real

  subroutine xyz2blh_double( X, Y, Z, b, l, h, major, minor, flat )
    real(8)          , intent(IN)  :: X
    real(8)          , intent(IN)  :: Y
    real(8)          , intent(IN)  :: Z
    real(8)          , intent(OUT) :: b
    real(8)          , intent(OUT) :: l
    real(8)          , intent(OUT) :: h
    real(8)          , intent(IN)  :: major
    real(8), optional, intent(IN)  :: minor
    real(8), optional, intent(IN)  :: flat

    real(8) :: e2
    real(8) :: p
    real(8) :: r
    real(8) :: mu
    real(8) :: sin3_mu
    real(8) :: cos3_mu
    real(8) :: db
    real(8) :: f

    b = -1.0d+37
    l = -1.0d+37
    h = -1.0d+37

    if( present( flat ) ) then
       f = flat
    else if( present( minor ) ) then
       f = ( major - minor ) / major
    else
       return
    end if
    
    e2 = f * ( 2.0d0 - f )
    p  = sqrt( X*X + Y*Y )
    r  = sqrt( X*X + Y*Y + Z*Z )
    mu = atan( Z/p * ( ( 1.0d0 - f ) + e2 * major / r ) )

    sin3_mu = sin(mu) * sin(mu) * sin(mu)
    cos3_mu = cos(mu) * cos(mu) * cos(mu)

    l = atan2( Y, X )
    b = atan(   ( Z * ( 1.0d0 - f ) + e2 * major * sin3_mu )                &
         &    / ( ( 1.0d0 - f ) * ( p - e2 * major * cos3_mu ) ) )
    h = p * cos( b ) + Z * sin( b ) - major * sqrt( 1.0d0 - e2*sin(b)*sin(b) )

    l = l * r2d
    b = b * r2d
    h = h

    return
  end subroutine xyz2blh_double

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine blh2xyz_real( B, L, H, x, y, z, major, minor, flat )
    real          , intent(IN)  :: B
    real          , intent(IN)  :: L
    real          , intent(IN)  :: H
    real          , intent(OUT) :: x
    real          , intent(OUT) :: y
    real          , intent(OUT) :: z
    real          , intent(IN)  :: major
    real, optional, intent(IN)  :: minor
    real, optional, intent(IN)  :: flat

    real(8) :: x_double
    real(8) :: y_double
    real(8) :: z_double

    x = -1.0e+37
    y = -1.0e+37
    z = -1.0e+37

    if( present( minor ) ) then
       call blh2xyz_double( dble(B),                                        &
            &               dble(L),                                        &
            &               dble(H),                                        &
            &               x_double,                                       &
            &               y_double,                                       &
            &               z_double,                                       &
            &               dble(major),                                    &
            &               minor=dble(minor) )
    else if( present( flat ) ) then
       call blh2xyz_double( dble(B),                                        &
            &               dble(L),                                        &
            &               dble(H),                                        &
            &               x_double,                                       &
            &               y_double,                                       &
            &               z_double,                                       &
            &               dble(major),                                    &
            &               minor=dble(minor) )
    else
       return
    end if

    return
  end subroutine blh2xyz_real

  subroutine blh2xyz_double( B, L, H, x, y, z, major, minor, flat )
    real(8)          , intent(IN)  :: B
    real(8)          , intent(IN)  :: L
    real(8)          , intent(IN)  :: H
    real(8)          , intent(OUT) :: x
    real(8)          , intent(OUT) :: y
    real(8)          , intent(OUT) :: z
    real(8)          , intent(IN)  :: major
    real(8), optional, intent(IN)  :: minor
    real(8), optional, intent(IN)  :: flat

    real(8) :: N
    real(8) :: e2
    real(8) :: f
    real(8) :: rB
    real(8) :: rL

    rB = B * d2r
    rL = L * d2r

    x = -1.0d+37
    y = -1.0d+37
    z = -1.0d+37
    
    if( present( flat ) ) then
       f = flat
    else if( present( minor ) ) then
       f = ( major - minor ) / major
    else
       return
    end if

    e2 = f * ( 2.0d0 - f )
    N  = major / sqrt( 1.0d0 - e2 * sin(rB) * sin(rB) )
    
    x  = ( N + H ) * cos(rB) * cos(rL)
    y  = ( N + H ) * cos(rB) * sin(rL)
    z  = ( N * ( 1.0d0 - e2 ) + H ) * sin(rB)

    return
  end subroutine blh2xyz_double

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! IN:
!     x, y         : grid indices
!     cproj        : map projection name
!     dx, dy       : grid space ( m )
!     clon, clat   : projection center longitude and latitude
!     xc, yc       : indices corresponding to center longitude and latitude
!     tlat1, tlat2 : true latitude 1 and 2 ( some map projection used )
!     en           : ellipsoid ID ( if earth is ellipsoidal. )
!     r            : earth radius ( or semimajor axis )
!     e            : eccentricity
!
! OUT:
!     lon, lat     : longitude, latitude corresponding to ( x, y )
!
  subroutine xy2ll( x, y, lon, lat, cproj, dx, dy, clon, clat, &
       xc, yc, tlat1, tlat2, en, earth, e )
    implicit none
    real                 , intent(OUT) :: lon, lat
    real                 , intent(IN)  :: x, y
    character ( len = * ), intent(IN)  :: cproj
    real                 , intent(IN)  :: dx, dy, clon, clat, xc, yc
    real   , optional              :: tlat1, tlat2, en, earth, e

    real :: A1, A2, YYC, psi, psi1, psi2, L1
    real :: POLE, SIGN, X2, Y2, R2, xn, T1, T2, r
    real :: tl1, tl2

    if( present( en ) .or. present( e ) ) Ellipsoid = .true.
    if( present( earth ) ) then
       r = earth
    else
       r = er
    end if
    if( r.lt.10000 ) r = r * 1000.

    if( Ellipsoid ) then
       write(6,*) 'Under construction, then treats as sphere'
       Ellipsoid = .false.

!    else
       
    end if

    if( clat.gt.0 ) then
       POLE =  90.
       SIGN = 1.
    else
       POLE = -90.
       SIGN = -1.
    end if

    if( cproj(1:6) .eq. 'LAMCON' ) then ! lambert comformal projection
       if( ( .not. present( tlat1 ) ) .or. ( .not. present( tlat2 ) ) ) then
          write(6,*) 'WARNING: TLAT1 or TLAT2 is not set'
          tl1 = SIGN * 60
          tl2 = SIGN * 30
       else
          tl1 = tlat1
          tl2 = tlat2
       end if
          
       psi1 = ( POLE - tl1 ) * d2r
       psi2 = ( POLE - tl2 ) * d2r
    
       xn = ( log( cos( tl1 * d2r ) ) - log( cos( tl2 * d2r ) ) ) / &
            & ( log( tan( .25 * pi - .5 * d2r * abs( tl1 ) ) ) - &
            &   log( tan( .25 * pi - .5 * d2r * abs( tl2 ) ) ) )

       YYC  = r / xn * sin( psi1 ) * &
            & ( tan( .5 * d2r * ( POLE - clat ) ) / tan( .5 * psi1 ) )**xn
       X2   = ( x - xc ) * dx
       Y2   = ( y - yc ) * dy - YYC
       R2   = sqrt( X2 * X2 + Y2 * Y2 )

       T1 = tan( .5 * psi1 )
       T2 = ( R2 * xn / r / abs( sin( psi1 ) ) )**(1./xn)

       lon = clon + atan2( X2, - SIGN * Y2 ) / d2r / xn
       lat = POLE - 2. * atan2( T1 * T2, 1. ) / d2r

    else if( cproj(1:7) .eq. 'PSTEREO' ) then  ! polar stereographic
       if( .not. present( tlat1 ) ) then
          write(6,*) 'WARNING: TLAT1 is not set'
          tl1 = SIGN * 90.
       else
          tl1 = tlat1
       end if

       psi1 = d2r * ( POLE - tl1 )
       if( abs(tl1).gt.89.999 ) then
          A1 = 2. * r
       else
          A1 = r * sin( psi1 ) / tan( .5 * psi1 )
       end if

       YYC = A1 * tan( d2r * .5 * ( POLE - clat ) )
       X2  = ( x - xc ) * dx
       Y2  = ( y - yc ) * dy - YYC
       R2  = sqrt( X2 * X2 + Y2 * Y2 )
       A2  = R2 / A1

       lon = clon + atan2( X2, - SIGN * Y2 )
       lat = SIGN * ( 90. - 2. * atan2( A2, 1. ) / d2r )

       if( lon.ge.360.) lon = lon - 360.

    else if( cproj(1:6) .eq. 'MERCAT' ) then  ! polar stereographic
       tl1 = 0.
       A1  = r * cos( tl1 )
       YYC = A1 * log( tan( 0.25 * pi + 0.5 * d2r * clat ) ) / dy

       lon = clon + dx / A1 * ( x - xc ) / d2r
       lat = 2. * atan2( exp( dy / a1 * ( y - yc + YYC ) ), 1. ) / d2r - 90.
       
    end if

    return
  end subroutine xy2ll

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! IN:
!     lon, lat     : longitude, latitude
!     cproj        : map projection ID
!     dx, dy       : grid space ( m )
!     clon, clat   : projection center longitude and latitude
!     xc, yc       : indices corresponding to center longitude and latitude
!     tlat1, tlat2 : true latitude 1 and 2 ( some map projection used )
!     en           : ellipsoid ID ( if earth is ellipsoidal. )
!     r            : earth radius ( or semimajor axis )
!     e            : eccentricity
!
! OUT:
!     x, y         : indices corresponding to ( lon, lat )
!
  subroutine ll2xy( lon, lat, x, y, cproj, dx, dy, clon, clat, &
       xc, yc, tlat1, tlat2, en, earth, e )
    real                , intent(IN)  :: lon, lat
    real                , intent(OUT) :: x, y
    character( len = * ), intent(IN)  :: cproj
    real                , intent(IN)  :: dx, dy, clon, clat, xc, yc
    real   , optional   , intent(IN)  :: tlat1, tlat2, en, earth, e

    real :: A1, A2, YYC, dlon, psi1, tl1, tl2, r, POLE, SIGN, xn

    if( present( en ) .or. present( e ) ) Ellipsoid = .true.
    if( present( earth ) ) then
       r = earth
    else
       r = er
    end if
    if( r.lt.10000 ) r = r * 1000.

    if( Ellipsoid ) then
       write(6,*) 'Under construction, then treats as sphere'
       Ellipsoid = .false.

!    else
       
    end if

    if( clat.gt.0 ) then
       POLE =  90.
       SIGN = 1.
    else
       POLE = -90.
       SIGN = -1.
    end if

    if( cproj(1:6) .eq. 'LAMCON' ) then ! lambert comformal
       if( ( .not.present( tlat1 ) ) .or. ( .not.present( tlat2 ) ) ) then
          write(6,*) 'WARNING: TLAT1 or TLAT2 is not set'
          tl1 = SIGN * 60.
          tl2 = SIGN * 30.
       else
          tl1 = tlat1
          tl2 = tlat2
       end if

       xn = ( log( cos( tl1 * d2r ) ) - log( cos( tl2 * d2r ) ) ) / &
            & ( log( tan( .25 * pi - .5 * d2r * abs( tl1 ) ) ) - &
            &   log( tan( .25 * pi - .5 * d2r * abs( tl2 ) ) ) )

       psi1 = ( POLE - tl1 ) * d2r
       YYC  = r / xn * sin( psi1 ) * &
            & ( tan( .5 * d2r * ( POLE - clat ) ) / tan( .5 * psi1 ) )**xn
       A1   = r / xn * sin( psi1 ) * &
            & ( tan( .5 * d2r * ( POLE -  lat ) ) / tan( .5 * psi1 ) )**xn
       dlon = xn * ( lon - clon ) * d2r

       x = xc + SIGN * A1 * sin( dlon ) / dx
       y = yc -        A1 * cos( dlon ) / dy + YYC / dy

    else if( cproj(1:7) .eq. 'PSTEREO' ) then  ! polar stereographic
       if( .not. present( tlat1 ) ) then
          write(6,*) 'WARNING: TLAT1 is not set'
          tl1 = POLE * 90.
       else
          tl1 = tlat1
       end if

       psi1 = d2r * ( POLE - tl1 )
       if( abs(tl1).gt.89.999 ) then
          A1 = 2. * r
       else
          A1 = r * sin( psi1 ) / tan( .5 * psi1 )
       end if
       A2   = A1 * tan( .5 * d2r * ( POLE -  lat ) )
       YYC  = A1 * tan( .5 * d2r * ( POLE - clat ) )
       dlon = d2r * ( lon - clon - 90. )

       x =   A2 * cos( dlon )         / dx + xc
       y = ( A2 * sin( dlon ) + YYC ) / dy + yc
       
    else if( cproj(1:6) .eq. 'MERCAT' ) then  ! Mercator
       tl1 = 0.
       A1  = r * cos( tl1 )
       YYC = A1 * log( tan( 0.25 * pi + 0.5 * d2r * clat ) )

       x = xc + A1 * ( lon - clon ) * d2r / dx
       y = yc + ( A1 * log( tan( 0.25 * pi + 0.5 * d2r * lat ) ) - YYC ) / dy

    end if

    return
  end subroutine ll2xy

  subroutine geodetic_convert_direct_real( from_a,                          &
       &                                   from_f,                          &
       &                                   from_dx,                         &
       &                                   from_dy,                         &
       &                                   from_dz,                         &
       &                                   from_lat,                        &
       &                                   from_lon,                        &
       &                                   from_height,                     &
       &                                   to_a,                            &
       &                                   to_f,                            &
       &                                   to_dx,                           &
       &                                   to_dy,                           &
       &                                   to_dz,                           &
       &                                   to_lat,                          &
       &                                   to_lon,                          &
       &                                   to_height )
    real, intent(IN)  :: from_a
    real, intent(IN)  :: from_f
    real, intent(IN)  :: from_dx
    real, intent(IN)  :: from_dy
    real, intent(IN)  :: from_dz
    real, intent(IN)  :: from_lat
    real, intent(IN)  :: from_lon
    real, intent(IN)  :: from_height
    real, intent(IN)  :: to_a
    real, intent(IN)  :: to_f
    real, intent(IN)  :: to_dx
    real, intent(IN)  :: to_dy
    real, intent(IN)  :: to_dz
    real, intent(OUT) :: to_lat
    real, intent(OUT) :: to_lon
    real, intent(OUT) :: to_height

    real(8) :: to_lat_double
    real(8) :: to_lon_double
    real(8) :: to_height_double

    call geodetic_convert_direct_double( dble(from_a),                      &
         &                               dble(from_f),                      &
         &                               dble(from_dx),                     &
         &                               dble(from_dy),                     &
         &                               dble(from_dz),                     &
         &                               dble(from_lat),                    &
         &                               dble(from_lon),                    &
         &                               dble(from_height),                 &
         &                               dble(to_a),                        &
         &                               dble(to_f),                        &
         &                               dble(to_dx),                       &
         &                               dble(to_dy),                       &
         &                               dble(to_dz),                       &
         &                               to_lat_double,                     &
         &                               to_lon_double,                     &
         &                               to_height_double )

    to_lat    = real( to_lat_double, kind=kind(4) )
    to_lon    = real( to_lon_double, kind=kind(4) )
    to_height = real( to_height_double, kind=kind(4) )
    return
  end subroutine geodetic_convert_direct_real

  subroutine geodetic_convert_direct_double( from_a,                        &
       &                                     from_f,                        &
       &                                     from_dx,                       &
       &                                     from_dy,                       &
       &                                     from_dz,                       &
       &                                     from_lat,                      &
       &                                     from_lon,                      &
       &                                     from_height,                   &
       &                                     to_a,                          &
       &                                     to_f,                          &
       &                                     to_dx,                         &
       &                                     to_dy,                         &
       &                                     to_dz,                         &
       &                                     to_lat,                        &
       &                                     to_lon,                        &
       &                                     to_height )
    real(8), intent(IN)  :: from_a
    real(8), intent(IN)  :: from_f
    real(8), intent(IN)  :: from_dx
    real(8), intent(IN)  :: from_dy
    real(8), intent(IN)  :: from_dz
    real(8), intent(IN)  :: from_lat
    real(8), intent(IN)  :: from_lon
    real(8), intent(IN)  :: from_height
    real(8), intent(IN)  :: to_a
    real(8), intent(IN)  :: to_f
    real(8), intent(IN)  :: to_dx
    real(8), intent(IN)  :: to_dy
    real(8), intent(IN)  :: to_dz
    real(8), intent(OUT) :: to_lat
    real(8), intent(OUT) :: to_lon
    real(8), intent(OUT) :: to_height

    real(8) :: x_from
    real(8) :: y_from
    real(8) :: z_from
    real(8) :: x_to
    real(8) :: y_to
    real(8) :: z_to

    to_lat    = -1.0d+37
    to_lon    = -1.0d+37
    to_height = -1.0d+37

    call blh2xyz_double( from_lat,                                          &
         &               from_lon,                                          &
         &               from_height,                                       &
         &               x_from,                                            &
         &               y_from,                                            &
         &               z_from,                                            &
         &               from_a,                                            &
         &               flat=from_f )

    x_to = x_from + from_dx - to_dx
    y_to = y_from + from_dy - to_dy
    z_to = z_from + from_dz - to_dz

    call xyz2blh_double( x_to,                                              &
         &               y_to,                                              &
         &               z_to,                                              &
         &               to_lat,                                            &
         &               to_lon,                                            &
         &               to_height,                                         &
         &               to_a,                                              &
         &               flat=to_f )

    return
  end subroutine geodetic_convert_direct_double

  subroutine geodetic_convert_real( from_datum,                             &
       &                            from_ellipsoid,                         &
       &                            from_lat,                               &
       &                            from_lon,                               &
       &                            from_height,                            &
       &                            to_datum,                               &
       &                            to_ellipsoid,                           &
       &                            to_lat,                                 &
       &                            to_lon,                                 &
       &                            to_height )
    character(len=*), intent(IN)  :: from_datum
    character(len=*), intent(IN)  :: from_ellipsoid
    real            , intent(IN)  :: from_lat
    real            , intent(IN)  :: from_lon
    real            , intent(IN)  :: from_height
    character(len=*), intent(IN)  :: to_datum
    character(len=*), intent(IN)  :: to_ellipsoid
    real            , intent(OUT) :: to_lat
    real            , intent(OUT) :: to_lon
    real            , intent(OUT) :: to_height

    real(8) :: to_lat_double
    real(8) :: to_lon_double
    real(8) :: to_height_double

    call geodetic_convert_double( from_datum,                               &
         &                        from_ellipsoid,                           &
         &                        dble(from_lat),                           &
         &                        dble(from_lon),                           &
         &                        dble(from_height),                        &
         &                        to_datum,                                 &
         &                        to_ellipsoid,                             &
         &                        to_lat_double,                            &
         &                        to_lon_double,                            &
         &                        to_height_double )

    to_lat    = real( to_lat_double, kind=kind(4) )
    to_lon    = real( to_lon_double, kind=kind(4) )
    to_height = real( to_height_double, kind=kind(4) )

    return
  end subroutine geodetic_convert_real

  subroutine geodetic_convert_double( from_datum,                           &
       &                              from_ellipsoid,                       &
       &                              from_lat,                             &
       &                              from_lon,                             &
       &                              from_height,                          &
       &                              to_datum,                             &
       &                              to_ellipsoid,                         &
       &                              to_lat,                               &
       &                              to_lon,                               &
       &                              to_height )
    character(len=*), intent(IN)  :: from_datum
    character(len=*), intent(IN)  :: from_ellipsoid
    real(8)         , intent(IN)  :: from_lat
    real(8)         , intent(IN)  :: from_lon
    real(8)         , intent(IN)  :: from_height
    character(len=*), intent(IN)  :: to_datum
    character(len=*), intent(IN)  :: to_ellipsoid
    real(8)         , intent(OUT) :: to_lat
    real(8)         , intent(OUT) :: to_lon
    real(8)         , intent(OUT) :: to_height

    integer :: num_from
    integer :: num_to
    real(8) :: a_from
    real(8) :: f_from
    real(8) :: x_from
    real(8) :: y_from
    real(8) :: z_from
    real(8) :: a_to
    real(8) :: f_to
    real(8) :: x_to
    real(8) :: y_to
    real(8) :: z_to

    real(8) :: dx
    real(8) :: dy
    real(8) :: dz

    to_lat    = -1.0d+37
    to_lon    = -1.0d+37
    to_height = -1.0d+37

    call geodetic_parameter( from_datum, from_ellipsoid,                    &
         &                   a_from, f_from, num_from )
    call geodetic_parameter( to_datum, to_ellipsoid, a_to, f_to, num_to )

    dx = GEOPARAM(num_from)%dx - GEOPARAM(num_to)%dx
    dy = GEOPARAM(num_from)%dy - GEOPARAM(num_to)%dy
    dz = GEOPARAM(num_from)%dz - GEOPARAM(num_to)%dz

    call blh2xyz_double( from_lat,                                          &
         &               from_lon,                                          &
         &               from_height,                                       &
         &               x_from,                                            &
         &               y_from,                                            &
         &               z_from,                                            &
         &               a_from,                                            &
         &               flat=f_from )

    x_to = x_from + dx
    y_to = y_from + dy
    z_to = z_from + dz

    call xyz2blh_double( x_to,                                              &
         &               y_to,                                              &
         &               z_to,                                              &
         &               to_lat,                                            &
         &               to_lon,                                            &
         &               to_height,                                         &
         &               a_to,                                              &
         &               flat=f_to )

    return
  end subroutine geodetic_convert_double

  subroutine geodetic_parameter( datum, ellipsoid, a, f, num )
    character(len=*)          , intent(IN)  :: datum
    character(len=*)          , intent(IN)  :: ellipsoid
    real(8)                   , intent(OUT) :: a
    real(8)                   , intent(OUT) :: f
    integer         , optional, intent(OUT) :: num

    integer :: i
    integer :: clen1
    integer :: clen2

    a   = -1.0d+37
    f   = -1.0d+37
    if( present( num ) ) num = -1
    clen1 = len_trim( datum )
    clen2 = len_trim( ellipsoid )

    if( GEOREAD == 0 ) call GEOINIT

    do i = 1, total_geoparam
       if(    GEOPARAM(i)%Datum(1:clen1+1)                                  &
            &        == datum(1:clen1)//' '   .and.                         &
            & GEOPARAM(i)%Ellipsoid(1:clen2+1)                              &
            &        == ellipsoid(1:clen2)//' ' ) then
          a = GEOPARAM(i)%WGS84_a - GEOPARAM(i)%da
          f = GEOPARAM(i)%WGS84_f - GEOPARAM(i)%df
          if( present( num ) ) num = i
          exit
       end if

    end do

    return
  end subroutine geodetic_parameter

  subroutine geodetic_parameter_print
    integer :: i
    real(8) :: finv

    if( GEOREAD == 0 ) call GEOINIT

    write(6,*)  '===================================' &
         &   // '==================================='
    write(6,*) 'GEODETIC DATUM: (NAME OF GEODETIC)'
    write(6,*) 'ELLIPSOID NAME: (NAME OF ELLIPSOID)'
    write(6,*) 'MAJOR_AXIS    INVERSE_OF_FLATTENING'
    write(6,*) 'DX    DY    DZ    ( WGS84 - THIS )'
    write(6,*)  '-----------------------------------' &
         &   // '-----------------------------------'
    do i = 1, total_geoparam
       finv = 1.0d0 / GEOPARAM(i)%WGS84_f - GEOPARAM(i)%df

       write(6,*) 'GEODETIC DATUM: ', GEOPARAM(i)%Datum
       write(6,*) 'ELLIPSOID NAME: ', GEOPARAM(i)%Ellipsoid
       write(6,*) GEOPARAM(i)%WGS84_a - GEOPARAM(i)%da, finv
       write(6,*) GEOPARAM(i)%dx, GEOPARAM(i)%dy, GEOPARAM(i)%dz
       write(6,*)
    end do
    write(6,*)  '===================================' &
         &   // '==================================='
    
    return
  end subroutine geodetic_parameter_print

  subroutine GEOINIT
    ! WGS-84
    GEOPARAM(1)%Datum     = 'WGS-84                                  '
    GEOPARAM(1)%Ellipsoid = 'WGS-84                                  '
    GEOPARAM(1)%da        = 0.0d0
    GEOPARAM(1)%df        = 0.0d0
    GEOPARAM(1)%dx        = 0.0d0
    GEOPARAM(1)%dy        = 0.0d0
    GEOPARAM(1)%dz        = 0.0d0

    ! Tokyo_mean
    GEOPARAM(2)%Datum     = 'TOKYO_MEAN                              '
    GEOPARAM(2)%Ellipsoid = 'BESSEL_1841                             '
    GEOPARAM(2)%da        =  739.845d0
    GEOPARAM(2)%df        =    0.000010037483d0 ! 0.10037483d0
    GEOPARAM(2)%dx        = -148.0d0
    GEOPARAM(2)%dy        =  507.0d0
    GEOPARAM(2)%dz        =  685.0d0

    ! Tokyo
    GEOPARAM(3)%Datum     = 'TOKYO                                   '
    GEOPARAM(3)%Ellipsoid = 'BESSEL_1841                             '
    GEOPARAM(3)%da        =  GEOPARAM(2)%WGS84_a - 6377397.155d0
    GEOPARAM(3)%df        =  GEOPARAM(2)%WGS84_f - 1.0d0/299.1528128d0
    GEOPARAM(3)%dx        = -148.0d0
    GEOPARAM(3)%dy        =  507.0d0
    GEOPARAM(3)%dz        =  685.0d0

    ! Tokyo_Okinawa
    GEOPARAM(4)%Datum     = 'TOKYO_OKINAWA                           '
    GEOPARAM(4)%Ellipsoid = 'BESSEL_1841                             '
    GEOPARAM(4)%da        =  739.845d0
    GEOPARAM(4)%df        =    0.000010037483d0 ! 0.10037483d0
    GEOPARAM(4)%dx        = -158.0d0
    GEOPARAM(4)%dy        =  507.0d0
    GEOPARAM(4)%dz        =  676.0d0

    ! Tokyo_South_Korea
    GEOPARAM(5)%Datum     = 'TOKYO_SOUTH_KOREA                       '
    GEOPARAM(5)%Ellipsoid = 'BESSEL_1841                             '
    GEOPARAM(5)%da        =  739.845d0
    GEOPARAM(5)%df        =    0.000010037483d0 ! 0.10037483d0
    GEOPARAM(5)%dx        = -147.0d0
    GEOPARAM(5)%dy        =  506.0d0
    GEOPARAM(5)%dz        =  687.0d0

    ! KOREAN_GEODETIC_SYSTEM_1995
    GEOPARAM(6)%Datum     = 'KOREAN_GEODETIC_SYSTEM_1995             '
    GEOPARAM(6)%Ellipsoid = 'WGS-84                                  '
    GEOPARAM(6)%da        =    0.0d0
    GEOPARAM(6)%df        =    0.0d0
    GEOPARAM(6)%dx        =    0.0d0
    GEOPARAM(6)%dy        =    0.0d0
    GEOPARAM(6)%dz        =    0.0d0

    ! ITRF-94 (JGD2000)
    GEOPARAM(7)%Datum     = 'JGD2000                                 '
    GEOPARAM(7)%Ellipsoid = 'GRS-80                                  '
    GEOPARAM(7)%da        =  GEOPARAM(2)%WGS84_a - 6378137.0d0
    GEOPARAM(7)%df        =  GEOPARAM(2)%WGS84_f - 1.0d0/298.257222101d0
    GEOPARAM(7)%dx        =    0.0d0
    GEOPARAM(7)%dy        =    0.0d0
    GEOPARAM(7)%dz        =    0.0d0

    ! ITRF-94 (JGD2000)
    GEOPARAM(8)%Datum     = 'JGD2000_MLITT                           '
    GEOPARAM(8)%Ellipsoid = 'GRS-80                                  '
    GEOPARAM(8)%da        =  GEOPARAM(2)%WGS84_a - 6378137.0d0
    GEOPARAM(8)%df        =  GEOPARAM(2)%WGS84_f - 1.0d0/298.257222101d0
    GEOPARAM(8)%dx        = -1.126d0
    GEOPARAM(8)%dy        = -0.077d0
    GEOPARAM(8)%dz        = -0.037d0

    GEOREAD = 1

    return
  end subroutine GEOINIT

end module GeoGraphic
