module ModelInfo
  use MM5
  use GRIB2
  use string
  use GeoGraphic
  implicit none
  private

  type MINFO_T
     character(len=24)  :: cdate
     character(len=128) :: File
     real               :: xtime
     integer            :: ipos
     character(len=128), dimension(:), pointer :: Files
  end type MINFO_T

  type MINFO_V
     character(len=20) :: name
     integer           :: ix, jx, kx, it
     integer           :: rec
     logical           :: staggering
     real              :: scale_factor
     real              :: add_offset
     character(len=80) :: desc
     character(len=50) :: unit
  end type MINFO_V

  type MINFO
     integer :: TotalVariables
     integer :: TotalInfo
     character(len=20) :: DataType
     character(len=6)  :: Endian
     character(len=20) :: vunit
     character(len=20) :: hunit
     integer           :: ByteSize
     logical           :: Signed
     logical           :: Vslice
     real              :: Undef
     integer(1)        :: Undef1i
     integer(2)        :: Undef2i
     integer           :: mproj
     character(len=10) :: cproj
     real              :: cone
     real              :: tlat1
     real              :: tlat2
     real              :: tlon
     real              :: clat
     real              :: clon
     real              :: slat
     real              :: slon
     real              :: xc
     real              :: yc
     real              :: er
     integer           :: ix
     integer           :: jx
     integer           :: kx
     integer           :: it
     integer           :: xsep
     integer           :: ysep
     integer           :: trec
     real              :: dx
     real              :: dy
     real              :: dt
     integer      , dimension(:), pointer :: isep
     integer      , dimension(:), pointer :: jsep
     real         , dimension(:), pointer :: lev
     type(MINFO_T), dimension(:), pointer :: Tinfo
     type(MINFO_V), dimension(:), pointer :: Vinfo
  end type MINFO

  integer :: Model_ix, Model_i
  integer :: Model_jx, Model_j
  integer :: Model_kx, Model_k
  integer :: Model_it, Model_t
  character(len=24) :: Model_date

  integer, parameter :: MaxModelData = 10
  integer, parameter :: MaxInFiles   = 1000

  integer, parameter :: IO  = 10
  integer, parameter :: DIO = 30

  character(len=128), parameter :: c128 = '                            ' &
       &            // '                                               ' &
       &            // '                                               '
  character(len=20) , parameter :: c20 = '                    '
  character(len=100), parameter :: c100 = '                    ' &
       &            // '                                       ' &
       &              // '                                       '
       
  integer, parameter :: MI_NOERROR           = 0
  integer, parameter :: MI_READ_ERROR        = 3001
  integer, parameter :: MI_BINARY_INFO_ERROR = 3002
  integer, parameter :: MI_MM5_INFO_ERROR    = 3003
  integer, parameter :: MI_GRIB1_INFO_ERROR  = 3004

  type(MINFO) :: MODEL(MaxModelData)

  public :: MINFO, MINFO_T, MINFO_V
  public :: Model_i, Model_ix
  public :: Model_j, Model_jx
  public :: Model_k, Model_kx
  public :: Model_t, Model_it
  public :: Model_date
  public :: MaxModelData
  public :: MODEL
  public :: DIO

  public :: ReadModelInfo

  real, parameter :: pi  = 3.14159265
  real, parameter :: d2r = 0.01745329252
  real, parameter :: MI_Undefined = -1.e+35

contains
  integer function ReadModelInfo( info )
    character(len=*), intent(IN)    :: info

    integer :: status, DataType, ip, nstr, cIO
    integer :: m = 0
    logical :: FileExist
    character(len=128) :: buf
    character(len=20)  :: ctitle
    character(len=100) :: cvalue

    m   = m + 1
    buf = c128
    cIO = IO + m - 1
    ReadModelInfo = MI_READ_ERROR

    if( m == 1 ) then
       MODEL%TotalInfo = 1
    else
       if( m > MODEL(1)%TotalInfo ) MODEL%TotalInfo = m
    end if

    inquire( file=info, EXIST=FileExist )
    if( .not. FileExist ) return
    open( cIO, file=info, IOSTAT=status )
    if( status /= 0 ) return

    MODEL(m)%Signed = .true.
    do
       read( cIO, '(a)', IOSTAT=status ) buf
       if( status /= 0 ) exit

       ip = 0
       DataType = InquireStringType( buf, ip )
       if( DataType == TypeChar ) then
          ctitle = c20
          status = InquireCharacter( buf, ip, nstr, ctitle )
          if( buf(1:8) == 'DataType' ) then
             cvalue   = c100
             DataType = InquireStringType( buf, ip )
             status   = InquireCharacter( buf, ip, nstr, cvalue )

             if( cvalue(1:6) == 'BINARY' ) then
                MODEL(m)%DataType(1:6) = cvalue(1:6)
                call ReadModelInfoBinary( cIO, m, ReadModelInfo )
                exit

             else if( cvalue(1:3) == 'MM5' ) then
                call ReadModelInfoMM5( cIO, m, ReadModelInfo )
                exit

             else if( cvalue(1:5) == 'GRIB1' ) then
                call ReadModelInfoGRIB1( cIO, m, ReadModelInfo )
                exit

             end if

          else
             return
          end if
       end if
    end do

    if( ReadModelInfo /= MI_NOERROR ) return

    if( m == 1 ) then
       Model_i = 1
       Model_j = 1
       Model_k = 1
       Model_t = 1
    end if

    ReadModelInfo = MI_NOERROR
    close( CIO )

    return
  end function ReadModelInfo

  include "module_modelinfo_mm5.f90"
  include "module_modelinfo_binary.f90"
  include "module_modelinfo_grib1.f90"

end module ModelInfo
