!
! Reverse Polish Notation module
!
!           Author: KATO Masaya <mkato@pastel.ocn.ne.jp>
!    FIRST RELEASE: 2007-03-02
!      LAST UPDATE: 2007-03-30
!
module RPN
  use string
  use ModelInfo
  use GeoGraphic
  implicit none
  private

  integer, parameter :: RPN_MaxLineBuffer = 1024

  type RPN_STACK
     integer           :: current
     character(len=80) :: type
     character(len=54) :: unit
     character(len=24) :: cdate
     integer           :: dim
     integer           :: ix
     integer           :: jx
     integer           :: kx
     integer           :: it
     integer           :: model
     integer           :: omodel
     logical           :: staggering
     real              :: undef

     real, dimension(:)      , pointer :: olev
     real, dimension(:,:,:,:), pointer :: val
  end type RPN_STACK

  type RPN_CONST
     integer           :: current
     character(len=28) :: name
     character(len=54) :: unit
     real              :: val
  end type RPN_CONST

  type RPN_EQN
     integer :: current
     character(len=28)   :: name
     character(len=54)   :: unit
     character(len=RPN_MaxLineBuffer) :: def
  end type RPN_EQN

  type Cache4D
     integer :: rank
     integer :: ix
     integer :: jx
     integer :: kx
     integer :: it
     integer :: itime
     integer :: model
     integer :: omodel
     integer :: order
     real    :: undef
     logical :: staggering
     character(len=28) :: name
     character(len=24) :: cdate
     real, dimension(:)      , pointer :: olev
     real, dimension(:,:,:,:), pointer :: val
  end type Cache4D

  ! parameter
  integer, parameter :: RPN_MaxStacks     = 100
  integer, parameter :: RPN_MaxConstants  = 500
  integer, parameter :: RPN_MaxEquations  = 500
  integer, parameter :: RPN_MaxModelName  = STRING_MaxModelName
  integer, parameter :: RPN_MaxVarName    = STRING_MaxVarName
  integer, parameter :: RPN_MaxAxisInfo   = STRING_MaxAxisInfo
  real   , parameter :: RPN_Undefined     = -1.e+35
  integer, parameter :: RPN_NO_MODEL_DATA = -1
  integer, parameter :: RPN_MaxCache      = 20
  !integer, parameter :: RPN_MaxCache      = 6

  ! error code
  integer, parameter :: RPN_NOERROR         = 0
  integer, parameter :: RPN_RDCNST_SYNTAX   = 2001
  integer, parameter :: RPN_RDEQN_SYNTAX    = 2002
  integer, parameter :: RPN_PUSH_ERROR      = 2003
  integer, parameter :: RPN_OPE_ERROR       = 2004
  integer, parameter :: RPN_STACK_ERROR     = 2005
  integer, parameter :: RPN_DIV_BY_ZERO     = 2006
  integer, parameter :: RPN_ZERO_IN_LOG     = 2007
  integer, parameter :: RPN_NEG_IN_SQRT     = 2008
  integer, parameter :: RPN_CALC_ERROR      = 2009
  integer, parameter :: RPN_STACK_OVERFLOW  = 2010
  integer, parameter :: RPN_EMPTY_LINE      = 2011
  integer, parameter :: RPN_INVALID_DATA    = 2012
  integer, parameter :: RPN_NO_READ_DATA    = 2013
  integer, parameter :: RPN_NOT_INITIALIZED = 2014
  integer, parameter :: RPN_NOT_OUT_LEVELS  = 2015
  integer, parameter :: RPN_FUNC_ARG_ERROR  = 2016
  integer, parameter :: RPN_FUNC_OPE_ERROR  = 2017
  integer, parameter :: RPN_NOT_DEFINED     = 2018
  integer, parameter :: RPN_UNDER_CONSTRUCTION = 2019

  integer, parameter :: cio = 11
  integer, parameter :: eio = 12

  logical :: RPN_ForceUndef = .false.
  logical :: RPN_debug      = .true.

  integer :: odr = 1

  integer :: status
  character(len=RPN_MaxLineBuffer) :: buf, buf2
  character(len=RPN_MaxLineBuffer) :: cbuf = ''

  type(Cache4D) :: CacheData4D(RPN_MaxCache)

  ! public structures, variables, functions, subroutines
  public :: RPN_STACK, RPN_CONST, RPN_EQN
  public :: RPN_MaxLineBuffer
  public :: RPN_MaxStacks
  public :: RPN_MaxConstants
  public :: RPN_MaxEquations
  public :: RPN_MaxModelName
  public :: RPN_MaxVarName
  public :: RPN_MaxAxisInfo
  public :: RPN_ForceUndef
  public :: RPN_Undefined
  public :: RPN_debug

  public :: RPN_NOERROR
  public :: RPN_NO_READ_DATA
  public :: RPN_NOT_OUT_LEVELS
  public :: RPN_FUNC_ARG_ERROR
  public :: RPN_FUNC_OPE_ERROR
  public :: RPN_NOT_DEFINED
  public :: RPN_UNDER_CONSTRUCTION

  public :: RPN_Initialize
  public :: RPN_PushStack
  public :: RPN_ReadConstants
  public :: RPN_ReadEquations
  public :: RPN_OPERATION
  public :: RPN_CALCULATION
  public :: RPN_RestartStacks
  public :: RPN_ClearStacks

  ! interface procedures
  interface RPN_PushStack
     module procedure PushStack0D
     module procedure PushStack4D
  end interface

  interface RPN_ClearStacks
     module procedure ClearStacks
  end interface

  integer :: CacheUseMax = RPN_MaxCache

contains
  integer function RPN_Initialize( stacks, MaxStacks,     &
       &                           consts, MaxConstants,  &
       &                           eqns,   MaxEquations,  &
       &                           im,     ModelInfoFile, &
       &                           ic,     ConstantFile,  &
       &                           ie,     EquationFile,  &
       &                           UseCacheMax            )
    integer         , intent(IN)    :: MaxStacks
    integer         , intent(IN)    :: MaxConstants
    integer         , intent(IN)    :: MaxEquations
    integer         , intent(IN)    :: im
    integer         , intent(IN)    :: ic
    integer         , intent(IN)    :: ie
    type(RPN_STACK) , intent(INOUT) :: stacks(MaxStacks)
    type(RPN_CONST) , intent(INOUT) :: consts(MaxConstants)
    type(RPN_EQN)   , intent(INOUT) :: eqns(MaxEquations)
    character(len=*), intent(IN)    :: ModelInfoFile(im)
    character(len=*), intent(IN)    :: ConstantFile(ic)
    character(len=*), intent(IN)    :: EquationFile(ie)

    integer         , optional, intent(IN)    :: UseCacheMax

    integer :: i

    RPN_Initialize = RPN_NOT_INITIALIZED

    if( present( UseCacheMax ) ) then
       if( UseCacheMax > 0 .and. UseCacheMax < RPN_MaxCache ) then
          CacheUseMax = UseCacheMax
       end if
    end if

    do i = 1, im
       if( ModelInfoFile(i)(1:1) == '' ) exit
       RPN_Initialize = ReadModelInfo( trim( ModelInfoFile(i) ) )
       if( RPN_Initialize /= RPN_NOERROR ) return
    end do

    consts(1)%current = 1
    do i = 1, ic
       if( ConstantFile(i)(1:1) == '' ) exit
       RPN_Initialize = RPN_ReadConstants( consts,         &
            &                              MaxConstants,   &
            &                              ConstantFile(i) )
       if( RPN_Initialize /= RPN_NOERROR ) return
    end do

    eqns(1)%current = 1
    do i = 1, ie
       if( EquationFile(i)(1:1) == '' ) exit
       RPN_Initialize = RPN_ReadEquations( eqns,           &
            &                              MaxEquations,   &
            &                              EquationFile(i) )
       if( RPN_Initialize /= RPN_NOERROR ) return
    end do

    stacks(1)%current = 1
    CacheData4D%rank  = -1

    RPN_Initialize = RPN_NOERROR
    return
  end function RPN_Initialize

  integer function  RPN_ReadConstants( consts, ic, file )
    integer                        , intent(IN)    :: ic
    type(RPN_CONST) , dimension(ic), intent(INOUT) :: consts
    character(len=*)               , intent(IN)    :: file

    integer :: n, ip, Type, nstr

    n  = consts(1)%current
    if( n < 1 ) n = 1
    RPN_ReadConstants = RPN_RDCNST_SYNTAX

    if( ic < 2 ) return
    open( cio, file=file )
    do
       buf = cbuf
       ip  = 1

       read( cio, '(a)', IOSTAT=status ) buf
       if( status /= 0 ) exit

       buf2 = buf
       call ToUpper( buf )

       ! name of constants
       Type = InquireStringType( buf, ip )
       if( Type /= TypeChar ) cycle
       if( buf(ip:ip) == '#' ) cycle
       n = n + 1
       consts(n)%name = '                           '
       status = InquireCharacter( buf, ip, nstr, consts(n)%name )

       ! value
       Type = InquireStringType( buf, ip )
       if( Type /= TypeNumber ) then
          n = n - 1
          cycle
       end if
       status = InquireNumber( buf, ip, consts(n)%val )

       ! unit of constants
       consts(n)%unit = 'undefined '
       Type = InquireStringType( buf, ip )
       if( Type == TypeEnd ) then
          consts(n)%unit = trim( buf2(ip+2:) )
       end if

    end do

    consts(1)%current = n
    close( cio )

    RPN_ReadConstants = RPN_NOERROR
  end function RPN_ReadConstants

  integer function RPN_ReadEquations( Eqns, ie, file )
    integer                     , intent(IN)    :: ie
    type(RPN_EQN), dimension(ie), intent(INOUT) :: Eqns
    character(len=*)            , intent(IN)    :: file

    integer :: n, ip, is1, is2, Type, nstr
    character(len=RPN_MaxLineBuffer) :: tmp

    n = Eqns(1)%current
    if( n < 1 ) n = 1
    RPN_ReadEquations = RPN_RDEQN_SYNTAX

    if( ie < 2 ) return
    open( eio, file=file )
    do
       buf = cbuf
       ip  = 1

       read( eio, '(a)', IOSTAT=status ) buf
       if( status /= 0 ) exit

       buf2 = buf
       call ToUpper( buf )

       ! name of equations
       Type = InquireStringType( buf, ip )
       if( Type /= TypeChar ) cycle
       if( buf(ip:ip) == '#' ) cycle
       n = n + 1
       Eqns(n)%name = '                           '
       status = InquireCharacter( buf, ip, nstr, Eqns(n)%name )

       ! definition
       Type = InquireStringType( buf, ip )
       if( Type == TypeEnd .or. Type == TypeError ) then
          n = n - 1
          cycle
       end if
       is1 = ip
       status = InquireCharacter( buf, ip, nstr, tmp )

       do
          Type = InquireStringType( buf, ip )
          if( Type == TypeEnd .or. Type == TypeError ) exit
          status = InquireCharacter( buf, ip, nstr, tmp )
       end do
       eqns(n)%def = buf(is1:ip-1)

       ! unit
       eqns(n)%unit = buf2(ip+1:)

    end do
    close( eio )

    eqns(1)%current   = n
    RPN_ReadEquations = RPN_NOERROR
    return
  end function RPN_ReadEquations

  ! push data into stack
  include "rpn/pushstack.f90"

  ! shurink stack
  subroutine ShurinStacks( stacks, MaxStacks )
    integer                          , intent(IN)    :: MaxStacks
    type(RPN_STACK), dimension(MaxStacks), intent(INOUT) :: stacks

    integer :: ic, i, j, k, t
    integer :: ix, jx, kx, it

    ic = stacks(1)%current
    ix = stacks(ic)%ix
    jx = stacks(ic)%jx
    kx = stacks(ic)%kx
    it = stacks(ic)%it
    if( ic < 3 ) return

    if( stacks(ic-1)%dim == 0 ) then
       stacks(ic-1)%cdate = stacks(ic)%cdate
    end if

!    if( associated( stacks(ic-1)%val ) ) nullify( stacks(ic-1)%val )
    if( associated( stacks(ic-1)%val ) ) deallocate( stacks(ic-1)%val )
    stacks(ic-1)%dim   = stacks(ic)%dim
    stacks(ic-1)%ix    = ix
    stacks(ic-1)%jx    = jx
    stacks(ic-1)%kx    = kx
    stacks(ic-1)%it    = it
    stacks(ic-1)%model = stacks(ic)%model
    stacks(ic-1)%omodel = stacks(ic)%omodel
    stacks(ic-1)%staggering = stacks(ic)%staggering
    stacks(ic-1)%undef = stacks(ic)%undef
!    if( associated( stacks(ic-1)%olev ) ) nullify( stacks(ic-1)%olev )
    if( associated( stacks(ic-1)%olev ) ) deallocate( stacks(ic-1)%olev )
    if( associated( stacks(ic)%olev ) ) then
       allocate( stacks(ic-1)%olev(kx) )
       stacks(ic-1)%olev(1:kx) = stacks(ic)%olev(1:kx)
    end if
    allocate( stacks(ic-1)%val(ix,jx,kx,it) )

    do t = 1, it
       do k = 1, kx
          do j = 1, jx
             do i = 1, ix
                stacks(ic-1)%val(i,j,k,t) = stacks(ic)%val(i,j,k,t)
             end do
          end do
       end do
    end do

    return
  end subroutine ShurinStacks

  !
  subroutine RPN_RestartStacks( stacks, MaxStacks )
    integer                              , intent(IN)    :: MaxStacks
    type(RPN_STACK), dimension(MaxStacks), intent(INOUT) :: stacks

    integer :: i, ic

    ic = stacks(1)%current
    do i = 2, ic
       call ClearStacks( stacks, MaxStacks )
    end do
    
    return
  end subroutine RPN_RestartStacks

  ! clear last stack data
  subroutine ClearStacks( stacks, MaxStacks )
    integer                              , intent(IN)    :: MaxStacks
    type(RPN_STACK), dimension(MaxStacks), intent(INOUT) :: stacks

    integer :: ic

    ic = stacks(1)%current
!    if( associated( stacks(ic)%val )  ) nullify( stacks(ic)%val )
!    if( associated( stacks(ic)%olev ) ) nullify( stacks(ic)%olev )
    if( associated( stacks(ic)%val )  ) deallocate( stacks(ic)%val )
    if( associated( stacks(ic)%olev ) ) deallocate( stacks(ic)%olev )
    stacks(1)%current = ic - 1

    return
  end subroutine ClearStacks

  integer function RPN_CALCULATION( stacks, MaxStacks,    &
       &                            consts, MaxConstants, &
       &                            Eqns  , MaxEquations, &
       &                            rpn_string            )
    integer                                 , intent(IN)    :: MaxStacks
    integer                                 , intent(IN)    :: MaxConstants
    integer                                 , intent(IN)    :: MaxEquations
    type(RPN_STACK), dimension(MaxStacks)   , intent(INOUT) :: stacks
    type(RPN_CONST), dimension(MaxConstants), intent(IN)    :: consts
    type(RPN_EQN)  , dimension(MaxEquations), intent(IN)    :: Eqns
    character(len=*)                        , intent(IN)    :: rpn_string

    integer           :: ip, nstr, Type
    real              :: var
    character(len=50) :: operation

    RPN_CALCULATION = RPN_CALC_ERROR

    ip = 0
    do
       Type = InquireStringType( rpn_string, ip )
       if( Type == TypeNumber ) then
          RPN_CALCULATION = InquireNumber( rpn_string, ip, var )
          RPN_CALCULATION = RPN_PushStack( Stacks, RPN_MaxStacks, var )
          
       else if( Type == TypeChar ) then
          RPN_CALCULATION = InquireCharacter( rpn_string, ip, nstr, operation )
          RPN_CALCULATION = RPN_OPERATION( stacks, MaxStacks,    &
               &                           consts, MaxConstants, &
               &                           Eqns  , MaxEquations, &
               &                           operation(1:nstr+1)    )
          if( RPN_CALCULATION /= 0 ) then
             write(6,*) 'RPN OPERATION ERROR (', RPN_CALCULATION, ')'
             write(6,*) 'OPERATION in  ', operation(1:nstr+1)
             return
          end if

       else if( Type == TypeEnd ) then
          if( stacks(1)%current > 2 ) then
!             RPN_CALCULATION = RPN_STACK_OVERFLOW
             RPN_CALCULATION = RPN_NOERROR
             return

          else if( ip == 1 ) then
             RPN_CALCULATION = RPN_EMPTY_LINE
             return
          end if
          exit
       end if

    end do

    RPN_CALCULATION = RPN_NOERROR
    return
  end function RPN_CALCULATION

  ! handle each operator
  recursive function RPN_OPERATION( stacks, MaxStacks,    &
       &                            consts, MaxConstants, &
       &                            Eqns  , MaxEquations, &
       &                            operation, mna      ) &
       & result( ope )
    integer                                  , intent(IN)    :: MaxStacks
    integer                                  , intent(IN)    :: MaxConstants
    integer                                  , intent(IN)    :: MaxEquations
    type(RPN_STACK) , dimension(MaxStacks)   , intent(INOUT) :: stacks
    type(RPN_CONST) , dimension(MaxConstants), intent(IN)    :: consts
    type(RPN_EQN)   , dimension(MaxEquations), intent(IN)    :: Eqns
    character(len=*)                         , intent(IN)    :: operation
    integer         , optional               , intent(IN)    :: mna

    integer :: ope

    integer :: ii, jj
    integer :: i, j, nstr, Type, ip, ipos, ns, ic, tp, ts, mn
    integer :: ix, jx, kx, it, mode, Model_t_tmp
    real    :: val
    character(len=50) :: roperation

    ope = RPN_OPE_ERROR
    mn  = -1
    if( present( mna ) ) mn = mna

    if( operation(1:1) == '[' ) then
       nstr = len_trim( operation )
       ip   = nstr
       ic   = 0
       tp   = 0
       ts   = -1
!       mn   = -1
       mode = 0
       do i = 2, nstr
          if( mode == 0 ) then
             if( operation(i:i) == ':' ) then
                if( ip == nstr ) ip = i
                mode = 1
             end if

          else if( mode == 1 ) then
             if( operation(i:i) == 'T' ) then
                mode = 2
                ic   = i + 1
             else if( operation(i:i) == 'M' ) then
                mode = 3
                ic   = i + 1
             else if( operation(i:i) == 'S' ) then
                mode = 4
                ic   = i + 1
             else if( operation(i:i) == 'N' ) then
                mode = 5
                ic   = i + 1
             else
                mode = -1
             end if

          else if( mode == 2 ) then
             if( operation(i:i) == ':' .or. operation(i:i) == ']' ) then
                ipos = 1
                ope  = InquireNumber( operation(ic:i-1), ipos, tp )
                mode = 1
             end if

          else if( mode == 3 ) then
             if( operation(i:i) == ':' .or. operation(i:i) == ']' ) then
                ipos = 1
                ope  = InquireNumber( operation(ic:i-1), ipos, odr )
                if( odr < 1 .and. odr > 9 ) odr = 1
                mode = 1
             end if

          else if( mode == 4 ) then
             if( operation(i:i) == ':' .or. operation(i:i) == ']' ) then
                ipos = 1
                ope  = InquireNumber( operation(ic:i-1), ipos, ts )
                mode = 1
             end if

          else if( mode == 5 ) then
             if( operation(i:i) == ':' .or. operation(i:i) == ']' ) then
                ipos = 1
                ope  = InquireNumber( operation(ic:i-1), ipos, mn )
                if( mn > MaxModelData ) mn = -1
                mode = 1
             end if

          end if
       end do
       nstr = ip

!!$       if( operation(nstr-2:nstr-2) == ':' ) then
!!$          odr = ichar(operation(nstr-1:nstr-1)) - ichar('0')
!!$          if( odr < 1 .and. odr > 9 ) odr = 1
!!$!          operation(nstr-2:nstr) == ']  '
!!$          nstr = nstr - 2
!!$       end if


       ! constant value
       do i = 2, consts(1)%current
          if( operation(2:nstr-1) == consts(i)%name(1:nstr-2) .and. &
               & consts(i)%name(nstr-1:nstr-1) == ' ' ) then
             ope = RPN_PushStack( stacks, MaxStacks, consts(i)%val )
             write(6,*) 'RETURN=>', operation(1:nstr), ope
             if( ope == RPN_NOERROR ) return
          end if
       end do

       Model_t_tmp = Model_t
       Model_t     = Model_t + tp
       if( ts > 0 ) Model_t = ts
       if( Model_t < 1 ) Model_t = 1

       ! equation
       do i = 2, Eqns(1)%current
          if( operation(2:nstr-1) == Eqns(i)%name(1:nstr-2) .and. &
               & Eqns(i)%name(nstr-1:nstr-1) == ' ' ) then

             if( RPN_debug ) then
!                do j = 1, RPN_MaxCache
                do j = 1, CacheUseMax
                   if( CacheData4D(j)%rank <= 0 ) cycle
                   write(6,'(a,i3,a)') '###CACHE RANK:', &
                        & CacheData4D(j)%rank, trim( CacheData4D(j)%name )
                end do
                write(6,*)
             end if

!             do j = 1, RPN_MaxCache
             do j = 1, CacheUseMax
                if( operation(2:nstr-1)//' ' &
                     & == CacheData4D(j)%name(1:nstr-1) .and. &
                     & ( Model_t == CacheData4D(j)%itime .or. &
                     &   CacheData4D(j)%cdate == 'YYYY-MM-DD_hh:mm:ss.ssss' ) &
                     & .and. CacheData4D(j)%order == odr ) then
                   ope = RPN_PushStack( stacks,                    &
                        &               MaxStacks,                 &
                        &                CacheData4D(j)%val,       &
                        &                CacheData4D(j)%ix,        &
                        &                CacheData4D(j)%jx,        &
                        &                CacheData4D(j)%kx,        &
                        &                CacheData4D(j)%it,        &
                        &                CacheData4D(j)%undef,     &
                        &                CacheData4D(j)%model,     &
                        &                CacheData4D(j)%omodel,    &
                        &                CacheData4D(j)%olev,      &
                        &                CacheData4D(j)%cdate,     &
                        &                CacheData4D(j)%staggering )
!                   if( CacheData4D(j)%rank /= RPN_MaxCache - 1 ) then
!                      CacheData4D(j)%rank = RPN_MaxCache
                   if( CacheData4D(j)%rank /= CacheUseMax - 1 ) then
                      CacheData4D(j)%rank = CacheUseMax
                      CacheData4D%rank    = CacheData4D%rank - 1
                   end if

                   Model_t = Model_t_tmp
                   ope     = RPN_NOERROR
                   return
                end if
             end do

             ip = 0
             do
                Type = InquireStringType( Eqns(i)%def, ip )
                if( Type == TypeNumber ) then
                   status = InquireNumber( Eqns(i)%def, ip, val )
                   ope    = RPN_PushStack( stacks, MaxStacks, val )

                else if( Type == TypeChar ) then
                   status = InquireCharacter( Eqns(i)%def, ip, ns, roperation )
                   ope    = RPN_OPERATION( stacks, MaxStacks,    &
                        &                  consts, MaxConstants, &
                        &                  Eqns  , MaxEquations, &
                        &                  roperation, mna=mn    )
                   if( ope /= RPN_NOERROR ) return

                else if( Type == TypeEnd .or. Type == TypeError ) then
                   exit
                end if
             end do

!             do j = 1, RPN_MaxCache
             do j = 1, CacheUseMax
                ic = stacks(1)%current
                if( stacks(ic)%dim == 0 ) exit
                if( CacheData4D(j)%rank <= 0 ) then
                   ix = stacks(ic)%ix
                   jx = stacks(ic)%jx
                   kx = stacks(ic)%kx
                   it = stacks(ic)%it

!                   CacheData4D(j)%rank = RPN_MaxCache
                   CacheData4D(j)%rank = CacheUseMax
                   CacheData4D(j)%name = operation(2:nstr-1)

                   if( CacheData4D(j)%kx /= kx .or. &
                        & ( .not. associated( CacheData4D(j)%olev ) ) ) then
                      if( associated( CacheData4D(j)%olev ) ) then
!                         nullify( CacheData4D(j)%olev )
                         deallocate( CacheData4D(j)%olev )
                      end if
                      allocate( CacheData4D(j)%olev(kx) )
                   end if

                   if(    CacheData4D(j)%ix /= ix .or.     &
                        & CacheData4D(j)%jx /= jx .or.     &
                        & CacheData4D(j)%kx /= kx .or.     &
                        & CacheData4D(j)%it /= it .or.     &
                        & ( .not. associated( CacheData4D(j)%val ) ) ) then
                      CacheData4D(j)%ix = ix
                      CacheData4D(j)%jx = jx
                      CacheData4D(j)%kx = kx
                      CacheData4D(j)%it = it
                      if( associated( CacheData4D(j)%val ) ) then
!                         nullify( CacheData4D(j)%val )
                         deallocate( CacheData4D(j)%val )
                      end if
                      allocate( CacheData4D(j)%val(ix,jx,kx,it) )
                   end if

                   do jj = 1,jx
                      do ii = 1, ix
                         CacheData4D(j)%val(ii,jj,1:kx,1:it) = &
                              & stacks(ic)%val(ii,jj,1:kx,1:it)
                      end do
                   end do
                   CacheData4D(j)%itime      = Model_t
                   CacheData4D(j)%cdate      = stacks(ic)%cdate
                   CacheData4D(j)%undef      = stacks(ic)%undef
                   CacheData4D(j)%model      = stacks(ic)%model
                   CacheData4D(j)%omodel     = stacks(ic)%omodel
                   CacheData4D(j)%staggering = stacks(ic)%staggering
                   CacheData4D(j)%order      = odr

                   if( associated( stacks(ic)%olev ) ) then
                      CacheData4D(j)%olev = stacks(ic)%olev
                   end if

                   CacheData4D%rank = CacheData4D%rank - 1
                   exit
                end if
             end do

          end if
       end do
       if( ope /= RPN_NOERROR ) then
          write(6,*) 'OPERATION=>', operation
          ope = RPN_NOT_DEFINED
       end if

       Model_t = Model_t_tmp
       return
    end if

    if( RPN_debug ) then
!       write(6,*) '*************DEBUG(OPERATION)***************'
       write(6,*) operation
!       write(6,*) '********************************************'
!       write(6,*)
    end if

    if( operation(1:5) == 'READ(' ) then
       call ReadFromData( operation, stacks, MaxStacks, ope, odr, mn )
    else if( operation(1:5) == 'FUNC(' ) then
       call ReadFuncData( operation, stacks, MaxStacks, ope )
    else if( operation(1:4) == 'ADD ' ) then
       ope = RPN_ADD( stacks, MaxStacks )
    else if( operation(1:4) == 'SUB ' ) then
       ope = RPN_SUB( stacks, MaxStacks )
    else if( operation(1:4) == 'MUL ' ) then
       ope = RPN_MUL( stacks, MaxStacks )
    else if( operation(1:4) == 'DIV ' ) then
       ope = RPN_DIV( stacks, MaxStacks )
    else if( operation(1:4) == 'POW ' ) then
       ope = RPN_POW( stacks, MaxStacks )
    else if( operation(1:4) == 'ABS ' ) then
       ope = RPN_ABS( stacks, MaxStacks )
    else if( operation(1:4) == 'LOG ' ) then
       ope = RPN_LOG( stacks, MaxStacks )
    else if( operation(1:6) == 'LOG10 ' ) then
       ope = RPN_LOG10( stacks, MaxStacks )
    else if( operation(1:5) == 'SQRT ' ) then
       ope = RPN_SQRT( stacks, MaxStacks )
    else if( operation(1:4) == 'INV ' ) then
       ope = RPN_INV( stacks, MaxStacks )
    else if( operation(1:4) == 'NEG ' ) then
       ope = RPN_NEG( stacks, MaxStacks )
    else if( operation(1:4) == 'EXP ' ) then
       ope = RPN_EXP( stacks, MaxStacks )
    else if( operation(1:3) == 'R2 ' ) then
       ope = RPN_R2( stacks, MaxStacks )
    else if( operation(1:4) == 'D2R ' ) then
       ope = RPN_D2R( stacks, MaxStacks )
    else if( operation(1:4) == 'R2D ' ) then
       ope = RPN_R2D( stacks, MaxStacks )
    else if( operation(1:4) == 'SIN ' ) then
       ope = RPN_SIN( stacks, MaxStacks )
    else if( operation(1:4) == 'COS ' ) then
       ope = RPN_COS( stacks, MaxStacks )
    else if( operation(1:4) == 'TAN ' ) then
       ope = RPN_TAN( stacks, MaxStacks )
    else if( operation(1:5) == 'ACOS ' ) then
       ope = RPN_ACOS( stacks, MaxStacks )
    else if( operation(1:5) == 'ASIN ' ) then
       ope = RPN_ASIN( stacks, MaxStacks )
    else if( operation(1:5) == 'ATAN ' ) then
       ope = RPN_ATAN( stacks, MaxStacks )
    else if( operation(1:6) == 'ATAN2 ' ) then
       ope = RPN_ATAN2( stacks, MaxStacks )
    else if( operation(1:4) == 'DDX ' ) then
       ope = RPN_DDX( stacks, MaxStacks )
    else if( operation(1:6) == 'D2DX2 ' ) then
       ope = RPN_D2DX2( stacks, MaxStacks )
    else if( operation(1:4) == 'DDY ' ) then
       ope = RPN_DDY( stacks, MaxStacks )
    else if( operation(1:6) == 'D2DY2 ' ) then
       ope = RPN_D2DY2( stacks, MaxStacks )
    else if( operation(1:7) == 'SMT121 ' ) then
       ope = RPN_SMT121( stacks, MaxStacks )
    else if( operation(1:4) == 'D2C ' ) then
       ope = RPN_D2C( stacks, MaxStacks )
    else if( operation(1:6) == 'DDX_S ' ) then
       ope = RPN_DDX_S( stacks, MaxStacks )
    else if( operation(1:6) == 'DDY_S ' ) then
       ope = RPN_DDY_S( stacks, MaxStacks )
    else if( operation(1:4) == 'DDZ ' ) then
       ope = RPN_DDZ( stacks, MaxStacks )
    else if( operation(1:6) == 'DDZ_2 ' ) then
       ope = RPN_DDZ_2( stacks, MaxStacks )
    else if( operation(1:5) == 'COPY ' ) then
       ope = RPN_COPY( stacks, MaxStacks )
    else if( operation(1:4) == 'POP ' ) then
       ope = RPN_POP( stacks, MaxStacks )
    else if( operation(1:6) == 'ROTUV ' ) then
       ope = RPN_ROTUV( stacks, MaxStacks, 1. )
    else if( operation(1:7) == 'IROTUV ' ) then
       ope = RPN_ROTUV( stacks, MaxStacks, -1. )
    else if( operation(1:7) == 'ROTUV2 ' ) then
       ope = RPN_ROTUV2( stacks, MaxStacks, 1. )
    else if( operation(1:8) == 'IROTUV2 ' ) then
       ope = RPN_ROTUV2( stacks, MaxStacks, -1. )
    else if( operation(1:7) == 'ROTVEC ' ) then
       ope = RPN_ROTVEC( stacks, MaxStacks )
    else if( operation(1:4) == 'SIM ' ) then
       ope = RPN_SIMILARITY( stacks, MaxStacks )
    else if( operation(1:8) == 'COMPBIG ' ) then
       ope = RPN_COMPBIG( stacks, MaxStacks )
    else if( operation(1:10) == 'COMPSMALL ' ) then
       ope = RPN_COMPSMALL( stacks, MaxStacks )
    else if( operation(1:5) == 'SWAP ' ) then
       ope = RPN_SWAP( stacks, MaxStacks )
    else if( operation(1:6) == 'SHIFT ' ) then
       ope = RPN_SHIFT( stacks, MaxStacks )
    else if( operation(1:5) == 'INFO ' ) then
       ope = RPN_INFO( stacks, MaxStacks )
    else if( operation(1:6) == 'HSTAT ' ) then
       ope = RPN_HSTAT( stacks, MaxStacks )
    else if( operation(1:6) == 'DDLON ' ) then
       ope = RPN_DDLON( stacks, MaxStacks )
    else if( operation(1:6) == 'DDLAT ' ) then
       ope = RPN_DDLAT( stacks, MaxStacks )
    else if( operation(1:7) == 'DDLATC ' ) then
       ope = RPN_DDLATC( stacks, MaxStacks )
    else if( operation(1:8) == 'FILLNAN ' ) then
       ope = RPN_FILLNAN( stacks, MaxStacks )
#ifndef NOFFT
    else if( operation(1:4) == 'FFT ' ) then
       ope = RPN_FFT( stacks, MaxStacks )
    else if( operation(1:5) == 'FFTC ' ) then
       ope = RPN_FFT( stacks, MaxStacks )
    else if( operation(1:5) == 'IFFT ' ) then
       ope = RPN_IFFT( stacks, MaxStacks )
#endif
    else if( operation(1:6) == 'CONJG ' ) then
       ope = RPN_CONJG( stacks, MaxStacks )
    else if( operation(1:5) == 'CMUL ' ) then
       ope = RPN_CMUL( stacks, MaxStacks )
    else if( operation(1:5) == 'CABS ' ) then
       ope = RPN_CABS( stacks, MaxStacks )
    else if( operation(1:5) == 'CADD ' ) then
       ope = RPN_CADD( stacks, MaxStacks )
    else if( operation(1:5) == 'CDIV ' ) then
       ope = RPN_CDIV( stacks, MaxStacks )
    else if( operation(1:5) == 'CEXP ' ) then
       ope = RPN_CEXP( stacks, MaxStacks )
    else if( operation(1:5) == 'CINV ' ) then
       ope = RPN_CINV( stacks, MaxStacks )
    else if( operation(1:5) == 'CSUB ' ) then
       ope = RPN_CSUB( stacks, MaxStacks )
    else if( operation(1:5) == 'CPOW ' ) then
       ope = RPN_CPOW( stacks, MaxStacks )
    else if( operation(1:5) == 'IPOW ' ) then
       ope = RPN_IPOW( stacks, MaxStacks )
    else if( operation(1:6) == 'CIPOW ' ) then
       ope = RPN_CIPOW( stacks, MaxStacks )
    else if( operation(1:6) == 'CCOPY ' ) then
       ope = RPN_CCOPY( stacks, MaxStacks )
    end if

    return
  end function RPN_OPERATION

  ! include files of each mathmatical operation
  include "rpn/add.f90"
  include "rpn/sub.f90"
  include "rpn/mul.f90"
  include "rpn/div.f90"
  include "rpn/pow.f90"
  include "rpn/abs.f90"
  include "rpn/log.f90"
  include "rpn/log10.f90"
  include "rpn/sqrt.f90"
  include "rpn/inv.f90"
  include "rpn/neg.f90"
  include "rpn/exp.f90"
  include "rpn/r2.f90"
  include "rpn/d2r.f90"
  include "rpn/r2d.f90"
  include "rpn/sin.f90"
  include "rpn/cos.f90"
  include "rpn/tan.f90"
  include "rpn/acos.f90"
  include "rpn/asin.f90"
  include "rpn/atan.f90"
  include "rpn/atan2.f90"
  include "rpn/ddx.f90"
  include "rpn/d2dx2.f90"
  include "rpn/ddy.f90"
  include "rpn/d2dy2.f90"
  include "rpn/smt121.f90"
  include "rpn/d2c.f90"
  include "rpn/ddx_s.f90"
  include "rpn/ddy_s.f90"
  include "rpn/ddz.f90"
  include "rpn/ddz_2.f90"
  include "rpn/copy.f90"
  include "rpn/pop.f90"
  include "rpn/rotuv.f90"
  include "rpn/rotuv2.f90"
  include "rpn/rotvec.f90"
  include "rpn/similarity.f90"
  include "rpn/compbig.f90"
  include "rpn/compsmall.f90"
  include "rpn/swap.f90"
  include "rpn/shift.f90"
  include "rpn/info.f90"
  include "rpn/hstat.f90"
  include "rpn/ddlon.f90"
  include "rpn/ddlat.f90"
  include "rpn/ddlatc.f90"
  include "rpn/fillnan.f90"
#ifndef NOFFT
  include "rpn/fft.f90"
  include "rpn/fftc.f90"
  include "rpn/ifft.f90"
#endif
  include "rpn/conjg.f90"
  include "rpn/cmul.f90"
  include "rpn/cabs.f90"
  include "rpn/cadd.f90"
  include "rpn/cdiv.f90"
  include "rpn/cexp.f90"
  include "rpn/cinv.f90"
  include "rpn/csub.f90"
  include "rpn/cpow.f90"
  include "rpn/ipow.f90"
  include "rpn/cipow.f90"
  include "rpn/ccopy.f90"

end module RPN
