!
! String Analyze module
!
!           Author: KATO Masaya <mkato@pastel.ocn.ne.jp>
!    FIRST RELEASE: 2007-03-02
!      LAST UPDATE: 2007-03-30
!
module string
  implicit none
  private

  integer, parameter :: Iplus  =  43   ! ichar( '+' )
  integer, parameter :: Iminus =  45   ! ichar( '-' )
  integer, parameter :: Izero  =  48   ! ichar( '0' )
  integer, parameter :: Inine  =  57   ! ichar( '9' )
  integer, parameter :: Iel    = 101   ! ichar( 'e' )
  integer, parameter :: IEu    =  69   ! ichar( 'E' )
  integer, parameter :: Idl    = 100   ! ichar( 'd' )
  integer, parameter :: IDu    =  68   ! ichar( 'D' )

  integer, parameter :: TypeChar   = 1
  integer, parameter :: TypeNumber = 2
  integer, parameter :: TypeEnd    = 10
  integer, parameter :: TypeError  = -1

  integer, parameter :: STRING_MaxModelName = 20
  integer, parameter :: STRING_MaxVarName   = 20
  integer, parameter :: STRING_MaxAxisInfo  = 20

  integer, parameter :: STRING_NOERROR      = 0
  integer, parameter :: STRING_NUM_SYNTAX   = 1001
  integer, parameter :: STRING_CHAR_SYNTAX  = 1002
  integer, parameter :: STRING_ARCMP_SYNTAX = 1003
  integer, parameter :: STRING_AFUNC_SYNTAX = 1004

  character(len=20), parameter :: c20 = '                    '

  ! variable
  public :: TypeChar, TypeNumber, TypeEnd, TypeError
  public :: STRING_MaxModelName
  public :: STRING_MaxVarName
  public :: STRING_MaxAxisInfo
  public :: STRING_NOERROR

  ! subroutine
  public :: ToUpper

  ! function
  public :: InquireStringType
  public :: InquireNumber
  public :: InquireCharacter
  public :: AnalyzeReadComponent
  public :: AnalyzeFuncComponent

  interface InquireNumber
     module procedure InquireNumberReal
     module procedure InquireNumberInteger
     module procedure InquireNumberInteger2
  end interface

contains
  ! convert all strings into upper case
  subroutine ToUpper( char_var )
    character(len=*), intent(INOUT) :: char_var

    integer :: i, length, code

    length = len( char_var )
    do i = 1, length
       code = ichar( char_var(i:i) )
       if( code >= 97 .and. code <= 122 ) char_var(i:i) = char( code-32 )
    end do

    return
  end subroutine ToUpper

  ! inquire the type
  integer function InquireStringType( str, ip )
    character(len=*), intent(IN)    :: str
    integer         , intent(INOUT) :: ip

    integer :: i, istr, length

    length = len( str )
    if( ip == 0 ) ip = 1
    do i = ip, length
       istr = ichar( str(i:i) )
       if( istr >= Izero .and. istr <= Inine ) then
          InquireStringType = TypeNumber
          ip                = i
          return

       else if( istr == Iplus .or. istr == Iminus ) then
          InquireStringType = TypeNumber
          ip                = i
          return

       else if( istr == ichar( ' ' ) ) then
          cycle

       else if( istr == ichar( '=' ) ) then
          InquireStringType = TypeEnd
          ip                = i
          return

       else
          InquireStringType = TypeChar
          ip                = i
          return

       end if
    end do
    InquireStringType = TypeEnd
    return
  end function InquireStringType

  ! convert character into float
  integer function InquireNumberReal( str, ip, var )
    character(len=*), intent(IN)    :: str
    integer         , intent(INOUT) :: ip
    real            , intent(OUT)   :: var

    integer :: i, istr, length, mode, inum
    integer :: v2p
    real    :: sign1, sign3, v1, v2, v3

    InquireNumberReal = STRING_NUM_SYNTAX
    mode              = 1
    sign1             = 1.
    sign3             = 1.
    var               = 0.
    v1                = 0.
    v2                = 0.
    v2p               = 0
    v3                = 0.

    length = len( str )
    if( ip < 1 .or. ip > length ) return

    do i = ip, length
       istr = ichar( str(i:i) )

       if( istr == Iplus ) then
       else if( istr == Iminus ) then
          if( mode == 1 ) then
             sign1 = -1.
          else if( mode == 4 ) then
             sign3 = -1.
          end if

       else if( istr >= Izero .and. istr <= Inine ) then
          inum = istr - Izero
          if( mode <= 2 ) then
             v1 = v1 * 10 + real( inum )
          else if( mode == 3 ) then
             v2p = v2p - 1
             v2  = v2 + real( inum ) * 10.**v2p
          else if( mode == 4 ) then
             v3 = v3 * 10 + real( inum )
          end if

       else if( istr == ichar( '.' ) ) then
          if( mode >= 3 ) return
          mode = 3

       else if( istr == IEu .or. istr == Iel .or. &
            &   istr == IDu .or. istr == Idl      ) then
          mode = 4

       else if( istr == ichar( ' ' ) ) then
          ip = i + 1
          exit

       else if( istr == ichar( '=' ) ) then
          ip = i
          exit

       else if( istr == ichar( ',' ) ) then
          ip = i + 1
          exit

       else
          return
       end if
    end do

    v3 = sign3 * v3
    var = sign1 * ( v1 + v2 ) * 10.**v3

    InquireNumberReal = STRING_NOERROR
    return
  end function InquireNumberReal

  ! convert character into 4-byte integer
  integer function InquireNumberInteger( str, ip, var )
    character(len=*), intent(IN)    :: str
    integer         , intent(INOUT) :: ip
    integer         , intent(OUT)   :: var

    real :: rtmp

    InquireNumberInteger = InquireNumberReal( str, ip, rtmp )
    var = int( rtmp )
    return
  end function InquireNumberInteger

  ! convert character into 2-byte integer
  integer function InquireNumberInteger2( str, ip, var )
    character(len=*), intent(IN)    :: str
    integer         , intent(INOUT) :: ip
    integer(2)      , intent(OUT)   :: var

    real :: rtmp

    InquireNumberInteger2 = InquireNumberReal( str, ip, rtmp )
    var = int( rtmp )
    return
  end function InquireNumberInteger2

  ! extract the word from string
  integer function InquireCharacter( str, ip, nstr, var )
    character(len=*), intent(IN)    :: str
    integer         , intent(INOUT) :: ip
    integer         , intent(OUT)   :: nstr
    character(len=*), intent(OUT)   :: var

    integer :: i, is, ie, length
    logical :: ReadDef

    InquireCharacter = STRING_CHAR_SYNTAX
    nstr             = 0
    ReadDef          = .false.

    length = len( str )
    if( ip < 1 .or. ip > length ) return
    if( length - ip > 4 .and. str(ip:ip+4) == 'READ(' ) ReadDef = .true.

    is = ip
    ie = ip
    do i = ip, length
       if( ReadDef ) then
          if( str(i:i) == ')' ) then
             ip = i + 1
             ie = i
             exit
          end if

       else if( str(i:i) == ' ' ) then
          ip = i + 1
          ie = i - 1
          exit

       else if( str(i:i) == '=' ) then
          ip = i
          ie = i - 1
          exit
       end if
    end do

    nstr = ie - is + 1
    var  = str(is:ie) // ' '

    InquireCharacter = STRING_NOERROR
    return
  end function InquireCharacter

  integer function AnalyzeFuncComponent( string, delimiter, out, ip, status )
    character(len=*), intent(IN)    :: string
    character(len=1), intent(IN)    :: delimiter
    character(len=*), intent(OUT)   :: out
    integer         , intent(INOUT) :: ip
    integer         , intent(OUT)   :: status

    integer :: i, ie, clen

    AnalyzeFuncComponent = STRING_AFUNC_SYNTAX

    clen = len( string )
    do i = ip, clen
       if( string(i:i) == delimiter(1:1) ) then
          ie     = i - 1
          out    = string(ip:ie)
          status = 0
          ip     = i + 1
          AnalyzeFuncComponent = STRING_NOERROR
          return

       else if( string(i:i) == ')' ) then
          ie     = i - 1
          out    = string(ip:ie)
          status = 1
          ip     = i - 1
          AnalyzeFuncComponent = STRING_NOERROR
          return
       end if
    end do

    out    = string(ip:clen)
    ip     = clen + 1
    status = 2
    AnalyzeFuncComponent = STRING_NOERROR
    return
  end function AnalyzeFuncComponent

  integer function AnalyzeReadComponent( string, Model, Variable, Tinfo )
    character(len=*)          , intent(IN)  :: string
    character(len=*), optional, intent(OUT) :: Model
    character(len=*), optional, intent(OUT) :: Variable
    character(len=*), optional, intent(OUT) :: Tinfo

    character(len=STRING_MaxModelName) :: model_tmp
    character(len=STRING_MaxVarName)   :: var_tmp
    character(len=STRING_MaxAxisInfo)  :: tinfo_tmp

    integer :: i, is, ie, clen
    logical :: ModelRead, VariableRead, TimeRead

    ! initialize
    AnalyzeReadComponent = STRING_ARCMP_SYNTAX
    model_tmp    = c20
    var_tmp      = c20
    tinfo        = c20
    ModelRead    = .false.
    VariableRead = .false.
    TimeRead     = .false.

    clen = len( string )
    is   = 6
    do i = 6, clen
       if( string(i:i) == ')' ) then
          ie = i - 1
          if( .not. ModelRead .or. .not. VariableRead ) return

          if( string(is:is+1) == 'T=' ) then
             tinfo_tmp = string(is+2:ie)
             TimeRead  = .true.
          end if
          exit

       else if( string(i:i) == ',' ) then
          ie = i - 1
          if( .not. ModelRead ) then
             model_tmp = string(is:ie)
             ModelRead = .true.

          else if( .not. VariableRead ) then
             if( string(is:is) == '"' ) is = is + 1
             if( string(ie:ie) == '"' ) ie = ie - 1
             var_tmp      = string(is:ie)
             VariableRead = .true.

          else
             if( string(is:is+1) == 'T=' ) then
                tinfo_tmp = string(is+2:ie)
                TimeRead  = .true.
             end if

          end if

          is = i + 1
       end if
    end do

    if( present( Model ) )    Model    = model_tmp
    if( present( Variable ) ) Variable = var_tmp
    if( present( Tinfo ) .and. TimeRead ) Tinfo = tinfo_tmp

    AnalyzeReadComponent = STRING_NOERROR
    return
  end function AnalyzeReadComponent
end module string
