!
! TIME HANDLING module
!
!  FIRST RELEASE: 2002/09/15
!    LAST UPDATE: 2003/04/21
!         AUTHOR: KATO Masaya <mkato@pastel.ocn.ne.jp>
!
module time_handle
  implicit none
  private

  integer :: ml(12) = (/31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31/)

  ! function
  public :: check_leap_year, JulianDay, get_total_date
  public :: get_now_date

  interface get_total_date
     module procedure get_total_year
     module procedure get_total_month
     module procedure get_total_day
     module procedure get_total_hour
     module procedure get_total_minute
     module procedure get_total_second
  end interface

  interface get_now_date
     module procedure get_now_year
     module procedure get_now_month
     module procedure get_now_day
     module procedure get_now_hour
     module procedure get_now_minute
     module procedure get_now_date_p
  end interface

contains
  integer function check_leap_year( year, calendar )
    integer                   , intent(IN) :: year
    character(len=*), optional, intent(IN) :: calendar

    character(len=10) :: cal = 'gregorian'

    if( present( calendar ) ) cal = calendar
    call tolower( cal )

    if( cal(1:6) == 'julian' ) then
       check_leap_year = year / 4 * 4 / year

    else if( cal(1:9) == 'gregorian' ) then
       check_leap_year = year / 4 * 4 / year - year / 100 * 100 / year &
            + year / 400 * 400 / year

    else if( cal(1:6) == 'noleap' ) then
       check_leap_year = 0

    else
       if( year >= 1583 ) then ! Gregorian
          check_leap_year = year / 4 * 4 / year - year / 100 * 100 / year &
               + year / 400 * 400 / year

       else                    ! Julian
          check_leap_year = year / 4 * 4 / year
       end if
    end if

    return
  end function check_leap_year

  integer function JulianDay( year, month, day, calendar )
    integer            , intent(IN) :: year, month, day
    character, optional, intent(IN) :: calendar

    integer             :: i

    if( present( calendar ) ) then
       ml(2) = 28 + check_leap_year( year, calendar=calendar )
    else
       ml(2) = 28 + check_leap_year( year )
    end if

    JulianDay = 0
    do i = 1, month-1
       JulianDay = JulianDay + ml(i)
    end do

    ml(2)     = 28
    JulianDay = JulianDay + day
    return
  end function JulianDay
  
  real(8) function get_total_year( base_year, now_year, calendar )
    integer                   , intent(IN) :: base_year
    integer                   , intent(IN) :: now_year
    character(len=*), optional, intent(IN) :: calendar

    character(len=10) :: cal = 'gregorian'
    if( present( calendar ) ) cal = calendar

    get_total_year = get_total_date_p( base_year, 1, 1, 0, 0, 0., &
         now_year, 1, 1, 0, 0, 0., 6, calendar=cal )
    return
  end function get_total_year

  real(8) function get_total_month( base_year, base_month, &
       & now_year, now_month, calendar )
    integer                   , intent(IN) :: base_year, base_month
    integer                   , intent(IN) :: now_year,  now_month
    character(len=*), optional, intent(IN) :: calendar

    character(len=10) :: cal = 'gregorian'
    if( present( calendar ) ) cal = calendar

    get_total_month = get_total_date_p( base_year, base_month, 1, 0, 0, 0., &
         now_year, now_month, 1, 0, 0, 0., 5, calendar=cal )
    return
  end function get_total_month

  real(8) function get_total_day( base_year, base_month, base_day, &
       & now_year, now_month, now_day, calendar )
    integer                   , intent(IN) :: base_year, base_month, base_day
    integer                   , intent(IN) :: now_year,  now_month,  now_day
    character(len=*), optional, intent(IN) :: calendar

    character(len=10) :: cal = 'gregorian'
    if( present( calendar ) ) cal = calendar

    get_total_day = get_total_date_p(&
         & base_year, base_month, base_day, 0, 0, 0., &
         & now_year, now_month, now_day, 0, 0, 0., 4, calendar=cal )
    return
  end function get_total_day

  real(8) function get_total_hour( base_year, base_month, base_day, &
       & base_hour, now_year, now_month, now_day, now_hour, calendar )
    integer, intent(IN) :: base_year, base_month, base_day, base_hour
    integer, intent(IN) :: now_year,  now_month,  now_day,  now_hour

    character(len=*), optional, intent(IN) :: calendar
    character(len=10)                      :: cal = 'gregorian'
    if( present( calendar ) ) cal = calendar

    get_total_hour = get_total_date_p(&
         & base_year, base_month, base_day, base_hour, 0, 0., &
         & now_year, now_month, now_day, now_hour, 0, 0., 3, &
         & calendar=calendar )
    return
  end function get_total_hour

  real(8) function get_total_minute(&
       & base_year, base_month, base_day, base_hour, base_minute, &
       & now_year, now_month, now_day, now_hour, now_minute, calendar )
    integer, intent(IN) :: base_year, base_month, base_day, base_hour
    integer, intent(IN) :: base_minute
    integer, intent(IN) :: now_year,  now_month,  now_day,  now_hour
    integer, intent(IN) :: now_minute

    character(len=*), optional, intent(IN) :: calendar
    character(len=10)                      :: cal = 'gregorian'
    if( present( calendar ) ) cal = calendar

    get_total_minute = get_total_date_p(&
         & base_year, base_month, base_day, base_hour, base_minute, 0., &
         & now_year, now_month, now_day, now_hour, now_minute, 0., 2, &
         & calendar=calendar)
    return
  end function get_total_minute

  real(8) function get_total_second(&
       & base_year, base_month, base_day, base_hour, base_minute, &
       & base_second, now_year, now_month, now_day, now_hour, now_minute, &
       & now_second, calendar )
    integer, intent(IN) :: base_year, base_month, base_day, base_hour
    integer, intent(IN) :: base_minute
    integer, intent(IN) :: now_year,  now_month,  now_day,  now_hour
    integer, intent(IN) :: now_minute
    real   , intent(IN) :: base_second
    real   , intent(IN) :: now_second

    character(len=*), optional, intent(IN) :: calendar
    character(len=10)                      :: cal = 'gregorian'
    if( present( calendar ) ) cal = calendar

    get_total_second = get_total_date_p(&
         & base_year, base_month, base_day, base_hour, base_minute, &
         & base_second, now_year, now_month, now_day, now_hour, &
         & now_minute, now_second, 1, calendar=cal )
    return
  end function get_total_second

  real(8) function get_total_date_p( base_year, base_month, base_day, &
       & base_hour, base_minute, base_second, &
       & now_year, now_month, now_day, now_hour, now_minute, now_second, id, &
       & calendar )

    integer, intent(IN) :: base_year,base_month,base_day,base_hour,base_minute
    integer, intent(IN) :: now_year, now_month, now_day, now_hour, now_minute
    real   , intent(IN) :: base_second
    real   , intent(IN) :: now_second
    integer, intent(IN) :: id

    character(len=*), optional, intent(IN) :: calendar
    character(len=10)                      :: cal = 'gregorian'

    integer :: mode
    real(8) :: base_total, now_total
    integer :: i, leap

    if( present( calendar ) ) cal = calendar
    mode = id
    ! mode: 
    !   1: return total seconds
    !   2: return total minutes
    !   3: return total hours
    !   4: return total days
    !   5: return total months
    !   6: return total years

    ! ml: total days of each month
    ml(2)            = 28
    get_total_date_p = 0.
    base_total       = 0.
    now_total        = 0.

    if( mode < 1 .or. mode > 6 ) mode = 1

    if( mode == 6 ) then
       get_total_date_p = dble( now_year - base_year )
       return

    else if( mode == 5 ) then
       get_total_date_p = &
            & dble( now_year * 12 + now_month - base_year * 12 - base_month )
       return
    end if

    ! year
    i = 0
    do
       i = i + 1
       if( i >= base_year .and. i >= now_year ) exit
       if( i < base_year ) &
            base_total = base_total + 365. &
            & + dble( check_leap_year(i, calendar=cal) )
       if( i < now_year) &
            now_total  = now_total  + 365. &
            & + dble( check_leap_year(i, calendar=cal) )
    end do

    ! month
    ml(2) = 28 + check_leap_year( base_year, calendar=cal )
    do i = 1,base_month-1
       base_total = base_total + dble( ml(i) )
    end do

    ml(2) = 28 + check_leap_year( now_year, calendar=cal )
    do i = 1,now_month-1
       now_total = now_total + dble( ml(i) )
    end do

    ! day
    base_total = base_total + dble( base_day )
    now_total  = now_total  + dble( now_day  )
    if( mode == 4 ) then
       get_total_date_p = now_total - base_total
       return
    end if

    ! hours , minutes and seconds
    if( mode == 3) then
       base_total = base_total * 24. + dble( base_hour )
       now_total  = now_total  * 24. + dble( now_hour  )

    else if( mode == 2 ) then
       base_total = &
            & base_total * 1440. + dble( base_hour* 60 ) + dble( base_minute )
       now_total  = &
            & now_total  * 1440. + dble( now_hour* 60 )  + dble( now_minute )

    else if( mode == 1 ) then
       base_total = base_total * 86400. + dble( base_hour * 3600 ) &
            & + dble( base_minute * 60 ) + dble( base_second )
       now_total  = now_total  * 86400. + dble( now_hour * 3600 ) &
            & + dble( now_minute * 60 ) + dble( now_second )
    end if

    get_total_date_p = now_total - base_total
    ml(2) = 28
    return
  end function get_total_date_p

  subroutine get_now_year( base_year, inc, now_year )
    integer, intent(IN)  :: base_year
    integer, intent(OUT) :: now_year
    integer, intent(IN)  :: inc

    now_year = base_year + inc
    return
  end subroutine get_now_year

  subroutine get_now_month( base_year, base_month, incd, now_year, now_month )
    integer, intent(IN)  :: base_year, base_month
    integer, intent(OUT) :: now_year,  now_month
    integer, intent(IN)  :: incd

    integer :: inc

    now_year = base_year
    inc      = incd + base_month
    do while( inc > 12 )
       now_year = now_year + 1
       inc      = inc - 12
    end do
    now_month = inc
    return
  end subroutine get_now_month

  subroutine get_now_day(base_year, base_month, base_day, inc, now_year, &
       & now_month, now_day, calendar )
    integer, intent(IN)  :: base_year, base_month, base_day
    integer, intent(OUT) :: now_year,  now_month,  now_day
    integer, intent(IN)  :: inc

    character(len=*), optional, intent(IN) :: calendar
    character(len=10)                      :: cal = 'gregorian'

    integer :: now_hour, now_minute
    real    :: now_second
    real(8) :: inc8

    if( present( calendar ) ) cal = calendar

    inc8 = dble(inc) * 86400d0
    call get_now_date_p( base_year, base_month, base_day, 0, 0, 0., inc8, &
         & now_year, now_month, now_day, now_hour, now_minute, now_second, &
         & calendar=cal )
    return
  end subroutine get_now_day

  subroutine get_now_hour( base_year, base_month, base_day, base_hour, inc, &
       & now_year, now_month, now_day, now_hour, calendar )
    integer, intent(IN)  :: base_year, base_month, base_day, base_hour
    integer, intent(OUT) :: now_year,  now_month,  now_day,  now_hour
    integer, intent(IN)  :: inc

    character(len=*), optional, intent(IN) :: calendar
    character(len=10)                      :: cal = 'gregorian'

    integer :: now_minute
    real    :: now_second
    real(8) :: inc8

    if( present( calendar ) ) cal = calendar

    inc8 = dble(inc) * 3600d0
    call get_now_date_p(base_year, base_month, base_day, base_hour, 0, 0., &
         & inc8, now_year, now_month, now_day, now_hour, now_minute, &
         & now_second, calendar=cal )
    return
  end subroutine get_now_hour

  subroutine get_now_minute(base_year, base_month, base_day, base_hour, &
       & base_minute, inc, now_year, now_month, now_day, now_hour, &
       & now_minute, calendar )
    integer, intent(IN)  :: base_year, base_month, base_day
    integer, intent(IN)  :: base_hour, base_minute
    integer, intent(OUT) :: now_year,  now_month,  now_day
    integer, intent(OUT) :: now_hour,  now_minute
    integer, intent(IN)  :: inc

    character(len=*), optional, intent(IN) :: calendar
    character(len=10)                      :: cal = 'gregorian'

    real :: now_second
    real(8) :: inc8

    if( present( calendar ) ) cal = calendar

    inc8 = dble(inc) * 60d0
    call get_now_date_p(base_year, base_month, base_day, base_hour, &
         & base_minute, 0., inc8, now_year, now_month, now_day, now_hour, &
         & now_minute, now_second, calendar=cal )
    return
  end subroutine get_now_minute
    
  subroutine get_now_date_p(base_year, base_month, base_day, base_hour, &
       & base_minute, base_second, incs, now_year, now_month, now_day, &
       & now_hour, now_minute, now_second, calendar )
    integer, intent(IN)  :: base_year, base_month, base_day
    integer, intent(IN)  :: base_hour, base_minute
    integer, intent(OUT) :: now_year, now_month, now_day
    integer, intent(OUT) :: now_hour, now_minute
    real   , intent(IN)  :: base_second
    real   , intent(OUT) :: now_second
    real(8), intent(IN)  :: incs

    character(len=*), optional, intent(IN) :: calendar
    character(len=10)                      :: cal = 'gregorian'

    integer :: i
    real(8) :: up, out, inc

    if( present( calendar ) ) cal = calendar

    inc        = incs
    now_year   = base_year
    now_month  = base_month
    now_day    = base_day
    now_hour   = base_hour
    now_minute = base_minute
    now_second = base_second

    ! second
!    now_second = now_second + mod( inc, 60d0 )
!    inc        = aint( inc / 60d0 )
    inc        = inc + now_second
    now_second = mod( inc, 60d0 )
    inc        = aint( inc / 60d0 )

    ! minute
!    now_minute = now_minute + mod( inc, 60d0 )
!    inc        = aint( inc / 60d0 )
    inc        = inc + now_minute
    now_minute = mod( inc, 60d0 )
    inc        = aint( inc / 60d0 )

    ! hour
!    now_hour = now_hour + mod( inc, 24d0 )
!    inc      = aint( inc / 24d0 )
    inc      = inc + now_hour
    now_hour = mod( inc, 24d0 )
    inc      = aint( inc / 24d0 )

    ! year, month, and day
    ml(2) = 28 + check_leap_year( now_year, calendar=cal )
    do i = 1, now_month-1
       inc = inc + dble( ml(i) )
    end do
    inc = inc + dble( now_day )

    YEAR: do
       ml(2) = 28 + check_leap_year( now_year, calendar=cal )
       MONTH: do i = 1, 12
          if( inc > dble( ml(i) ) ) then
             inc = inc - dble( ml(i) )
          else
             now_month = i
             now_day = int(inc)
             exit YEAR
          end if
       end do MONTH
       now_year = now_year + 1
    end do YEAR

    ml(2) = 28
    return
  end subroutine get_now_date_p

  subroutine tolower( word )
    character(len=*), intent(INOUT) :: word
    integer :: i, length, code

    length = len( word )
    do i = 1, length
       code = ichar( word(i:i) )
       if( code >= 65 .and. code <= 90 ) word(i:i) = char( code + 32 )
    end do
    return
  end subroutine tolower

end module time_handle

