#!/usr/bin/perl -w

use strict;
use Time::Local;

my $define_file = 'def/model_info_GSMjp.def';
my $work_file   = 'hokuriku_5km_draw.txt';

my $verbose  = 1;
my $LST      = 9;
my $LST_name = 'JST';

my $latlon_map = 'LAMCON';

my $ps_width  = 6;   # inch
my $tz_width  = 6.8; # inch
my $tz_height = 4;   # inch
my $tz_valid  = '0/15000';
my $NaN       = -0.1000000E+36;
my $xs = 1;
my $ys = 1;
my $xe = -1;
my $ye = -1;

###############################################################################

if( @ARGV ){
    $work_file = shift @ARGV;
}
if( @ARGV ){
    $define_file = shift @ARGV;
}

my %def = ();
my $tcheck = '';

# command
my $GMT_path = '/usr/local/gmt4/bin';
my $CMD = {};
$CMD->{gmtset}     = $GMT_path . '/gmtset';
$CMD->{grdimage}   = $GMT_path . '/grdimage';
$CMD->{grdbarb}    = $GMT_path . '/grdbarb';
$CMD->{grdvector}  = $GMT_path . '/grdvector';
$CMD->{grdcontour} = $GMT_path . '/grdcontour';
$CMD->{pscontour}  = $GMT_path . '/pscontour';
$CMD->{psscale}    = $GMT_path . '/psscale';
$CMD->{psxy}       = $GMT_path . '/psxy';
$CMD->{pscoast}    = $GMT_path . '/pscoast';
$CMD->{psbasemap}  = $GMT_path . '/psbasemap';
$CMD->{pstext}     = $GMT_path . '/pstext';
$CMD->{makecpt}    = $GMT_path . '/makecpt';

$CMD->{convert}    = '/usr/bin/convert';

###############################################################################
# constant
my $pi = 3.14159265;
my $dr = $pi / 180.;           # coefficient of conversion degree to radian

###############################################################################
my $default = {};
my $common  = {};
my @list    = ();
my @draw    = ();
my @drawz1d = ();

# command check
&command_check( \$CMD );

# read work list file
&read_work_list( $work_file, \@list );

# check work list data
&check_work_list( \@list, \$common, \@draw, \@drawz1d );

if( defined( $common->{h2dmax} ) ){
    $ps_width = $common->{h2dmax};
}

if( defined( $common->{z1dwidth} ) ){
    $tz_width = $common->{z1dwidth};
}

if( defined( $common->{z1dheight} ) ){
    $tz_height = $common->{z1dheight};
}

if( defined( $common->{NaN} ) ){
    $NaN = $common->{NaN};
}

my $latlon_map_default = $latlon_map;
if( defined( $common->{latlon_map} ) ){
    $latlon_map = $common->{latlon_map};
}

###############################################################################
my ( $nx, $ny, $dx, $dy, $er, $mproj, $tlat1, $tlat2, $tlon,
     $rlat, $rlon, $rx, $ry, @date ) = &read_define_file( $define_file );

&check_data( 'xs', 1, \$common );
&check_data( 'ys', 1, \$common );
&check_data( 'xe', $nx, \$common );
&check_data( 'ye', $ny, \$common );
&check_data( 'LST', $LST, \$common );
&check_data( 'LST_name', $LST_name, \$common );
( defined( $common->{west}  ) ) and ( $common->{xs} = $common->{west} );
( defined( $common->{east}  ) ) and ( $common->{xe} = $common->{east} );
( defined( $common->{south} ) ) and ( $common->{ys} = $common->{south} );
( defined( $common->{north} ) ) and ( $common->{ye} = $common->{north} );

# main loop
my $ndate = @date;

###############################################################################
# time check
if( $tcheck ne '' ){
    $ndate = 0;
    while( <${tcheck}*> ){
	$ndate++;
    }
}
&check_data( 'from', 1     , \$common );
&check_data( 'to'  , $ndate, \$common );

# horizontal plane
my $mode = 'h2d';
for( my $t = 0 ; $t < $ndate ; $t++ ){
    next if( $t < $common->{from}-1 );
    next if( $t > $common->{to}-1 );

    $date[$t] =~ /(\d{4})-(\d\d)-(\d\d)_(\d\d):(\d\d):(\d\d\.\d\d\d\d)/;
    my $year   = $1;
    my $month  = $2;
    my $day    = $3;
    my $hour   = $4;
    my $minute = $5;
    my $second = $6;

    my $ndraw  = @draw;
    my $status = 0;
    my $skip   = 0;
    my $key    = '';
    my $subkey = '';

    my $cur    = {};
    if( defined( $common->{latlon_map} ) ){
	$latlon_map = $common->{latlon_map};
    }else{
	$latlon_map = $latlon_map_default;
    }
    $cur->{draw}      = 1;
    $cur->{overwrite} = undef;
    $cur->{end}       = 0;
    $cur->{ps}        = 'z.ps';

    for( my $i = 0 ; $i < $ndraw ; $i++ ){
	if( $status == 0 && $draw[$i] =~ /<DRAW>/ ){
	    $status = 1;

	    if( defined( $common->{gmtset} ) ){
		unlink( <.gmtdefaults*> );
		foreach my $gmtset ( @{$common->{gmtset}} ){
		    if( $verbose > 1 ){
			print( "$CMD->{gmtset} $gmtset\n" );
		    }
		    system( "$CMD->{gmtset} $gmtset" );
		}
	    }

	}elsif( $status == 1 && $draw[$i] =~ /<\/DRAW>/ ){
	    if( $cur->{end} > 0 && $skip == 0 && $cur->{draw} == 1 ){
		&draw_ps( \$common, \$cur, 'caption',
			  $year, $month, $day, $hour, $minute, $second,
			  $mode );
	    }

	    $status = 0;
	    $skip   = 0;
	    $cur    = {};
	    $cur->{draw}      = 1;
	    $cur->{overwrite} = undef;
	    $cur->{end}       = 0;
	    $cur->{ps}        = 'z.ps';

	}elsif( $status == 1 ){
	    if( $key eq '' && $draw[$i] =~ /<([^\>]+)>/ ){
		$key = $1;
	    }elsif( $key ne '' && $draw[$i] =~ /<\/${key}>/ ){
		if( $key eq 'draw' ){	
		    if( $cur->{draw} != 1 ){
			$skip = 1;
		    }
		}

		if( defined( $cur->{ps} )        &&
		    defined( $cur->{overwrite} ) &&
		    $cur->{end} == 0 ){
#		    $t > 0 ){
		    if( $cur->{ps} !~ /[^\\]\%/ &&
			$cur->{overwrite} == 0  &&
			-e $cur->{ps} ){
			$skip = 1;
		    }
		}

		if( $skip == 0 && $cur->{draw} == 1 ){
		    &draw_ps( \$common, \$cur, $key,
			      $year, $month, $day, $hour, $minute, $second,
			      $mode );
		}
		$key = '';
		
	    }elsif( $key ne '' ){
		if( $subkey eq '' && $draw[$i] !~ /(<|>)/ ){
		    $cur->{$key} = $draw[$i];

		}elsif( $subkey eq '' && $draw[$i] =~ /<([^>]+)>/ ){
		    $subkey = $1;
		    
		}elsif( $subkey ne '' && $draw[$i] =~ /<\/${subkey}>/ ){
		    $subkey = '';

		}elsif( $subkey ne '' ){
		    if( defined( $cur->{$key}->{$subkey} ) ){
			$cur->{$key}->{$subkey} .= "\n$draw[$i]";
		    }else{
			$cur->{$key}->{$subkey} = $draw[$i];
		    }

		}
	    }
	}

    }
}

# time-height cross section
$mode = 'tz';

$date[0] =~ /(\d{4})-(\d\d)-(\d\d)_(\d\d):(\d\d):(\d\d\.\d\d\d\d)/;
my $year   = $1;
my $month  = $2;
my $day    = $3;
my $hour   = $4;
my $minute = $5;
my $second = $6;

my $stime_lst = '';
my $syshift   =  0;
if( $year > 2037 ){
    $syshift = -100;
    $year   -=  100;
}elsif( $year < 1936 ){
    $syshift = 100;
    $year   +=  100;
}elsif( $year < 1970 ){
    $syshift = 40;
    $year   +=  40;
}
$stime_lst = timelocal( $second, $minute, $hour, $day,
			${month}-1, ${year}-1900 ) + $common->{LST} * 3600;

my @stime = localtime( $stime_lst );
$common->{GMT_time} = sprintf( "%04d-%02d-%02dT%02d:%02d:%07.2f/",
			       $stime[5]+1900-$syshift,
			       $stime[4]+1,
			       $stime[3],
			       $stime[2],
			       $stime[1],
			       $stime[0] );

$date[${ndate}-1] =~ /(\d{4})-(\d\d)-(\d\d)_(\d\d):(\d\d):(\d\d\.\d\d\d\d)/;
my $etime_lst = '';
my $eyshift   =  0;
if( $1 < 2037 ){
    if( $1 < 1936 ){
	$eyshift = 100;
	$etime_lst = timelocal( $6, $5, $4, $3, $2-1, $1-1900+$eyshift )
	    + $common->{LST}*3600;
    }elsif( $1 < 1970 ){
	$eyshift = 40;
	$etime_lst = timelocal( $6, $5, $4, $3, $2-1, $1-1900+$eyshift )
	    + $common->{LST}*3600;
    }else{
	$etime_lst = timelocal( $6, $5, $4, $3, $2-1, $1-1900 )
	    + $common->{LST}*3600;
    }
}else{
    $eyshift = -100;
    $etime_lst = timelocal( $6, $5, $4, $3, $2-1, $1-1900+$eyshift )
	+ $common->{LST}*3600;
}
my @etime = localtime( $etime_lst );
$common->{GMT_time} .= sprintf( "%04d-%02d-%02dT%02d:%02d:%07.2f",
				$etime[5]+1900-$eyshift,
				$etime[4]+1,
				$etime[3],
				$etime[2],
				$etime[1],
				$etime[0] );

$common->{GMT_xy} = "-JX${tz_width}i/${tz_height}i";

my $ndraw  = @drawz1d;
my $status = 0;
my $skip   = 0;
my $key    = '';
my $subkey = '';

my $cur    = {};
$cur->{draw}      = 1;
$cur->{overwrite} = 0;
$cur->{end}       = 0;
$cur->{ps}        = 'z.ps';
$cur->{GMT_time}  = $common->{GMT_time};

for( my $i = 0 ; $i < $ndraw ; $i++ ){

    if( $status == 0 && $drawz1d[$i] =~ /<DRAWZ1D>/ ){
	$status = 1;

	if( defined( $common->{gmtset} ) ){
	    unlink( <.gmtdefaults*> );
	    foreach my $gmtset ( @{$common->{gmtset}} ){
		if( $verbose ){
		    print( "$CMD->{gmtset} $gmtset\n" );
		}
		system( "$CMD->{gmtset} $gmtset" );
	    }
	}


    }elsif( $status == 1 && $drawz1d[$i] =~ /<\/DRAWZ1D>/ ){
	if( $cur->{end} > 0 && $skip == 0 && $cur->{draw} == 1 ){
	    &draw_ps( \$common, \$cur, 'caption',
		      $year, $month, $day, $hour, $minute, $second,
		      $mode );
	}

	$status = 0;
	$skip   = 0;
	$cur    = {};
	$cur->{draw}      = 1;
	$cur->{overwrite} = 0;
	$cur->{end}       = 0;
	$cur->{ps}        = 'z.ps';
	$cur->{GMT_time}  = $common->{GMT_time};

    }elsif( $status == 1 ){
	if( $key eq '' && $drawz1d[$i] =~ /<([^\>]+)>/ ){
	    $key = $1;
	}elsif( $key ne '' && $drawz1d[$i] =~ /<\/${key}>/ ){
	    if( $key eq 'draw' ){
		if( $cur->{draw} != 1 ){
		    $skip = 1;
		}
	    }

	    if( $skip == 0 && $cur->{draw} == 1 ){
		if( $key ne 'read' ){
		    $cur->{ps_width} = $tz_width;
		    $cur->{ps_height} = $tz_height;

		    &draw_ps( \$common, \$cur, $key,
			      $year, $month, $day, $hour, $minute, $second,
			      $mode );
		}

	    }
	    $key = '';
		
	}elsif( $key ne '' ){
	    if( $subkey eq '' && $drawz1d[$i] !~ /(<|>)/ ){
		$cur->{$key} = $drawz1d[$i];
		
	    }elsif( $subkey eq '' && $drawz1d[$i] =~ /<([^>]+)>/ ){
		$subkey = $1;
		    
	    }elsif( $subkey ne '' && $drawz1d[$i] =~ /<\/${subkey}>/ ){
		$subkey = '';

	    }elsif( $subkey ne '' ){
		if( defined( $cur->{$key}->{$subkey} ) ){
		    $cur->{$key}->{$subkey} .= "\n$drawz1d[$i]";
		}else{
		    $cur->{$key}->{$subkey} = $drawz1d[$i];
		}

		if( $key eq 'read' ){
		    for( my $i = 0 ; $i < $ndate ; $i++ ){
			$date[$i] =~ /(\d{4})-(\d\d)-(\d\d)_(\d\d):(\d\d):(\d\d\.\d\d\d\d)/;
			my $cyear   = $1;
			my $cmonth  = $2;
			my $cday    = $3;
			my $chour   = $4;
			my $cminute = $5;
			my $csecond = $6;

			&draw_ps( \$common, \$cur, $key,
				  $cyear, $cmonth, $cday, $chour,
				  $cminute, $csecond,
				  $mode );
		    }
		}

	    }
	}
    }

}

###############################################################################
sub draw_ps{
    my ( $common, $cur, $key,
	 $year, $month, $day, $hour, $minute, $second,
	 $mode ) = @_;

    my %mon = (
	       '1' => 'JAN', '01' => 'JAN',
	       '2' => 'FEB', '02' => 'FEB',
	       '3' => 'MAR', '03' => 'MAR',
	       '4' => 'APR', '04' => 'APR',
	       '5' => 'MAY', '05' => 'MAY',
	       '6' => 'JUN', '06' => 'JUN',
	       '7' => 'JUL', '07' => 'JUL',
	       '8' => 'AUG', '08' => 'AUG',
	       '9' => 'SEP', '09' => 'SEP',
	       '10' => 'OCT', '11' => 'NOV', '12' => 'DEC'
	       );

    my %keys = ( 'grdimage'   => 1,
		 'grdcontour' => 1,
		 'grdbarb'    => 1,
		 'grdvector'  => 1,
		 'psxy'       => 1,
		 'pscoast'    => 1,
		 'pscontour'  => 1,
		 'psbasemap'  => 1,
		 'makecpt'    => 1,
		 'gmtset'     => 1,
		 'caption'    => 1,
		 'read'       => 1,
		 );

    return if( -e $$cur->{ps} && $$cur->{end} == 0 && $$cur->{overwrite} == 0 );
    if( $key eq 'convert' && $$cur->{end} < 0 ){
	if( defined( $$cur->{convert}->{option} ) ){
	    $$cur->{png} = $$cur->{ps};
	    $$cur->{png} =~ s/\.ps$/\.png/;
	    if( $verbose ){
		print( "$CMD->{convert} $$cur->{convert}->{option} $$cur->{ps} $$cur->{png}\n" );
	    }
	    system( "$CMD->{convert} $$cur->{convert}->{option} $$cur->{ps} $$cur->{png}" );
	}

	return;
    }

    return unless( defined( $keys{$key} ) );

    return if ( $$cur->{end} < 0 );

    my $decoded_ps = '';
    if( $$cur->{end} == 0 ){
	return if( -e $$cur->{ps} && $$cur->{overwrite} == 0 );

	&check_data( 'xs', $$common->{xs}, $cur );
	&check_data( 'ys', $$common->{ys}, $cur );
	&check_data( 'xe', $$common->{xe}, $cur );
	&check_data( 'ye', $$common->{ye}, $cur );

	my $xs = $$cur->{xs};
	my $xe = $$cur->{xe};
	my $ys = $$cur->{ys};
	my $ye = $$cur->{ye};
	( defined( $$cur->{west}  ) ) and ( $xs = $$cur->{west} );
	( defined( $$cur->{east}  ) ) and ( $xe = $$cur->{east} );
	( defined( $$cur->{south} ) ) and ( $ys = $$cur->{south} );
	( defined( $$cur->{north} ) ) and ( $ye = $$cur->{north} );
#	my $NX = $$cur->{xe} - $$cur->{xs} + 1;
#	my $NY = $$cur->{ye} - $$cur->{ys} + 1;
	my $NX = $xe - $xs + 1;
	my $NY = $ye - $ys + 1;
	my $ps_height = $ps_width;
	my $ps_width0 = $ps_width;
	if( $NX > $NY ){
	    $ps_height = $ps_width0 * $NY / $NX;
	}else{
	    $ps_width0 = $ps_height * $NX / $NY;
	}
	$ps_height *= 0.996;

	if( defined( $$cur->{latlon_map} ) ){
	    $latlon_map = $$cur->{latlon_map};
	}
=backup
	$$cur->{GMT_map} = &set_GMT_coast_region( $$cur->{xs},
						  $$cur->{ys},
						  $$cur->{xe},
						  $$cur->{ye},
						  $mproj,
						  $er,
						  $dx,
						  $dy,
						  $tlat1,
						  $tlat2,
						  $tlon,
						  $rlat,
						  $rlon,
						  $rx,
						  $ry,
						  $ps_width0 );
=cut

	$$cur->{GMT_map} = &set_GMT_coast_region( $xs,
						  $ys,
						  $xe,
						  $ye,
						  $mproj,
						  $er,
						  $dx,
						  $dy,
						  $tlat1,
						  $tlat2,
						  $tlon,
						  $rlat,
						  $rlon,
						  $rx,
						  $ry,
						  $ps_width0 );

	unless( defined( $$cur->{ps_width} ) ){
	    $$cur->{ps_width}  = $ps_width0;
	    $$cur->{ps_height} = $ps_height;
	}

	$$cur->{GMT_xy}  = "-JX${ps_width0}i/${ps_height}i";
	$$cur->{GMT_xy} .= " -R${xs}/${xe}/${ys}/${ye}";

#	if( $key ne 'read'    &&
#	    $key ne 'makecpt' &&
#	    $key ne 'gmtset' ){

#	    $$cur->{ps} = &check_file( $$cur->{ps},
	    $decoded_ps = &check_file( $$cur->{ps},
				       'create',
				       $$common->{psdir},
				       $year,
				       $month,
				       $day,
				       $hour,
				       $minute,
				       $second );
#	}
    }

    if( $decoded_ps ne '' ){
	return if( -e $decoded_ps    &&
		   $$cur->{end} == 0 &&
		   $$cur->{overwrite} == 0 );

	if( $key ne 'read' &&
	    $key ne 'makecpt' &&
	    $key ne 'gmtset' ){
	    $$cur->{ps} = $decoded_ps;
	}
    }else{
	return if( -e $$cur->{ps}    &&
		   $$cur->{end} == 0 &&
		   $$cur->{overwrite} == 0 );
    }

    my ( $grd, $txt, $cpt );

    ###########################################################################
    #  grdimage
    if( $key eq 'grdimage' ){
	$grd = &check_file( $$cur->{grdimage}->{grd},
			    'check',
			    $$common->{outdir},
			    $year,
			    $month,
			    $day,
			    $hour,
			    $minute,
			    $second );
	
	$cpt = &check_file( $$cur->{grdimage}->{cpt}, 'check' );

	if( $mproj eq 'LATLON' ){
	    &GMT_draw( 'grdimage', "$grd $$cur->{GMT_map} -C${cpt}", $cur, 0 );
	}else{
	    &GMT_draw( 'grdimage', "$grd $$cur->{GMT_xy} -C${cpt}", $cur, 0 );
	}

	my $x = $ps_width + 0.25;
	if( defined( $$cur->{grdimage}->{scalepos} ) ){
	    $x = $ps_width + $$cur->{grdimage}->{scalepos};
	}

	if( defined( $$cur->{grdimage}->{cpt0} ) ){
	    my $cpt0 = &check_file( $$cur->{grdimage}->{cpt0}, 'check' );
	    &GMT_draw( 'psscale', "-D${x}i/1.5i/3i/0.2i -C${cpt0} -L -E -I", $cur, 0 );

	    if( defined( $$cur->{grdimage}->{cpt1} ) ){
		my $cpt1 = &check_file( $$cur->{grdimage}->{cpt1}, 'check' );
		my $x2   = $x + 0.9;

		&GMT_draw( 'psscale', "-D${x2}i/1.5i/3i/0.2i -C${cpt1} -L -E -I", $cur, 0 );
	    }

	}else{
	    &GMT_draw( 'psscale', "-D${x}i/1.5i/3i/0.2i -C${cpt} -L -E -I", $cur, 0 );
	}
	$$cur->{$key} = {};

    ###########################################################################
    #  grdcontour
    }elsif( $key eq 'grdcontour' ){
	$grd = &check_file( $$cur->{grdcontour}->{grd},
			    'check',
			    $$common->{outdir},
			    $year,
			    $month,
			    $day,
			    $hour,
			    $minute,
			    $second );

	my $option = "-C1";
	if( defined( $$cur->{grdcontour}->{option} ) ){
	    $option = $$cur->{grdcontour}->{option};
	}

	if( $mproj ne 'LATLON' ){
	    &GMT_draw( 'grdcontour', "$grd $$cur->{GMT_xy} $option", $cur, 0 );
	}else{
	    &GMT_draw( 'grdcontour', "$grd $$cur->{GMT_map} $option", $cur, 0 );
	}
	$$cur->{$key} = {};

    ###########################################################################
    #  grdbarb
    }elsif( $key eq 'grdbarb' ){
	my $grd1 = &check_file( $$cur->{grdbarb}->{grd1},
				'check',
				$$common->{outdir},
				$year,
				$month,
				$day,
				$hour,
				$minute,
				$second );
	my $grd2 = &check_file( $$cur->{grdbarb}->{grd2},
				'check',
				$$common->{outdir},
				$year,
				$month,
				$day,
				$hour,
				$minute,
				$second );

	my $option = "-Q0.1i/0.2i/120/5";
	if( defined( $$cur->{grdbarb}->{option} ) ){
	    $option = $$cur->{grdbarb}->{option};
	}

	if( $mproj ne 'LATLON' ){
	    &GMT_draw( 'grdbarb', "$grd1 $grd2 $$cur->{GMT_xy} $option", $cur, 0 );
	}else{
	    &GMT_draw( 'grdbarb', "$grd1 $grd2 $$cur->{GMT_map} $option", $cur, 0 );
	}
	$$cur->{$key} = {};

    ###########################################################################
    #  grdvector
    }elsif( $key eq 'grdvector' ){
	my $grd1 = &check_file( $$cur->{grdvector}->{grd1},
				'check',
				$$common->{outdir},
				$year,
				$month,
				$day,
				$hour,
				$minute,
				$second );
	my $grd2 = &check_file( $$cur->{grdvector}->{grd2},
				'check',
				$$common->{outdir},
				$year,
				$month,
				$day,
				$hour,
				$minute,
				$second );

	my $option = "-Q0.005i/0.03i/0.02in0.1";
	if( defined( $$cur->{grdvector}->{option} ) ){
	    $option = $$cur->{grdvector}->{option};
	}

	if( $mproj ne 'LATLON' ){
	    &GMT_draw( 'grdvector', "$grd1 $grd2 $$cur->{GMT_xy} $option", $cur, 0 );
	}else{
	    &GMT_draw( 'grdvector', "$grd1 $grd2 $$cur->{GMT_map} $option", $cur, 0 );
	}
	$$cur->{$key} = {};

    ###########################################################################
    #  makecpt
    }elsif( $key eq 'makecpt' ){
	my $option = '-Crainbow -T270/370/5';
	if( defined( $$cur->{$key}->{option} ) ){
	    $option = $$cur->{$key}->{option};
	}
	my $cpt = 'z.cpt';
	if( defined( $$cur->{$key}->{cpt} ) ){
	    $cpt = $$cur->{$key}->{cpt};
	}

	if( $verbose == 1 ){
	    print( "$CMD->{$key} $option > $cpt\n" );
	}
	system( "$CMD->{$key} $option > $cpt" );
	$$cur->{$key} = {};

    ###########################################################################
    #  gmtset
    }elsif( $key eq 'gmtset' ){
	if( $verbose == 1 ){
	    print( "$CMD->{$key} $$cur->{$key}\n" );
	}
	system( "$CMD->{$key} $$cur->{$key}" );
	$$cur->{$key} = {};

    ###########################################################################
    #  pscontour
    }elsif( $key eq 'pscontour' ){
	my $option = "-W1";
	if( defined( $$cur->{$key}->{option} ) ){
	    $option = $$cur->{$key}->{option};
	}
	if( defined( $$cur->{$key}->{cpt} ) ){
	    $option .= " -C$$cur->{$key}->{cpt}";
	}
	
	$$cur->{GMT_time} = $$common->{GMT_time};
	$$cur->{GMT_xy}   = $$common->{GMT_xy};

	return unless( defined( $$cur->{lonp} ) );
	return unless( defined( $$cur->{latp} ) );
	my $lonlat = "$$cur->{lonp}:$$cur->{latp}";


	unless( defined( $$cur->{pscontour}->{valid} ) ){
	    $$cur->{pscontour}->{valid} = $tz_valid;
	}

	my $canvus = "$$cur->{GMT_xy} -R$$cur->{GMT_time}/$$cur->{pscontour}->{valid}";
	my $row  = 0;
	my $zkey = 'z1dz';
	if( defined( $$cur->{$key}->{z1dz} ) ){
	    $row = int( $$cur->{$key}->{z1dz} - 1 );

	}elsif( defined( $$cur->{$key}->{z1d} ) ){
	    $row  = int( $$cur->{$key}->{z1d} - 1 );
	    $zkey = 'z1d';
	}

	$$cur->{NaN} = $NaN;
	if( defined( $$cur->{$key}->{NaN} ) ){
	    $$cur->{NaN} = $$cur->{$key}->{NaN};
	}

	my $draw_text = '';
	foreach my $ctime ( sort keys %{$$cur->{$zkey}->{$lonlat}} ){
	    my @ctime = localtime( $ctime );
	    my $ctime_fmt = sprintf( "%04d-%02d-%02dT%02d:%02d:%07.4f",
				     $ctime[5]+1900,
				     $ctime[4]+1,
				     $ctime[3],
				     $ctime[2],
				     $ctime[1],
				     $ctime[0]
				     );

	    if( $zkey eq 'z1dz' ){
		foreach my $z ( sort { $a <=> $b } keys %{$$cur->{$zkey}->{$lonlat}->{$ctime} } ){
		    my $val = $$cur->{$zkey}->{$lonlat}->{$ctime}->{$z}->{$row};
		    if( $val == $$cur->{NaN} ){
			$val = 'NaN';
		    }
		    $draw_text .= "$ctime_fmt $z $val\n";
		}
	    }
	}

	&GMT_draw_stdout( 'pscontour', "$canvus $option",
			  $draw_text, $cur, 0 );

	if( defined( $$cur->{$key}->{cpt} ) ){
	    my $x = $tz_width + 0.25;
	    &GMT_draw( 'psscale', "-D${x}/1.5i/3i/0.2i -C$$cur->{$key}->{cpt} -L -E -I", $cur, 0 );
	}

	$$cur->{$key} = {};

    ###########################################################################
    #  psxy
    }elsif( $key eq 'psxy' ){
	my $option = "-Sc0.1i";
	if( defined( $$cur->{$key}->{option} ) ){
	    $option = $$cur->{$key}->{option};
	}
	if( defined( $$cur->{$key}->{cpt} ) ){
	    $option .= " -C$$cur->{$key}->{cpt}";
	}

	if( $mode eq 'tz' ){
	    $$cur->{GMT_time} = $$common->{GMT_time};
	    $$cur->{GMT_xy}   = $$common->{GMT_xy};

	    if( defined( $$cur->{$key}->{z1dwidth} ) &&
		defined( $$cur->{$key}->{z1dheight} ) ){
		$$cur->{GMT_xy} = "-JX$$cur->{$key}->{z1dwidth}i/$$cur->{$key}->{z1dheight}i";
	    }elsif( defined( $$cur->{$key}->{z1dwidth} ) ){
		$$cur->{GMT_xy} = "-JX$$cur->{$key}->{z1dwidth}i/${tz_height}i";
	    }elsif( defined( $$cur->{$key}->{z1dheight} ) ){
		$$cur->{GMT_xy} = "-JX${tz_width}i/$$cur->{$key}->{z1dheight}i";
	    }

	    return unless( defined( $$cur->{lonp} ) );
	    return unless( defined( $$cur->{latp} ) );
	    my $lonlat = "$$cur->{lonp}:$$cur->{latp}";


	    unless( defined( $$cur->{$key}->{valid} ) ){
		$$cur->{$key}->{valid} = $tz_valid;
	    }

	    my $canvus = "$$cur->{GMT_xy} -R$$cur->{GMT_time}/$$cur->{$key}->{valid}";

	    my @row  = ( 0 );
	    my $zkey = 'z1dz';

	    if( defined( $$cur->{$key}->{z1dz} ) ){
		@row = split( /\s+/, $$cur->{$key}->{z1dz} );
		( shift @row ) if( $row[0] eq '' );

	    }elsif( defined( $$cur->{$key}->{z1d} ) ){
		@row = split( /\s+/, $$cur->{$key}->{z1d} );
		( shift @row ) if( $row[0] eq '' );
		$zkey = 'z1d';
	    }
	    my $nrow = @row;

	    $$cur->{NaN} = $NaN;
	    if( defined( $$cur->{$key}->{NaN} ) ){
		$$cur->{NaN} = $$cur->{$key}->{NaN};
	    }

	    my $draw_text = '';
	    foreach my $ctime ( sort keys %{$$cur->{$zkey}->{$lonlat}} ){
		my @ctime = localtime( $ctime );
		my $ctime_fmt = sprintf( "%04d-%02d-%02dT%02d:%02d:%07.4f",
					 $ctime[5]+1900,
					 $ctime[4]+1,
					 $ctime[3],
					 $ctime[2],
					 $ctime[1],
					 $ctime[0]
					 );

		if( $zkey eq 'z1dz' ){
		    foreach my $z ( sort { $a <=> $b } keys %{$$cur->{$zkey}->{$lonlat}->{$ctime} } ){
			my $dtext = "$ctime_fmt $z";
			my $nan_check = 0;

			for( my $j = 0 ; $j < $nrow ; $j++ ){
			    my $row = $row[$j] - 1;
			    my $val = $$cur->{$zkey}->{$lonlat}->{$ctime}->{$z}->{$row};
			    if( $val == $$cur->{NaN} ){
				$nan_check = 1;
			    }
			    $dtext .= " $val";
			}
			next if( $nan_check > 0 );

			$draw_text .= "$dtext\n";
		    }
		}elsif( $zkey eq 'z1d' ){
		    my $dtext = "$ctime_fmt";
		    my $nan_check = 0;
		    if( defined( $$cur->{$key}->{level} ) ){
			$dtext .= " $$cur->{$key}->{level}";
		    }

		    for( my $j = 0 ; $j < $nrow ; $j++ ){
			my $row = $row[$j] - 1;
			my $val = $$cur->{$zkey}->{$lonlat}->{$ctime}->{$row};
			if( $val == $$cur->{NaN} ){
			    $nan_check = 1;
			}
			$dtext .= " $val";
		    }
		    next if( $nan_check > 0 );

		    $draw_text .= "$dtext\n";
		}
	    }

	    &GMT_draw_stdout( 'psxy', "$canvus $option",
			      $draw_text, $cur, 0 );

	    if( defined( $$cur->{$key}->{cpt} ) ){
		my $x = $tz_width + 0.25;
		&GMT_draw( 'psscale', "-D${x}/1.5i/3i/0.2i -C$$cur->{$key}->{cpt} -L -E -I", $cur, 0 );
	    }

	    $$cur->{$key} = {};

	}elsif( $mode eq 'h2d' ){
	    my $GMTMAP = $$cur->{GMT_map};
	    if( defined( $$cur->{$key}->{canvas} ) ){
		if( $$cur->{$key}->{canvas} =~ /xy/i ){
		    $GMTMAP = $$cur->{GMT_xy};
		}elsif( $$cur->{$key}->{canvas} =~ /-J/ &&
			$$cur->{$key}->{canvas} =~ /-R/ ){
		    $GMTMAP = $$cur->{$key}->{canvas};
		}
	    }

	    if( defined( $$cur->{$key}->{file} ) ){
		$option = "$$cur->{$key}->{file} $option";
		if( -e $$cur->{$key}->{file} ){
		    &GMT_draw( $key, "$GMTMAP $option", $cur, 0 );
		}

	    }elsif( defined( $$cur->{$key}->{data} ) ){
		&GMT_draw_stdout( 'psxy', "$GMTMAP $option",
				  $$cur->{$key}->{data}, $cur, 0 );
	    }
	    
	    $$cur->{$key} = {};
	}


    ###########################################################################
    #  pscoast
    }elsif( $key eq 'pscoast' ){
	my $option = "-W1";
	if( defined( $$cur->{pscoast}->{option} ) ){
	    $option = $$cur->{pscoast}->{option};
	}
	
	&GMT_draw( 'pscoast', "$$cur->{GMT_map} $option", $cur, 0 );
	$$cur->{pscoast} = {};

    ###########################################################################
    #  psbasemap
    }elsif( $key eq 'psbasemap' ){
	my $option = "";
	if( defined( $$cur->{$key}->{option} ) ){
	    $option = $$cur->{$key}->{option};
	}
	
	&GMT_draw( 'psbasemap', "$$cur->{GMT_map} $option", $cur, 0 );
	$$cur->{$key} = {};

    ###########################################################################
    #  caption
    }elsif( $key eq 'caption' ){
	my $x = $$cur->{ps_width} + 2;
	my $y = $$cur->{ps_height} + 1;
	if( defined( $$cur->{caption}->{hmargin} ) ){
	    $x = $$cur->{ps_width} + $$cur->{caption}->{hmargin};
	}
	if( defined( $$cur->{caption}->{vmargin} ) ){
	    $y = $$cur->{ps_height} + $$cur->{caption}->{vmargin};
	}
	my $GMT_xy = "-JX${x}i/${y}i -R0/${x}/0/${y}";

	my $option = "";
	if( defined( $$cur->{caption}->{option} ) ){
	    $option = $$cur->{caption}->{option};
	}

	if( $verbose ){
	    print( "$CMD->{pstext} $GMT_xy $option -O >> $$cur->{ps}\n" );
	}

	my $date_utc = sprintf( "%02d UTC %02d $mon{$month} %04d",
				$hour, $day, $year );
	my $ldate;
	my $yshift = 0;
	if( $year > 2037 ){
	    $yshift = -100;
	    $year  -=  100;
	}elsif( $year < 1936 ){
	    $yshift = 100;
	    $year  += 100;
	}elsif( $year < 1970 ){
	    $yshift = 40;
	    $year  += 40;
	}
	$ldate = timelocal( $second, $minute, $hour, $day, ${month}-1, ${year}-1900 ) + $$common->{LST} * 3600;
	my @ldate = localtime( $ldate );
	my $month_lst = $ldate[4] + 1;
	my $date_lst = sprintf( "%02d $$common->{LST_name} %02d $mon{$month_lst} %04d",
				$ldate[2], $ldate[3], $ldate[5]+1900-$yshift );
	if( defined( $$cur->{caption}->{HHMM} ) ){
	    $date_lst = sprintf( "%02d:%02d $$common->{LST_name} %02d $mon{$month_lst} %04d",
				 $ldate[2], $ldate[1], $ldate[3], $ldate[5]+1900-$yshift );
	    $date_utc = sprintf( "%02d:%02d UTC %02d $mon{$month} %04d",
				 $hour, $minute, $day, $year );
	}elsif( defined( $$cur->{caption}->{hhmm} ) ){
	    $date_lst = sprintf( "%02d $mon{$month_lst} %04d",
				 $ldate[3], $ldate[5]+1900-$yshift );
	    $date_utc = sprintf( "%02d $mon{$month} %04d",
				 $day, $year );
	}


	$y = $$cur->{ps_height} + 0.1;
	open( TEXT, "| $CMD->{pstext} $GMT_xy $option -O >> $$cur->{ps}" );
	if( defined( $$cur->{caption}->{text} ) ){
	    $$cur->{caption}->{text} =~ s/\%UTC/$date_utc/g;
	    $$cur->{caption}->{text} =~ s/\%LST/$date_lst/g;

	    print TEXT $$cur->{caption}->{text};
	    print TEXT "\n";

	    if( $verbose ){
		print "$$cur->{caption}->{text}";
		print "\n";
	    }
	}
	close( TEXT );

	$$cur->{end} = -1;

    ###########################################################################
    #  read
    }elsif( $key eq 'read' ){
	next unless( defined( $$cur->{lonp} ) );
	next unless( defined( $$cur->{latp} ) );

	my $lonlat = "$$cur->{lonp}:$$cur->{latp}";

	if( defined( $$cur->{$key}->{z1dz} ) ){
	    $txt = &check_file( $$cur->{$key}->{z1dz},
				'skip',
				$$common->{outdir},
				$year,
				$month,
				$day,
				$hour,
				$minute,
				$second );

	    open( TXT, $txt ) || warn "$txt is not found\n";
	    while( <TXT> ){
		chomp;
		my @line = split( /\s+/, $_ );
		( shift @line ) if ( $line[0] eq '' );

		my $t    = shift @line;
		my $z    = shift @line;
		my $lon  = shift @line;
		my $lat  = shift @line;
		my $nrow = @line;

		next if( abs( $lon - $$cur->{lonp} ) > 0.001 );
		next if( abs( $lat - $$cur->{latp} ) > 0.001 );
		
		$t =~ /(\d{4})-(\d\d)-(\d\d)_(\d\d):(\d\d):(\d\d\.\d\d\d\d)/;
		my $stime_lst = timelocal( $6, $5, $4, $3, $2-1, $1-1900 )
		    + $$common->{LST} * 3600;

		for( my $i = 0 ; $i < $nrow ; $i++ ){
		    $$cur->{z1dz}->{$lonlat}->{$stime_lst}->{$z}->{$i} = $line[$i];
		}
	    }

	}elsif( defined( $$cur->{$key}->{z1d} ) ){
	    $txt = &check_file( $$cur->{$key}->{z1d},
				'skip',
				$$common->{outdir},
				$year,
				$month,
				$day,
				$hour,
				$minute,
				$second );
	    open( TXT, $txt ) || warn "$txt is not founda\n";
	    while( <TXT> ){
		chomp;
		my @line = split( /\s+/, $_ );
		( shift @line ) if ( $line[0] eq '' );

		my $t    = shift @line;
		my $lon  = shift @line;
		my $lat  = shift @line;
		my $nrow = @line;

		next if( abs( $lon - $$cur->{lonp} ) > 0.001 );
		next if( abs( $lat - $$cur->{latp} ) > 0.001 );
		
		$t =~ /(\d{4})-(\d\d)-(\d\d)_(\d\d):(\d\d):(\d\d\.\d\d\d\d)/;
		my $stime_lst = timelocal( $6, $5, $4, $3, $2-1, $1-1900 )
		    + $$common->{LST} * 3600;

		for( my $i = 0 ; $i < $nrow ; $i++ ){
		    $$cur->{z1d}->{$lonlat}->{$stime_lst}->{$i} = $line[$i];
		}
	    }
	}
    }

}

###############################################################################
sub GMT_draw_stdout{
    my ( $command, $arg, $stdout, $cur, $end ) = @_;

    my $ps = "-O -K >> $$cur->{ps}";
    if( $end == 1 ){
	if( $$cur->{end} == 0 ){
	    $ps = "> $$cur->{ps}";
	}else{
	    $ps = "-O >> $$cur->{ps}";
	}
    }else{
	if( $$cur->{end} == 0 ){
	    $ps = "-K > $$cur->{ps}";
	}
    }

    if( $verbose ){
	print( "$CMD->{$command} $arg $ps\n" );
    }
    open( PST, "| $CMD->{$command} $arg $ps" );
    print PST $stdout;
    close( PST );

    $$cur->{end}++;
    if( $end == 1 ){ $$cur->{end} = -1; }
}

###############################################################################
sub GMT_draw{
    my ( $command, $arg, $cur, $end ) = @_;

    my $ps = "-O -K >> $$cur->{ps}";
    if( $end == 1 ){
	if( $$cur->{end} == 0 ){
	    $ps = "> $$cur->{ps}";
	}else{
	    $ps = "-O >> $$cur->{ps}";
	}
    }else{
	if( $$cur->{end} == 0 ){
	    $ps = "-K > $$cur->{ps}";
	}
    }

    if( $verbose ){
	print( "$CMD->{$command} $arg $ps\n" );
    }
    system( "$CMD->{$command} $arg $ps" );

    $$cur->{end}++;
    if( $end == 1 ){ $$cur->{end} = -1; }
}

###############################################################################
sub check_file{
    my ( $file, $check,$dir,
	 $year, $month, $day, $hour, $minute, $second ) = @_;

    if( defined( $file ) ){
	my $cfile = $file;
	if( defined( $dir ) ){
	    $cfile = "${dir}/$file";
	}

	if( defined( $year ) ){
	    $year = sprintf( "%04d", $year );
	    $cfile =~ s/%04y/$year/;
	}

	if( defined( $month ) ){
	    $month = sprintf( "%02d", $month );
	    $cfile =~ s/%02m/$month/;
	}

	if( defined( $day ) ){
	    $day = sprintf( "%02d", $day );
	    $cfile =~ s/%02d/$day/;
	}

	if( defined( $hour ) ){
	    $hour = sprintf( "%02d", $hour );
	    $cfile =~ s/%02H/$hour/;
	}

	if( defined( $minute ) ){
	    $minute = sprintf( "%02d", $minute );
	    $cfile =~ s/%02M/$minute/;
	}

	if( -e $cfile ){
	    return ( $cfile );
	}else{
	    if( $check eq 'create' ){
		return ( $cfile );
	    }elsif( $check eq 'skip' ){
		return ( $cfile );
	    }
	    print STDERR "$cfile is not found\n";
	    exit;
	}
    }
    
    print STDERR "$file is not found\n";
    exit;
}

###############################################################################
sub check_data{
    my ( $key, $default, $common ) = @_;

    unless( defined( $$common->{$key} ) ){
	$$common->{$key} = $default;
    }
}

###############################################################################
sub check_work_list{
    my ( $list, $common, $draw, $drawz1d ) = @_;

    my $nlist = @$list;
    my $key   = '';
    for( my $i = 0 ; $i < $nlist ; $i++ ){
	foreach my $def ( keys %def ){
	    if( $list[$i] =~ /\@${def}\@/ ){
		$list[$i] =~ s/\@${def}\@/$def{$def}/g;
	    }
	}

	if( $key eq '' && $$list[$i] =~ /<([^>]+)>/ ){
	    $key = $1;

	    ( $key eq 'DRAW'    ) and ( push( @$draw   , $$list[$i] ) );
	    ( $key eq 'DRAWZ1D' ) and ( push( @$drawz1d, $$list[$i] ) );
	    next;

	}elsif( $$list[$i] =~ /<\/${key}>/ ){
	    ( $key eq 'DRAW'    ) and ( push( @$draw   , $$list[$i] ) );
	    ( $key eq 'DRAWZ1D' ) and ( push( @$drawz1d, $$list[$i] ) );

	    $key = '';
	    next;

	}

	if( $key eq 'DRAW' ){ 
	    push( @$draw, $$list[$i] );

	}elsif( $key eq 'DRAWZ1D' ){ 
	    push( @$drawz1d, $$list[$i] );

	}elsif( $key ne '' && $$list[$i] !~ /(<|>)/ ){
	    &set_work_list( $key, $$list[$i], $common );

	}
	
    }
}

###############################################################################
sub set_work_list{
    my ( $key, $value, $list ) = @_;

    if( $key eq 'gmtset' ){
	push( @{$$list->{$key}}, $value );
	
    }elsif( $key eq 'outdir' ||
	    $key eq 'psdir'  ){
	if( -d $value ){
	    $$list->{$key} = $value;
	}
    }elsif( $key eq 'LST'      ||
	    $key eq 'LST_name' ||
	    $key eq 'NaN'      ||
	    $key eq 'h2dmax'   ||
	    $key eq 'west'     ||
	    $key eq 'east'     ||
	    $key eq 'south'    ||
	    $key eq 'north'    ||
	    $key eq 'latlon_map' ){
	$$list->{$key} = $value;

    }elsif( $key eq 'define_file' ){
	if( -e $value ){
	    $define_file = $value;
	}else{
	    print "$value is not found\n";
	    exit;
	}

    }elsif( $key eq 'overwrite' ||
	    $key eq 'from'      ||
	    $key eq 'to'        ){
	$$list->{$key} = int( $value );

    }

}

###############################################################################
sub read_work_list{
    my ( $worklist, $list, $lv ) = @_;
    my $MaxNestLevel = 5;

    no strict "refs";

    unless( defined( $lv ) ){ $lv = 0; }

    $lv++;
    return if( $lv > $MaxNestLevel );

    my $fh = 'LST' . $lv;
    open( $fh, $worklist ) || return;
    while( <$fh> ){
	chomp;

	if( /^\#include\s+(\S+)/ ){
	    my $wfile = $1;
	    foreach my $def ( keys %def ){
		if( $wfile =~ /\@${def}\@/ ){
		    $wfile =~ s/\@${def}\@/$def{$def}/g;
		}
	    }
	    &read_work_list( $wfile, $list, $lv );
	    next;
	}elsif( /^\#def\s+(\S+)\s+(.*)$/ ){
	    $def{$1} = $2;
	}elsif( /^\#tcheck\s+(\S+)/ ){
	    $tcheck = $1;
	}

	s/([^\#]*)\#.*$/$1/;
	next unless ( /\S/ );

	if( />/ ){
	    s/>/>\n/g;
	    s/</\n</g;
	    s/\n\n/\n/g;
	    foreach my $line ( split( /\n/, $_ ) ){
		next unless ( $line =~ /\S/ );
		push( @{$list}, $line );
	    }
	}else{
	    s/^\s+//;
	    s/\s+$//;
	    push( @{$list}, $_ );
	}
    }
    close( $fh );
}

###############################################################################
#
sub set_GMT_coast_region{
    my ( $xs, $ys, $xe, $ye, $mproj, $er, $dx, $dy, $tlat1, $tlat2, $tlon,
	 $rlat, $rlon, $rx, $ry, $ps_width ) = @_;

    my ( $GMT_map );
    my ( $xc, $yc );

    if( $mproj eq 'LAMCON' ||
	$mproj eq 'POLSTR' ||
	$mproj eq 'MERCAT' ){

	# inquire XC and YC
	( $xc, $yc ) = &ll2xy( $rlon,
			       $rlat,
			       $mproj,
			       $dx,
			       $dy,
			       $tlon,
			       $rlat,
			       0.,
			       0.,
			       $tlat1,
			       $tlat2,
			       $er     );

	$xc = $rx - $xc;
	$yc = $ry - $yc;

	my ( $xp_sw, $yp_sw ) = &xy2ll( $xs,
					$ys,
					$mproj,
					$dx,
					$dy,
					$tlon,
					$rlat,
					$xc,
					$yc,
					$tlat1,
					$tlat2,
					$er     );

	my ( $xp_ne, $yp_ne ) = &xy2ll( $xe,
					$ye,
					$mproj,
					$dx,
					$dy,
					$tlon,
					$rlat,
					$xc,
					$yc,
					$tlat1,
					$tlat2,
					$er     );

	if( $mproj eq 'LAMCON' ){
	    $GMT_map = "-JL${tlon}/${rlat}/${tlat1}/${tlat2}/${ps_width}i";
	    $GMT_map .= " -R${xp_sw}/${yp_sw}/${xp_ne}/${yp_ne}r";
	}elsif( $mproj eq 'POSTR' ){
	}elsif( $mproj eq 'MERCAT' ){
	}
    }elsif( $mproj eq 'LATLON' ){
	my $slon = $rlon - ( $rx - 1 ) * $dx;
	my $slat = $rlat - ( $ry - 1 ) * $dy;
	my $elon = $slon + ( $xe - 1 ) * $dx;
	my $elat = $slat + ( $ye - 1 ) * $dy;

	$tlon  = 0.5 * ( $slon + $elon );
	$tlat1 = $slat;
	$tlat2 = $elat;

	if( $latlon_map eq 'LAMCON' ){
	    $GMT_map = "-JL${tlon}/${rlat}/${tlat1}/${tlat2}/${ps_width}i";
	    $GMT_map .= " -R${slon}/${elon}/${slat}/${elat}";
	}elsif( $latlon_map eq 'ROBINSON' ){
	    $GMT_map = "-JN${tlon}/${ps_width}i";
	    $GMT_map .= " -R${slon}/${elon}/${slat}/${elat}";
	}elsif( $latlon_map =~ /-J/ && $latlon_map =~ /-R/ ){
	    $GMT_map =  $latlon_map;
	}else{
	    print "$latlon_map has not been supported yet.\n";
	    exit;
	}

    }else{
	exit;
    }

    return ( $GMT_map );
}

###############################################################################
sub read_define_file{
    my ( $def_file ) = @_;

    my %CPROJ = ( 2 => 'LAMCON',        # Lambert Conformal
		  1 => 'POLSTR',        # Polar Stereographic
		  3 => 'MERCAT',        # Mercator
		  0 => 'LATLON');

    my $nfiles = 0;
    my ( $nx, $ny, $dx, $dy );
    my ( $tlat1, $tlat2, $tlon, $rlat, $rlon, $rx, $ry, $mproj, $er );
    my ( @date );

    open( DEF, $def_file ) || die "$def_file is not found\n";
    while( <DEF> ){
	chomp;

	if( /^TotalDates/i ){
	    while( <DEF> ){
		chomp;
		last if ( /^\/TotalDates/ );

		push( @date, $_ );
	    }
	}

	if( /^TotalFiles/i ){
	    while( <DEF> ){
		chomp;
		last if( /^\/TotalFiles/ );

		/\S+\s+(\S+)\s+\S+$/;
		my $file = $1;
		( $nfiles++ ) if ( -e $file );
	    }
	}

	if( /GridSizeX\s+(\S+)/ ){
	    $nx = $1;
	}elsif( /GridSizeY\s+(\S+)/ ){
	    $ny = $1;
	}elsif( /DeltaX\s+(\S+)/ ){
	    $dx = $1;
	}elsif( /DeltaY\s+(\S+)/ ){
	    $dy = $1;
	}elsif( /EarthRadius\s+(\S+)/ ){
	    $er = $1;
	}elsif( /MPROJ\s+(\S+)/ ){
	    $mproj = $1;
	}elsif( /TrueLat1\s+(\S+)/ ){
	    $tlat1 = $1;
	}elsif( /TrueLat2\s+(\S+)/ ){
	    $tlat2 = $1;
	}elsif( /TrueLon\s+(\S+)/ ){
	    $tlon = $1;
	}elsif( /RefLat\s+(\S+)/ ){
	    $rlat = $1;
	}elsif( /RefLon\s+(\S+)/ ){
	    $rlon = $1;
	}elsif( /RefX\s+(\S+)/ ){
	    $rx = $1;
	}elsif( /RefY\s+(\S+)/ ){
	    $ry = $1;
	}
    }
    close( DEF );

    return ( $nx,
	     $ny,
	     $dx,
	     $dy,
	     $er,
	     $CPROJ{$mproj},
	     $tlat1,
	     $tlat2,
	     $tlon,
	     $rlat,
	     $rlon,
	     $rx,
	     $ry,
	     @date );
}

###############################################################################
# Cartesian -> lat-lon
#
#  usage: ( $lon, $lat ) = &xy2ll( $x, $y, $cproj, $dx, $dy, $clon, $clat,
#                                  $xc, $yc, $tlat1, $tlat2, $gr );
#
#   Return value:
#     ( $lon, $lat ) : longitude and latitude corresponding to ( $x, $y )
#
#   Argument:
#       $x, $y : map position
#       $cproj : map projection name
#                     'LAMCON' -> Lambert Conformal
#                     'POLSTR' -> Polar Stereographic
#                     'MERCAT' -> Mercator
#          $dx : grid space for the x-axis ( km )
#          $dy : grid space for the y-axis ( km )
#        $clon : center longitude of map
#        $clat : center latitude of map
#          $xc : grid number for the x-axis corresponding to ( $clon, $clat )
#          $yc : grid number for the y-axis corresponding to ( $clon, $clat )
#       $tlat1 : map true latitude 1
#       $tlat2 : map true latitude 2
#          $gr : earth radius (km) ( note; if $gr is omitted, set to $er )
#
sub xy2ll{
    my ( $x, $y, $cproj, $dx, $dy, $clon, $clat, 
	 $xc , $yc, $tlat1, $tlat2, $gr ) = @_;
    my ( $xn, $XC, $YC, $X, $Y, $R, $T1, $T2, $lon, $lat, $a1, $a2 );
    my ( $psi1, $psi2, $pole );

    ( !$gr ) and ( $gr = $er );
    ( $clat >= 0 ) ? ( $pole = 1 ) : ( $pole = -1 );
    $psi1 = $pole * .5 * $pi - $tlat1 * $dr;
    $psi2 = $pole * .5 * $pi - $tlat2 * $dr;

    if( $cproj eq 'LAMCON' ){
	$xn = ( log( cos( $tlat1 * $dr ) ) - log( cos( $tlat2 * $dr ) ) ) /
	    ( log( tan( .25 * $pi - .5 * $dr * abs( $tlat1 ) ) ) - 
	      log( tan( .25 * $pi - .5 * $dr * abs( $tlat2 ) ) ) );

	$YC = $gr / $xn * sin( $psi1 ) * 
	    ( &tan( $pole * $pi * .25 - .5 * $clat * $dr ) /
	      &tan( .5 * $psi1 ) )**$xn;
	$X   = ( $x - $xc ) * $dx;
	$Y   = ( $y - $yc ) * $dy - $YC;
	$R   = sqrt( $X * $X + $Y * $Y );

	$T1  = &tan( .5 * $psi1 );
	$T2  = ( $R * $xn / $gr / abs(sin( $psi1 )) )**(1/$xn);

	$lon = $clon  + atan2( $X, - $pole * $Y ) / $dr / $xn;
	$lat = $pole * 90 - 2 * atan2( $T1 * $T2, 1. ) / $dr;

    }elsif( $cproj eq 'POLSTR'){
	if( abs( $tlat1 ) > 89.999 ){
	    $a1 = 2. * $gr;
	}else{
	    $a1 = $gr * sin( $psi1 ) / &tan( .5 * $psi1 );
	}
	$YC  = $a1 * &tan( $pole * .25 * $pi - .5 * $clat * $dr );
	$X   = ( $x - $xc ) * $dx;
	$Y   = ( $y - $yc ) * $dy - $YC;
	$R   = sqrt( $X * $X + $Y * $Y );
	$a2  = $R / $a1;

	$lon = $clon  + atan2( $X, - $pole * $Y ) / $dr;
	$lat = $pole * ( 90 - 2 * atan2( $a2, 1. ) / $dr );

    }elsif( $cproj eq 'MERCAT' ){
	$a1 = $gr * cos( $tlat1 );
	$YC = &tan( .25 * $pi + .5 * $dr * $clat );
	$YC = $a1 * log( $YC ) / $dy;

	$lon = $clon + $dx / $a1 * ( $x - $xc ) / $dr;
	$lat = 2 * atan2( exp( $dy / $a1 * ( $y - $yc + $YC) ), 1 ) / $dr - 90;
    }

    return( $lon, $lat );
}

###############################################################################
# lat-lon -> Cartesian
#
#  usage: ( $x, $y ) = &xy2ll( $lon, $lat, $cproj, $dx, $dy, $clon, $clat,
#                              $xc, $yc, $tlat1, $tlat2, $gr );
#
#   Return value:
#     ( $x, $y ) : map position corresponding to ( $lon, $lat )
#
#   Argument:
#    $lon, $lat : longitude and latitude
#        $cproj : map projection name
#                      'LAMCON' -> Lambert Conformal
#                      'POLSTR' -> Polar Stereographic
#                      'MERCAT' -> Mercator
#           $dx : grid space for the x-axis ( km )
#           $dy : grid space for the y-axis ( km )
#         $clon : center longitude of map
#         $clat : center latitude of map
#           $xc : grid number for the x-axis corresponding to ( $clon, $clat )
#           $yc : grid number for the y-axis corresponding to ( $clon, $clat )
#        $tlat1 : map true latitude 1
#        $tlat2 : map true latitude 2
#           $gr : earth radius (km) ( note; if $gr is omitted, set to $er )
#
sub ll2xy{
    my ( $lon, $lat, $cproj, $dx, $dy, $clon, $clat, 
	 $xc , $yc, $tlat1, $tlat2, $gr ) = @_;
    my ( $pole, $psi, $psi1, $psi2 );
    my ( $xn, $XC, $YC, $a1, $dlon, $x, $y, $a2 );

    ( !$gr ) and ( $gr = $er );
    ( $clat >= 0 ) ? ( $pole = 1 ) : ( $pole = -1 );
    $psi  = $pole * .5 * $pi - $lat   * $dr;
    $psi1 = $pole * .5 * $pi - $tlat1 * $dr;
    $psi2 = $pole * .5 * $pi - $tlat2 * $dr;

    if( $cproj eq 'LAMCON' ){
	$xn = ( log( cos( $tlat1 * $dr ) ) - log( cos( $tlat2 * $dr ) ) ) /
	    ( log( tan( .25 * $pi - .5 * $dr * abs( $tlat1 ) ) ) - 
	      log( tan( .25 * $pi - .5 * $dr * abs( $tlat2 ) ) ) );
	$YC = $gr / $xn * sin( $psi1 ) * 
	    ( &tan( $pole * $pi * .25 - .5 * $clat * $dr ) /
	      &tan( .5 * $psi1 ) )**$xn;
	$a1   = $gr / $xn * sin( $psi1 ) *
	    ( &tan( $pole * $pi * .25 - .5 * $lat * $dr ) /
	      &tan( .5 * $psi1 ) )**$xn;
	$dlon = $xn * ( $lon - $clon ) * $dr;

	$x  = $xc + $pole * $a1 * sin( $dlon ) / $dx;
	$y  = $yc -         $a1 * cos( $dlon ) / $dy + $YC / $dy;

    }elsif( $cproj eq 'POLSTR' ){

	if( abs( $tlat1 ) > 89.999 ){
	    $a1 = 2. * $gr;
	}else{
	    $a1 = $gr * sin( $psi1 ) / &tan( .5 * $psi1 );
	}
	$a2   = $a1 * &tan( $pole * .25 * $pi - .5 * $lat  * $dr );
	$YC   = $a1 * &tan( $pole * .25 * $pi - .5 * $clat * $dr );
	$dlon = $dr * ( $lon - $clon - 90. );

	$x =   $a2 * cos( $dlon )         / $dx + $xc;
	$y = ( $a2 * sin( $dlon ) + $YC ) / $dy + $yc;

    }elsif( $cproj eq 'MERCAT' ){
	$a1 = $gr * cos( $tlat1 );
	$YC = $a1 * log( &tan( .25 * $pi + .5 * $dr * $clat ) );

	$x = $xc + $a1 * ( $lon - $clon ) * $dr / $dx;
	$y = $yc + ( $a1 * log( &tan( .25 * $pi + .5 * $dr * $lat ) )
		     - $YC ) / $dy;
    }

    return( $x, $y );
}

###############################################################################
# tangent
sub tan{
    my $rad = shift @_;
    my $tan;

    $tan = sin( $rad ) / cos( $rad );
    return( $tan );
}

###############################################################################
# external command check
sub command_check{
    my ( $CMD ) = @_;

    my $err = 0;
    foreach my $path ( keys %$$CMD ){
	unless( -e $$CMD->{$path} ){
	    print "$$CMD->{$path} does not exist.\n";
	    $err++;
	}
    }

    if( $err > 0 ){ exit; }
}

